/*********************************************************************************\
*     File:                 C A N Y M R E S . C P P                               *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 * CanyMRes.cpp: multiresolution Canny border detection
 *
 *  Contains:
 *         void ImgBasic::doCannyMultires24   ();
 *         void ImgBasic::doCannyMultiresGrey ();
 */


#undef NO_COMPILAR_CANNYMULTIRES24
#define _MUESTRA_GRADIENTE_
/******************************************************************************/

/* estructura no usada */
struct struct_punto {
  float* offset; /* puntero directo respecto el origen de la imagen */
  float modulo; /* modulo del gradiente para ese punto             */
  byte tita;    /* angulo 0..255 del gradiente                     */
} ;

typedef struct struct_punto * pPUNTO;


struct struct_segmento {
  /* nptos -> 1 + puntos hallados recoriendo en las dos direcciones */
  /* Nota: nptos =  puntos totales en segmento + 1 */
  int    nptos;
  int   *offset;     /* puntero a la tabla de direcciones de los puntos del
                      * segmento (referidos a la esquina sup. dcha de la
                      * imagen ). Tiene nptos elementos */
  float  media;      /* media del modulo del gradiente en el segmento */
  byte  *tita_grad;  /* puntero a la tabla de angulos (0..256) de los puntos
                      * del segmento (referidos a la esquina sup. dcha de la
                      * imagen ). Tiene nptos elementos */
} ;

typedef struct struct_segmento * pSEGMENTO;

/* NOTA: A nivel de funcionamiento, el inicio del monton de memoria que
 * utiliza SEGMENTO comienza en Imagen.seg[0].offset, y cada vez que
 * se aade un segmento nuevo, la primera direccin de memoria disponible
 * siempre empieza en Imagen.seg[Imagen.nsegs+1].offset
 * La tabla de los ngulos, tita_grad, puede contigua con la tabla de
 * direcciones del segmento(tabla apuntada por offset) o puden estar por
 * las todas las tablas de offset y luego todas las tablas de angulos,
 * nicamente se ha de respetar que la prxima direccin de memoria libre
 * sea Imagen.seg[Imagen.nsegs+1].offset
 */
struct struct_capa {
  int nsegs;        /* numero de segmentos en la capa (o escala de resolucion) */
  struct struct_segmento *seg;
} ;

typedef struct struct_capa CAPAS_MULTIRES [MAX_LEVEL-1];

struct casamiento_segs {
  u_int *segSup;    /* nmero del contorno equivalente en la escala
                     * superior, empezando a contar por 1 (en vez de
                     * 0). Si vale 0, indica que no se ha encontrado
                     * ningn contorno superior que case. */
  float *ptosComun; /* ratio (por unidad) de los puntos casados en 
                     * el segmento superior partido por el nmero de
                     * puntos del segmento inferior */
} ;

typedef struct casamiento_segs ARBOL_CASAMIENTO [MAX_LEVEL-2];

/* Maxima iteracin, para cada nivel, que se puede producir
 * buscando un contorno equivalente en la escala inferior/superior
 * Slo para el algoritmo Canny Multiresolucin
 */

const static int iter_casamiento[3][MAX_LEVEL-2] = {
/*Niveles: 1-2  2-3  3-4  4-5  5-6 */
          { 4,   6,   8,  10,   12 },
          { 2,   2,   4,   6,   10 },
          { 0,   2,   2,   4,    6 } };

static iter_CMres[MAX_LEVEL-2];



/******************************************************************************/

/* Use a once-calculated offset: 
 * vecinoAlineado = vecino_x + vecino_Y*columnas;
 */
#define VECINOS_ALINEADOS 11
static int vecinoAlineado[2][VECINOS_ALINEADOS];
static byte HolguraAngular = HOLGURA_ANG; 

static int* __fastcall SigueContorno( float* mod_grad, byte* tita_grad, int offset_inic,
                               struct struct_segmento* seg );

static void SegmentaGradienteImagen  (float* mod_grad,byte* tita_grad,
                               struct struct_capa* pImagen,
                               int columnas, int filas, float threshold,
                               float escala, int noise_len );

static int fast_analiza_nivel ( struct struct_capa* pSuperior,
                         struct struct_capa* pInferior,
                         struct casamiento_segs* resultado, 
                         int* Segs, byte* Tita, int max_iter,
                         float reclass, int columnas, int filas,
                         int OrdenCasamiento);

static byte* AsignaColoresSegs (struct struct_capa  *ImagenMultires,
                         struct casamiento_segs *Casam,
                         int maxlevel,int minlevel, int minlen,
                         int umbral);

static void __fastcall SegmentacionAbitmapConMedia
    (struct struct_capa* ImagenSegmentada, byte * pic);

static void __fastcall SegmentacionAbitmapConGradiente
    (struct struct_capa* ImagenSegmentada, byte * pic);

static void __fastcall SegmentacionAbitmapConTabla
    (struct struct_capa* Capa, byte* tabla, byte * pic);

static void __fastcall BorraSegmentos
    (struct struct_capa* pImagen, int* mod_grad);


static void __fastcall InicializaCannyMRes(int filas, int columnas,
                                           int holgura, int num_iter);

static int InicializaCasamientoCapa( struct casamiento_segs *CasamientoCapa,
                              struct struct_capa *CapaMultiNivel,
                              int maxlevel, int OrdenCasamiento );
static int InicializaCanyMultiNivel( struct struct_capa *CapaMultiNivel,
                              int columnas, int filas,int maxlevel);
static void LiberaCanyMultiNivel( struct struct_capa *CapaMultiNivel);
static void LiberaCasamientoCapa( struct casamiento_segs *CasamientoCapa,
                           int maxlevel);


#define sizeof_int 4
#define sizeof_float 4
#if (sizeof_int > sizeof_float)
#error Revisar la reutilizacin de mod_grad como matriz de floats
#endif

/******************************************************************************/

byte *ImgBasic::doCannyMultires24( byte *pic24, int columnas, int filas, int minlevel,
                                   int maxlevel, int umbral_grad, int noise_level,
                                   float reclass, int minlen,
                                   int OrdenCasamiento, int debug, int num_iter,
                                   int noise_len, int holg_ang, int noise_comp)
{
 int *imagen,nivel;
 float *mod_grad, *SupressCanny;
 byte  *tita_grad;
 struct struct_capa  CapaMultiNivel[MAX_LEVEL-1];
 struct casamiento_segs CasamientoCapa[MAX_LEVEL-2];
 double factor=1.0;
 char   mensaje[128];
 float    factor_gradiente[MAX_LEVEL] = {5.32,8.0,9.32,10.20,10.55,10.64,10.0};

 ProgressMeter(maxlevel*2, 0, 0, "Canny Multirresolucin");

 /* Iniciacin dimensiones */
 InicializaCannyMRes(filas, columnas, holg_ang, num_iter);
 if (noise_comp < 0) /* gradient factor */
   for(int i=MAX_LEVEL;i;) {i--; factor_gradiente[i]=1.0; }; 

 
 /* Busco la memoria necesaria para poder calcular el gradiente
  * (modulo y argumento)
  */
 imagen= (int*) malloc(filas*columnas*3*sizeof(int));
 if (!imagen) return NULL;
 mod_grad= (float*) busca_memoria (filas*columnas*sizeof(float));
 if (!mod_grad) {
   free(imagen);
   return NULL;
 }
 tita_grad = (byte*) busca_memoria (filas*columnas*sizeof(byte));
 if (!tita_grad) {
   free(imagen);
   free(mod_grad);
   return NULL;
 }
 SupressCanny = (float*) busca_memoria (filas*columnas*sizeof(float));
 if (!SupressCanny) {
   free(imagen);
   free(mod_grad);
   free(tita_grad);
   return NULL;
 }
 if ( !InicializaCanyMultiNivel( CapaMultiNivel, columnas, filas, maxlevel) ) {
    free(imagen);
    free(mod_grad);
    free(tita_grad);
    free(SupressCanny);
    Error("Fallo requiriendo memoria en doCannyMultiresGrey");
    return NULL;
 }
 /* Fin de busqueda de memoria */

 /* CALCULO EL GRADIENTE DE LA IMAGEN SUAVIZADA, EXTRAIGO SEGMENTOS Y LOS CASO
  * CON LOS DEL NIVEL INFERIOR (salvo en el primer nivel)
  */ 

 for ( nivel=1 ; nivel <= maxlevel ; nivel++ ) {
    
    TRACE("SUAVIZADO ESCALA %d\n",nivel);
    if (nivel==1) {
      ConvByteToInt(pic24,imagen,columnas*filas*3);
    } else {
      FastBlur24  ( imagen, columnas, filas, BlurWindow[nivel]);
      /* Actualizo el valor de factor  imagen = imagen ./ factor
       * rango imagen:  0..255 
       * rango imagen:  0..255*factor */
      factor *= FactorWindow[nivel];
    }

    if (debug) {
      ConvIntToByte(imagen,(byte*)mod_grad,columnas*filas*3,(int) factor );
      sprintf(mensaje,"Imagen filtrada %d (ventana %d)",
              nivel,BlurWindow[nivel]);
      ShowImage((byte*)mod_grad,F_FULLCOLOR,mensaje);
    }

    if ( noise_comp < 0) {
      /* 2x2 window */
      FastGradient24(imagen,mod_grad,tita_grad,columnas,filas);      
    } else {
      /* 4x4 window */
      FastGradient24_4x4(imagen,mod_grad,tita_grad,columnas,filas,noise_comp);      
    }
        
    SupresionPuntosCannyMRes(mod_grad,tita_grad,SupressCanny,
		    columnas,filas,
        (float)((double)(factor_gradiente[nivel]*noise_level*COLOR_GRADIENT_FACTOR)*factor
               *ReduccionRuidoNivel[nivel]/EscalaGradient[nivel]));
    
    /* Para grandes ventanas, el valor mnimo del gradiente ha de ser bajo
     * -> utilizo el factor anterior de escala */    
    SegmentaGradienteImagen(SupressCanny,tita_grad,&CapaMultiNivel[nivel-1],
		    columnas, filas, (float) umbral_grad*ReduccionRuidoNivel[nivel],
        /*factor de escala*/
        (float) (factor_gradiente[nivel]*factor*COLOR_GRADIENT_FACTOR/EscalaGradient[nivel]),
        noise_len);

        /* Guarda para la siguiente resolucin el principio de la 
     * pila libre */
    if (nivel < MAX_LEVEL-1) {
      CapaMultiNivel[nivel].nsegs=0;
      CapaMultiNivel[nivel].seg =
        &CapaMultiNivel[nivel-1].seg[CapaMultiNivel[nivel-1].nsegs];
      /* Dado que el siguiente segmento de CapaMultiNivel coincide con el
       * siguiente al ltimo de la capa anterior, y offset ya est inicia-
       * lizado, no es necesario repetir la operacin (similar a "x"="x")*/
      /*CapaMultiNivel[nivel].seg[0].offset =
        CapaMultiNivel[nivel-1].seg[CapaMultiNivel[nivel-1].nsegs].offset;
       */
    }
        /* Ajusto la imagen para evitar sobrepasar el limite de precision
     * de los numeros enteros y la muestro si es necesario */

    if (debug) {
#ifdef _MUESTRA_GRADIENTE_
      sprintf(mensaje,"Orientacin del gradiente %d", nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);

      FitFltInByte(mod_grad,tita_grad, columnas*filas, (float)4);
      sprintf(mensaje,"Mdulo del gradiente %d", nivel);
      ShowImage(tita_grad,F_GREYSCALE,"Gradiente");
#endif  // _MUESTRA_GRADIENTE_
      
      /* Clear solution (to erase lonely points with [1..threshold] values) */  
      memset((void*) tita_grad, 0, filas*columnas); 
      SegmentacionAbitmapConMedia(&CapaMultiNivel[nivel-1],tita_grad);
      sprintf(mensaje,"Puntos contorno %d",nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);

      /* Clear solution (to erase lonely points with [1..threshold] values) */  
      memset((void*) tita_grad, 0, filas*columnas); 
      SegmentacionAbitmapConGradiente(&CapaMultiNivel[nivel-1],tita_grad);
      sprintf(mensaje,"Orientacin del contorno %d",nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);

    }

    /* Ajusto la imagen para evitar sobrepasar el limite de precision
     * de los numeros enteros y la muestro si es necesario */
    /* Dont excess integer 32 bits representation -> decrease factor */
    if ( nivel < maxlevel && 
           (factor_gradiente[nivel+1]*factor*FactorWindow[nivel+1]) > MAX_FACTOR_COLOR) {
        int escala = (int) (log(factor)/log(2.0)) - 7;  
        /* Minimum resolution: 1/(1<<7)= 1/128 (to avoid loss precision)*/
        if (escala >1 ) { 
          factor /= (double) (1<<escala);
          for(int *ptr=imagen, i=filas*columnas*3; i ; i--, ptr++)
            *ptr = *ptr>>escala; 
        }
    }
   ProgressMeter(maxlevel*2, 0, nivel, "Canny Multirresolucin");
 }
 /* free unused memory */
 free(SupressCanny);
 free(imagen);


  /* Asignacin de memoria para el casamiento de segmentos [0..maxlevel-2]*/
 if (!InicializaCasamientoCapa( CasamientoCapa, CapaMultiNivel,
                                maxlevel,  OrdenCasamiento )) {
    free(mod_grad);
    free(tita_grad);
    LiberaCanyMultiNivel( CapaMultiNivel);
    Error("Fallo requiriendo memoria en doCannyMultiresGrey");
    return NULL;
 }

 /* RELACION ENTRE SEGMENTOS */
 /* Borro mod_grad para poder reutilizarlo en
  * fast_analiza_nivel y ahorrar memoria */ 
 /* Si orden de casamiento == 0 => ORDEN_ASCENDENTE
  *    Se busca un punto en el segmento superior que concuerde
  *    con el segmento inferior
  * Si orden de casamiento == 1 => ORDEN_DESCENDENTE
  *    Se busca un punto en el segmento inferior que concuerde
  *    con el segmento superior
  */
 memset((void*) mod_grad, 0, filas*columnas*sizeof(float));
 for ( nivel=0 ; nivel<maxlevel-1 ; nivel++ ) {
    /* RELACION ENTRE SEGMENTOS */
    /* Caso cada segmento con el superior (salvo para nivel 0) */
    if ( !fast_analiza_nivel ( &CapaMultiNivel[nivel+1],
           &CapaMultiNivel[nivel], &CasamientoCapa[nivel],
           (int*) mod_grad, tita_grad, iter_CMres[nivel],
           reclass, columnas, filas,OrdenCasamiento ) ) {
         free(mod_grad);
         free(tita_grad);
         LiberaCanyMultiNivel( CapaMultiNivel );
         LiberaCasamientoCapa( CasamientoCapa, maxlevel);
         Error("Fallo requiriendo memoria");
         return NULL;
      }
      if (nivel < maxlevel)
        BorraSegmentos ( &CapaMultiNivel[nivel+1-OrdenCasamiento],
           (int*) mod_grad);    
   ProgressMeter(maxlevel*2, 0, maxlevel+nivel,
     "Canny Multirresolucin");
 }

 /* Libero memoria y vuelvo */
 free(mod_grad);

 /* Asignacin de memoria para el casamiento de segmentos [0..maxlevel-1]*/
 byte *ColoresSegs = 
         AsignaColoresSegs ( CapaMultiNivel, CasamientoCapa,
                             maxlevel, minlevel, minlen,umbral_grad);
 
 if (ColoresSegs==NULL) {
   free(tita_grad);
   LiberaCanyMultiNivel( CapaMultiNivel );
   LiberaCasamientoCapa( CasamientoCapa, maxlevel);
   Error("Fallo requiriendo memoria");
   return NULL;
 }
 
 /* Clear solution (to erase lonely points with [1..threshold] values)  */ 
 memset((void*) tita_grad, 0, filas*columnas);
 /* Transform the solution into a bitmap */
 SegmentacionAbitmapConTabla(&CapaMultiNivel[0],ColoresSegs,tita_grad);
 
 /* free allocated memory and return */
 free(ColoresSegs);
 LiberaCanyMultiNivel( CapaMultiNivel );
 LiberaCasamientoCapa( CasamientoCapa, maxlevel);
 /* end of freeing memory */
 ProgressMeter(maxlevel*2, 0, maxlevel*2,
     "Canny Multirresolucin");
 return tita_grad;
}

/*********************************************************************************/
 
byte *ImgBasic::doCannyMultiresGrey  ( byte *picGrey, int columnas, int filas,
                                       int minlevel, int maxlevel, int umbral_grad,
                                       int noise_level, float reclass,
                                       int minlen, int OrdenCasamiento, int debug,
                                       int num_iter,int noise_len, int holg_ang,
                                       int noise_comp)
{
 int   *imagen, nivel;
 float *mod_grad, *SupressCanny;
 byte  *tita_grad;
 struct struct_capa  CapaMultiNivel[MAX_LEVEL-1];
 struct casamiento_segs CasamientoCapa[MAX_LEVEL-2];
 double factor=1.0;
 char   mensaje[128];
 float    factor_gradiente[MAX_LEVEL] = {5.32,8.0,9.32,10.20,10.55,10.64,10.0};

 /* Iniciacin dimensiones */
 InicializaCannyMRes(filas,columnas,holg_ang,num_iter);
 if (noise_comp < 0) /* gradient factor */
   for(int i=MAX_LEVEL;i;) {i--; factor_gradiente[i]=1.0;}; 


 /* Busco la memoria necesaria para poder calcular el gradiente
  * (modulo y argumento)
  */
 imagen    = (int*) busca_memoria (filas*columnas*sizeof(int));
 if (!imagen) return NULL;
 mod_grad  = (float*) busca_memoria (filas*columnas*sizeof(float));
 if (!mod_grad) {
   free(imagen);
   Error("Fallo requiriendo memoria en doCannyMultiresGrey");
   return NULL;
 }
 tita_grad = (byte*) busca_memoria (filas*columnas*sizeof(byte));
 if (!tita_grad) {
   free(imagen);
   free(mod_grad);
   Error("Fallo requiriendo memoria en doCannyMultiresGrey");
   return NULL;
 }
 SupressCanny = (float*) busca_memoria (filas*columnas*sizeof(float));
 if (!SupressCanny) {
   free(imagen);
   free(mod_grad);
   free(tita_grad);
   Error("Fallo requiriendo memoria en doCannyMultiresGrey");
   return NULL;
 }
 if ( !InicializaCanyMultiNivel( CapaMultiNivel, columnas, filas, maxlevel) ) {
    free(imagen);
    free(mod_grad);
    free(tita_grad);
    free(SupressCanny);
    Error("Fallo requiriendo memoria en doCannyMultiresGrey");
    return NULL;
 }
 /* Fin de busqueda de memoria */
 
 /* CALCULO EL GRADIENTE DE LA IMAGEN SUAVIZADA, EXTRAIGO SEGMENTOS Y LOS CASO
  * CON LOS DEL NIVEL INFERIOR (salvo en el primer nivel)
  */
 
 for ( nivel=1 ; nivel<=maxlevel ; nivel++ ) {
    
    TRACE("SUAVIZADO ESCALA %d\n",nivel);
    if (nivel==1) {
      ConvByteToInt (picGrey,imagen,columnas*filas);
    } else {
      FastBlurGrey  ( imagen, columnas, filas, BlurWindow[nivel]);
      /* Actualizo el valor de factor  imagen = imagen ./ factor
       * rango imagen:  0..255 
       * rango imagen:  0..255*factor */
      factor *= FactorWindow[nivel];
    }

    if (debug) {
      ConvIntToByte(imagen,tita_grad,columnas*filas,(int) factor);
      sprintf(mensaje,"Imagen filtrada %d (ventana %d)",
              nivel,BlurWindow[nivel]);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);
    }

    if (noise_comp < 0) {
      /* 2x2 window */
      FastGradientGrey(imagen,mod_grad,tita_grad,columnas,filas);      
    } else {
      /* 4x4 window */
      FastGradientGrey4x4(imagen,mod_grad,tita_grad,columnas,filas,noise_comp);         
    }
    
    SupresionPuntosCannyMRes(mod_grad,tita_grad,SupressCanny,
		    columnas,filas,
        (float)((double)factor_gradiente[nivel]*noise_level*factor/EscalaGradient[nivel])
               *ReduccionRuidoNivel[nivel]);
    
    /* Para grandes ventanas, el valor mnimo del gradiente ha de ser bajo
     * -> utilizo el factor anterior de escala */    
    SegmentaGradienteImagen(SupressCanny,tita_grad,&CapaMultiNivel[nivel-1],
		    columnas, filas, (float)umbral_grad*ReduccionRuidoNivel[nivel],
        /*factor de escala*/
        (float) (factor_gradiente[nivel]*factor/EscalaGradient[nivel]), noise_len );

    /* Guarda para la siguiente resolucin el principio de la 
     * pila libre */
    if (nivel<MAX_LEVEL-1) {
      CapaMultiNivel[nivel].nsegs=0;
      CapaMultiNivel[nivel].seg =
        &CapaMultiNivel[nivel-1].seg[CapaMultiNivel[nivel-1].nsegs];
      /* Dado que el siguiente segmento de CapaMultiNivel coincide con el
       * siguiente al ltimo de la capa anterior, y offset ya est inicia-
       * lizado, no es necesario repetir la operacin (similar a "x"="x")*/
      /*CapaMultiNivel[nivel].seg[0].offset =
        CapaMultiNivel[nivel-1].seg[CapaMultiNivel[nivel-1].nsegs].offset;
       */
    }
        /* Ajusto la imagen para evitar sobrepasar el limite de precision
     * de los numeros enteros y la muestro si es necesario */

    if (debug) {

#ifdef _MUESTRA_GRADIENTE_
      sprintf(mensaje,"Orientacin del gradiente %d", nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);

      float escala  = MaximumFlt(mod_grad,filas*columnas);
      escala /= (float) 255.49;
      printf("Factor Gradiente mximo relativo en nivel %d: %f\n",
              nivel,escala/(factor/EscalaGradient[nivel]));

      FitFltInByte(mod_grad,tita_grad, columnas*filas, (float)4);
      sprintf(mensaje,"Mdulo del gradiente %d", nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);
#endif // _MUESTRA_GRADIENTE_
      /* Clear solution (to erase lonely points with [1..threshold] values) */  
      memset((void*) tita_grad, 0, filas*columnas); 
      SegmentacionAbitmapConMedia(&CapaMultiNivel[nivel-1],tita_grad);
      sprintf(mensaje,"Puntos contorno %d",nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);

      /* Clear solution (to erase lonely points with [1..threshold] values) */  
      memset((void*) tita_grad, 0, filas*columnas); 
      SegmentacionAbitmapConGradiente(&CapaMultiNivel[nivel-1],tita_grad);
      sprintf(mensaje,"Orientacin del contorno %d",nivel);
      ShowImage(tita_grad,F_GREYSCALE,mensaje);

    }

    /* Ajusto la imagen para evitar sobrepasar el limite de precision
     * de los numeros enteros y la muestro si es necesario */
    /* Dont excess integer 32 bits representation -> decrease factor */
    if (nivel < maxlevel && 
          (factor_gradiente[nivel]*factor*FactorWindow[nivel+1]) > MAX_FACTOR) {
        int escala = (int) (log(factor)/log(2.0)) - 7;  
        /* Minimum resolution: 1/(1<<7)= 1/128 (to avoid loss precision)*/
        if (escala >1 ) { 
          factor /= (double) (1<<escala);
          for(int *ptr=imagen, i=filas*columnas; i ; i--, ptr++)
            *ptr = *ptr>>escala; 
        }
    }
 }
 /* free unused memory */
 free(SupressCanny);
 free(imagen);


 /* Asignacin de memoria para el casamiento de segmentos [0..maxlevel-2]*/
 if (!InicializaCasamientoCapa( CasamientoCapa, CapaMultiNivel,
                                maxlevel,  OrdenCasamiento)) {
    free(mod_grad);
    free(tita_grad);
    LiberaCanyMultiNivel( CapaMultiNivel);
    Error("Fallo requiriendo memoria en doCannyMultiresGrey");
    return NULL;
 }

 /* RELACION ENTRE SEGMENTOS */
 /* Borro mod_grad para poder reutilizarlo en
  * fast_analiza_nivel y ahorrar memoria */ 
 /* Si orden de casamiento == 0 => ORDEN_ASCENDENTE
  *    Se busca un punto en el segmento superior que concuerde
  *    con el segmento inferior
  * Si orden de casamiento == 1 => ORDEN_DESCENDENTE
  *    Se busca un punto en el segmento inferior que concuerde
  *    con el segmento superior
  */
 memset((void*) mod_grad, 0, filas*columnas*sizeof(int));
 for ( nivel=0 ; nivel<maxlevel-1 ; nivel++ ) {
    /* RELACION ENTRE SEGMENTOS */
    /* Caso cada segmento con el superior (salvo para nivel 0) */
    if ( !fast_analiza_nivel ( &CapaMultiNivel[nivel+1],
           &CapaMultiNivel[nivel], &CasamientoCapa[nivel],
           (int*) mod_grad, tita_grad, iter_CMres[nivel],
           reclass, columnas, filas, OrdenCasamiento ) ) {
         free(mod_grad);
         free(tita_grad);
         LiberaCanyMultiNivel( CapaMultiNivel );
         LiberaCasamientoCapa( CasamientoCapa, maxlevel);
         Error("Fallo requiriendo memoria");
         return NULL;
      } 
    /*
    memset((void*) CasamientoCapa[nivel].segSup, 0,
           (CapaMultiNivel[nivel].nsegs+1)*sizeof(u_int));
    memset((void*) CasamientoCapa[nivel].ptosComun, 0,
           (CapaMultiNivel[nivel].nsegs+1)*sizeof(float));
    */
    if (nivel < maxlevel-2)
        BorraSegmentos ( &CapaMultiNivel[nivel+1-OrdenCasamiento],
          (int*) mod_grad);
 }

 /* Libero memoria que no se utiliza */
 free(mod_grad);

 /* Asignacin de memoria para el casamiento de segmentos [0..maxlevel-1] */
 byte *ColoresSegs = 
         AsignaColoresSegs ( CapaMultiNivel, CasamientoCapa,
                             maxlevel, minlevel, minlen, umbral_grad);
 
 if (ColoresSegs==NULL) {
   free(tita_grad);
   LiberaCanyMultiNivel( CapaMultiNivel );
   LiberaCasamientoCapa( CasamientoCapa, maxlevel);
   Error("Fallo requiriendo memoria en doCannyMultiresGrey");
   return NULL;
 }
 
 /* Clear solution (to erase lonely points with [1..threshold] values)  */ 
 memset((void*) tita_grad, 0, filas*columnas);
 /* Transform the solution into a bitmap */
 SegmentacionAbitmapConTabla(&CapaMultiNivel[0],ColoresSegs,tita_grad); 
 
 /* free allocated memory and return  */
 free(ColoresSegs);
 LiberaCanyMultiNivel( CapaMultiNivel );
 LiberaCasamientoCapa( CasamientoCapa, maxlevel);
 /* end of freeing memory */

 return tita_grad;
}

/*********************************************************************************/
static int* __fastcall SigueContorno( float* mod_grad, byte* tita_grad, int offset_inic,
                               struct struct_segmento* pSeg )
{
  /* Inicializo los punteros, e ubico la tabla de los puntos (offset) */
  int *pPtoEstudiado=pSeg->offset, *pFinSegmento=pSeg->offset, *pCambioSentido;
  /* Sentido = 1  2, en cual se buscan puntos contiguos de las dos posibles */ 
  int UltimoPtoPrimeraPasada;
  byte sentido = 2;
  
  /* Primer punto del segmento -> inicializo media y borro punto */
  *pPtoEstudiado=offset_inic;
  pSeg->media = mod_grad[offset_inic];
  mod_grad[offset_inic]=(float)0;

  /* condiciones necesarias aadir un nuevo punto a uno considerado en el
   * segmento:
   * - que sean puntos contiguos y estn aliniados en direccin perpendicular
   *   a la direccin del gradiente (hay 2 direcciones perpendiculares de sen-
   *   tido contrario -> hay 2 segmentos ptos1 y ptos2 a partir de una semilla).
   * - que la direccin del gradiente se diferencie en menos de HOLGURA_ANG
   *   grados en rango 0..256.
   */
  /* Al inicio de cada bucle de sentido, pFinSegmento debe ser una unidad
   * mayor que pPtoEstudiado */
  pFinSegmento++;
  do {
    /* actualizo los vecinos que tomo para la segunda pasada (sentido=0) */
    sentido--;
    int* pVecinos = (sentido) ? 
      &vecinoAlineado[0][1] : 
      &vecinoAlineado[1][1];
    do {
      /* aumento tita para que los dos vecinos estudiados estn 
       * centrados respecto la direccin del gradiente
       * (se estudian tres octantes: el centrado en la direccin
       * perpendicular al gradiente y los dos adyacentes) */
     byte tita = tita_grad[*pPtoEstudiado];
     int *vecino_tita = pVecinos + 
                         (( tita + 16 ) >> 5);
     /* Allow extra third check to avoid breaking edges 
     if ( (tita & 0x3F) == 0x2F ) vecino_tita--; */
      /* divido la circunferencia en 8 partes con ocho vecinos distintos */
      /* 0..256 -> 0..8 . Por tanto divido por 32, o equivalentemente >>5*/
      /* tomo un desfase del 25% (8 grados 0..255) para tener ms centra-*/
      /* dos los vecinos */
      int offset = *pPtoEstudiado + *vecino_tita;
      if ( mod_grad[offset] ) { /* first try */
        byte distancia = tita_grad[*pPtoEstudiado] - tita_grad[offset];
        if ( distancia < HolguraAngular || distancia > (~HolguraAngular)
             /*(256-HOLGURA_ANG)*/ ) {
          pSeg->media += mod_grad[offset]; mod_grad[offset]=(float)0.0; 
          *pFinSegmento=offset; pFinSegmento++;
        }
      } else {
        offset = *pPtoEstudiado + *(vecino_tita+1);
        if ( mod_grad[offset] ) {  /* second try */
          byte distancia = tita - tita_grad[offset];
          if ( distancia < HolguraAngular || distancia > ((byte) ~HolguraAngular)
               /*(256-HOLGURA_ANG)*/ ) {
            pSeg->media += mod_grad[offset]; mod_grad[offset]=(float)0.0; 
            *pFinSegmento=offset; pFinSegmento++;
          }
        } else /* if ( (tita & 0x3F) ^ 0x3F )*/ { 
          /* if orentation is near 45, allow third check to avoid breaking edges */
          offset = *pPtoEstudiado + *(vecino_tita-1);
          if ( mod_grad[offset] ) {  /* third try */
            byte distancia = tita - tita_grad[offset];
            if ( distancia < HolguraAngular || distancia > ((byte) ~HolguraAngular)
                 /*(256-HOLGURA_ANG)*/ ) {
              pSeg->media += mod_grad[offset]; mod_grad[offset]=(float)0.0; 
              *pFinSegmento=offset; pFinSegmento++;
            }
          }
        }
      }
      pPtoEstudiado++;
      /* Se termina el bucle cuando pFinSegmento == pPtoEstudiado; */
    } while (pPtoEstudiado<pFinSegmento);
  
   /* Incluimos el punto de inicio de la busqueda en el semisegmento de
   * sentido contrario (dos veces)
   * para simplificar el bucle (mismo comportamiento para cada sentido)
   * En realidad el punto de inicio se incluye dos veces en el segmento
   */
   /* Modificacin: utilizar dos variables auxiliares para evitar
    * la traperada. Se guarda el ltimo valor calculado en la primera  
    * pasada y la posicin dentro del segmento. Se machaca ese valor en
    * la segunda pasada pero luego se restituye */
    if (sentido) {
      pPtoEstudiado--;
      pCambioSentido = pPtoEstudiado;
      UltimoPtoPrimeraPasada = *pPtoEstudiado;
      //pPtoEstudiado++;
      *pPtoEstudiado /* pFinSegmento */ = *pSeg->offset;  
    } else {
      *pCambioSentido = UltimoPtoPrimeraPasada;
      break;
    }
  } while ( 1/*sentido*/ ); /* para cuando sentido == 1 (se repite el bucle 2 veces)*/

  /* Guarda el segmento en la estructura*/
  /* Calcula la direccin mirando la direfencia entre los punteros al 
   * principio y final del segmento */
  /* nptos -> 1 + puntos hallados recoriendo en direcciones 1 y 2  */
  int nptos = pSeg->nptos = pPtoEstudiado - pSeg->offset;

  /* Aqu se guarda a partir de seg[Imagen.nsegs+1].offset tita_grad */
   pSeg->tita_grad=(byte*) pFinSegmento;
  
   for ( ; nptos ; ) {
      nptos--;
      pSeg->tita_grad[nptos] = tita_grad[pSeg->offset[nptos]];
   }

  /* Devuelve un puntero de la primera posicin libre del almacen (a continuacion
   * del segmento actual */
  /* Primera posicin de memoria libre despus de pFinSegmento, alineada en
   * una direccin mltiplo de 8 (para acelerar accesos) */

  return ((int*)(&pSeg->tita_grad[(pSeg->nptos + 7) & (~0x000f)]));
}

/*********************************************************************************/

static void __fastcall InicializaCannyMRes(int filas, int columnas, int holg_ang,
                                           int tipo_iteracion)
{
   /* ajusta la holgura o diferencia de orientacin del gradiente
    * permitida entre dos puntos contorno vecinos
    */
   HolguraAngular = holg_ang;

   /* Calcula los vecinos, que dependen de la anchura de la
    * imagen */
   vecinoAlineado[0][0] = 1+columnas;        
   vecinoAlineado[0][1] = columnas;        
   vecinoAlineado[0][2] = columnas-1;        
   vecinoAlineado[0][3] = -1;        
   vecinoAlineado[0][4] = -1-columnas;        
   vecinoAlineado[0][5] = -columnas;        
   vecinoAlineado[0][6] = 1-columnas;        
   vecinoAlineado[0][7] = 1;        
   vecinoAlineado[0][8] = 1+columnas;        
   vecinoAlineado[0][9] = columnas;        
   vecinoAlineado[0][10] = columnas-1;        

   /* Direcciones contrarias */
   {
     int i=VECINOS_ALINEADOS;
     for (;i;) {
       i--;
       vecinoAlineado[1][i] = -vecinoAlineado[0][i];
     }
   }

   /* Inicializa vecino (utilizado en  fast_analiza_nivel) */
   int i=8*16;
   for(; i ; ) {
     i--;
     vecino[0][i] = vecino_x[0][i] + columnas*vecino_y[0][i];
   }

   /* Calcula la maxima iteracin para cada nivel en el
    * casamiento de segmentos */
   for(i=MAX_LEVEL-2; i ; ) {
     i--;
     iter_CMres[i] = iter_casamiento[tipo_iteracion][i];
   }
} // Fin de InicializaCannyMRes
                         
/*********************************************************************************/

static void SegmentaGradienteImagen  (float* mod_grad,byte* tita_grad,
                               struct struct_capa* pImagen,
		                           int columnas, int filas, float threshold,
                               float escala, int noise_len )
{ 
 float minimo = threshold*noise_len*escala;
 threshold *= escala;

 /* los segmentos se van guardando en un montn que empieza en 
    pImagen->seg[0].offset=0 */
 pImagen->nsegs = 0;

 for (int y=(filas-1)*columnas; y -= columnas; ) {
   int x=columnas-1;
   register float *punto = mod_grad + y;
   for ( ;--x; ) {
     if ( punto[x] > threshold ) {

        pImagen->seg[pImagen->nsegs+1].offset =
           SigueContorno(mod_grad,tita_grad,
                       x+y, &(pImagen->seg[pImagen->nsegs]));
        
        /* Acepta slo los segmentos de suma mayor a la 
           longitud mnima entre niveles * nivel medio de
           mnimo gradiente y longitud menor que noise_len */
        if ( pImagen->seg[pImagen->nsegs].media > minimo ) {
          /* Guarda en nptos la longitud total del segmento +1*/
          pImagen->seg[pImagen->nsegs].media /= 
            escala*(pImagen->seg[pImagen->nsegs].nptos /*- 1*/);
          pImagen->nsegs++; /* Acepta segmento */
        }
        /*
        if ( pImagen->seg[pImagen->nsegs].nptos > noise_len 
             && pImagen->seg[pImagen->nsegs].media > threshold )
          pImagen->nsegs++;
        */
	   }
   }
 }
}

/*********************************************************************************/

static void __fastcall SegmentacionAbitmapConMedia
    (struct struct_capa* pImagen, byte * pic)
{

  for ( int i=pImagen->nsegs; i ; ) {
    i--;
    int *offset = pImagen->seg[i].offset;
    byte media = (pImagen->seg[i].media > 63.4)
                 ? 255 : ((byte)(pImagen->seg[i].media*4));
    for ( int j=pImagen->seg[i].nptos; j ; ) {
      j--;
      pic[offset[j]]= /*255*/ media;
    }
  }
}

/*********************************************************************************/

static void __fastcall BorraSegmentos
    (struct struct_capa* pImagen, int* mod_grad)
{
  for ( int i=pImagen->nsegs; i ; ) {
    i--;
    int *offset = pImagen->seg[i].offset;
    for ( int j=pImagen->seg[i].nptos; j ; ) {
      j--;
      mod_grad[offset[j]]= 0;
    }
  }
}

/*********************************************************************************/

static void __fastcall SegmentacionAbitmapConGradiente
    (struct struct_capa* pImagen, byte * pic)
{

  for ( int i=pImagen->nsegs; i ; ) {
    i--;
    int *offset = pImagen->seg[i].offset;
    byte *gradiente = pImagen->seg[i].tita_grad;
    for ( int j=pImagen->seg[i].nptos; j ; ) {
      j--;
	/* register int val = gradiente[j]>>2;
         pic[offset[j]]= (val>255) ? 255 : val; */
         pic[offset[j]]= gradiente[j];
    }
  }
}

/*********************************************************************************/

static void __fastcall SegmentacionAbitmapConTabla
    (struct struct_capa* pImagen,byte* tabla, byte * pic)
{
  for ( int i=pImagen->nsegs; i ; ) {
    i--;
    int *offset = pImagen->seg[i].offset;
    for ( int j=pImagen->seg[i].nptos; j ; ) {
      j--;
      pic[offset[j]]= tabla[i] /*255*/;
    }
  }
}

/*********************************************************************************/

static byte* AsignaColoresSegs (struct struct_capa  *ImagenMultires,
                         struct casamiento_segs *Casam,
                         int maxlevel,int minlevel, int minlen,
                         int umbral)
/* This function returns NULL if an error occurs allocating memory */
/* noncero value otherwise */
{
  int nsegsNivel_0 = ImagenMultires[0].nsegs;
  
  byte *Tabla = (byte *) busca_memoria (
                  (nsegsNivel_0+1) );
  if (!Tabla) {
    return NULL; /* memory allocation failure */
  }

  pSEGMENTO *SegPadre = (pSEGMENTO*) busca_memoria ( maxlevel*
              (nsegsNivel_0+1)*(sizeof(pSEGMENTO)) );
  if (!SegPadre) {
    free(Tabla);
    return NULL; /* memory allocation failure */
  }

  int* BufferNPuntos = (int*) busca_memoria ( 
                       (nsegsNivel_0+1)*(sizeof(int)) );
  if (!BufferNPuntos) {
    free(Tabla);
    free(SegPadre);
    return NULL; /* memory allocation failure */
  }

  /* Utilizo SegPadre[0][j] para guardar el segmento superior
   * correspondiente a la escala ms alta en que exita
   * (top of parent segment). Por defecto, SegPadre[0][j]
   * contiene el segmento en el nivel cero.
   */

  /* Inicializo SegPadre, de forma que el segmento padre del
   * nivel 0 son los propios segmentos del nivel base 0.
   */
  for (int i=nsegsNivel_0 ; i; ) {
    i--;
    SegPadre[i]= &ImagenMultires[0].seg[i];
  }

 /* TablaSegmentos[i][j]. i -> rango 0..max_level-1 */
 /* j -> rango 1..ImageMultires[0].nsegs */
 pSEGMENTO TablaSegmentos[MAX_LEVEL-1];
 for (i = maxlevel ; i; ) {
    i--;
    TablaSegmentos[i] = ImagenMultires[i].seg - 1;
  }

  /* Formacin del rbol de segmentos y clculo de
   * medias */

  float factor =(float) 4.0/maxlevel; 
  for (i = nsegsNivel_0 ; i; ) {
    /* segSuperior: rango 1..ImageMultires[0].nsegs */
    u_int segSuperior = i; 
    i--;
    /* PadreNivelSup: rango 0..ImageMultires[0].nsegs - 1*/
    pSEGMENTO *PadreNivelSup = &SegPadre[i];
    /* Capa del segmento superior en el nivel ms alto
     * en que existe padre (0..maxlevel-2) */
    for ( int j=0; j<maxlevel-1 ; ) {
      segSuperior = Casam[j].segSup[segSuperior];
      if ( segSuperior ) {
        j++;
        /* almaceno el padre del siguiente nivel */
        PadreNivelSup += nsegsNivel_0; 
        *PadreNivelSup = &TablaSegmentos[j][segSuperior];
      } else {
       /* El segmento no tiene padre en el nivel j + 1
        * ni superiores. Por tanto el padre del nivel
        * superior es PadreNivelSup.
        */
        /* almaceno el padre del siguiente nivel como nulo */
        PadreNivelSup[nsegsNivel_0] = (pSEGMENTO) NULL ;
        break;
      }
    }

    /* Utilizo SegPadre[0][j] para guardar el segmento superior
     * correspondiente a la escala ms alta en que exista
     * (top of parent segment). Por defecto, SegPadre[0][j]
     * contiene el segmento en el nivel cero.
     */
    SegPadre[i] = *PadreNivelSup;

    /* Inicio del clculo de la media de los segmentos casados */
    if ( j < minlevel-1) {
     /* El segmento no tiene padre en el nivel j ni superiores */  
      Tabla[i] = 0;
    }
    else {
      register float media = ImagenMultires[0].seg[i].media;
      for ( ; j; ) {
        j--;
       /* j = 1 .. maxlevel-1 */
       /* El segmento no tiene padre en el nivel TopParentLevel
        * ni superiores */
        media += (*PadreNivelSup)->media;
        PadreNivelSup -= nsegsNivel_0;
      }
      int nivelgris = (int) (media*factor) ;
      Tabla[i]=  (nivelgris > 255 ) ? 255 : (byte) (nivelgris);
    }
    /* Fin del clculo de la media de los segmentos casados */      
  }

  /* Clculo de puntos en contorno Padre Superior */
  /* NOTA: A ESTAS ALTURAS YA NO SE NECESITA LA INFORMACIN CONTENIDA
   * EN EL CAMPO nptos DE LOS SEGMENTOS, POR LO CUAL LA PUEDO MODIFICAR.
   * (De todas formas, guardo una copia y la restituyo posteriormente).
   */

  /* Inicializacin del nmero de puntos en segmento padre*/
  for ( i = nsegsNivel_0; i ;  ) {
    i--;
    BufferNPuntos[i] = SegPadre[i]->nptos;
    SegPadre[i]->nptos = 0;
  }

  /* Clculo del nmero total de puntos en la capa base (0)
   * correspondiente al Padre Superior
   */
  for ( i=nsegsNivel_0; i ;  ) {
    i--;
    SegPadre[i]->nptos += ImagenMultires[0].seg[i].nptos;
  }

  /* Eliminacin de los puntos contorno que no tienen
   * un momento
   * min_len puntos hermanos de nivel * nivel gradiente */
  int momento = (minlen*umbral);
  for ( i=nsegsNivel_0; i ;  ) {
    i--; 
    /*
    if ( SegPadre[i]->nptos < minlen )
      Tabla[i] = 0; 
    else */   
    if ( ((int) ((int) Tabla[i])*SegPadre[i]->nptos) < momento)
        Tabla[i] = 0;
  }

  /* Por si ImagenMultires se pudiera utilizar, restituyo los
   * valores nptos de los segmentos padre contenidos 
   * en ImagenMultires a partir de BufferNPuntos */
  for ( i = nsegsNivel_0; i ;  ) {
    i--;
    SegPadre[i]->nptos = BufferNPuntos[i];
  }

  /* free allocated memory and return Tabla */
  free(SegPadre);
  free(BufferNPuntos);

  return Tabla;
}

/*********************************************************************************/

static byte* AsignaColoresSegs2 (struct struct_capa  *ImagenMultires,
                         struct casamiento_segs *Casam,
                         int maxlevel,int minlevel, int minlen)
/* This function returns NULL if an error occurs allocating memory */
/* noncero value otherwise */
{
  int nsegsNivel_0 = ImagenMultires[0].nsegs;
  
  byte *Tabla = (byte *) busca_memoria (
                  (nsegsNivel_0+1) );
  if (!Tabla) {
    return NULL; /* memory allocation failure */
  }

  pSEGMENTO *SegPadre = (pSEGMENTO*) busca_memoria ( maxlevel*
              (nsegsNivel_0+1)*(sizeof(pSEGMENTO)) );
  if (!SegPadre) {
    free(Tabla);
    return NULL; /* memory allocation failure */
  }

  int* BufferNPuntos = (int*) busca_memoria ( 
                       (nsegsNivel_0+1)*(sizeof(int)) );
  if (!BufferNPuntos) {
    free(Tabla);
    free(SegPadre);
    return NULL; /* memory allocation failure */
  }

  /* Utilizo SegPadre[0][j] para guardar el segmento superior
   * correspondiente a la escala ms alta en que exita
   * (top of parent segment). Por defecto, SegPadre[0][j]
   * contiene el segmento en el nivel cero.
   */

  /* Inicializo SegPadre, de forma que el segmento padre del
   * nivel 0 son los propios segmentos del nivel base 0.
   */
  for (int i=nsegsNivel_0 ; i; ) {
    i--;
    SegPadre[i]=&ImagenMultires[0].seg[i];
  }

 /* TablaSegmentos[i][j]. i -> rango 0..max_level-1 */
 /* j -> rango 1..ImageMultires[0].nsegs */
 pSEGMENTO TablaSegmentos[MAX_LEVEL-1];
 for (i = maxlevel ; i; ) {
    i--;
    TablaSegmentos[i] = ImagenMultires[i].seg - 1;
  }

  /* Formacin del rbol de segmentos y clculo de
   * medias */

  float factor =(float) 4.0/maxlevel; 
  for (i = nsegsNivel_0 ; i; ) {
    /* segSuperior: rango 1..ImageMultires[0].nsegs */
    u_int segSuperior = i; 
    i--;
    /* PadreNivelSup: rango 0..ImageMultires[0].nsegs - 1*/
    pSEGMENTO *PadreNivelSup = SegPadre + i ;
    /* Capa del segmento superior en el nivel ms alto
     * en que existe padre (0..maxlevel-1) */
    for ( int j=0; j<maxlevel-1 ; ) {
      segSuperior = Casam[j].segSup[segSuperior];
      if ( segSuperior ) {
        j++;
        PadreNivelSup += nsegsNivel_0;
        *PadreNivelSup = TablaSegmentos[j] + segSuperior;
      } else {
       /* El segmento no tiene padre en el nivel j + 1
        * ni superiores. Por tanto el padre del nivel
        * superior es PadreNivelSup.
        */
        PadreNivelSup[nsegsNivel_0] = (pSEGMENTO) NULL ;
        break;
      }
    }

    /* Utilizo SegPadre[0][j] para guardar el segmento superior
     * correspondiente a la escala ms alta en que exista
     * (top of parent segment). Por defecto, SegPadre[0][j]
     * contiene el segmento en el nivel cero.
     */
    SegPadre[i] = *PadreNivelSup;

    //int TopParentLevel=j;

    /* Inicio del clculo de la media de los segmentos casados */
    if ( j < minlevel-1) {
     /* El segmento no tiene padre en el nivel j ni superiores */  
      Tabla[i] = 0;
    }
    else {
      register float media = ImagenMultires[0].seg[i].media;
      for ( ; j; ) {
        j--;
       /* j = 1 .. maxlevel-1 */
       /* El segmento no tiene padre en el nivel TopParentLevel
        * ni superiores */
        media += (*PadreNivelSup)->media;
        PadreNivelSup -= nsegsNivel_0;
      }
      int nivelgris = (int) (media*factor) ;
      Tabla[i]=  (nivelgris > 255 ) ? 255 : (byte) (nivelgris);
    }
    /* Fin del clculo de la media de los segmentos casados */      
  }

  /* Clculo de puntos en contorno Padre Superior */
  /* NOTA: A ESTAS ALTURAS YA NO SE NECESITA LA INFORMACIN CONTENIDA
   * EN EL CAMPO nptos DE LOS SEGMENTOS, POR LO CUAL LA PUEDO MODIFICAR.
   * (De todas formas, guardo una copia y la restituyo posteriormente).
   */

  /* Inicializacin del nmero de puntos en segmento padre*/
  for ( i = nsegsNivel_0; i ;  ) {
    i--;
    BufferNPuntos[i] = SegPadre[i]->nptos;
    SegPadre[i]->nptos = 0;
  }

  /* Clculo del nmero total de puntos en la capa base (0)
   * correspondiente al Padre Superior
   */
  for ( i=nsegsNivel_0; i ;  ) {
    i--;
    SegPadre[i]->nptos += ImagenMultires[0].seg[i].nptos;
  }

  /* Eliminacin de los puntos contorno que no tienen
   * min_len puntos hermanos */
  for ( i=nsegsNivel_0; i ;  ) {
    i--; 
    if ( SegPadre[i]->nptos < minlen )
      Tabla[i] = 0; 
  }

  /* Por si ImagenMultires se pudiera utilizar, restituyo los
   * valores nptos de los segmentos padre contenidos 
   * en ImagenMultires a partir de BufferNPuntos */
  for ( i = nsegsNivel_0; i ;  ) {
    i--;
    SegPadre[i]->nptos = BufferNPuntos[i];
  }

  /* free allocated memory and return Tabla */
  free(SegPadre);
  free(BufferNPuntos);

  return Tabla;
}

/*********************************************************************************/
static int fast_analiza_nivel (struct struct_capa* pSuperior,
                         struct struct_capa* pInferior,
                         struct casamiento_segs* resultado, 
                         int* Segs, byte* Tita, int max_iter,
                         float reclass, int columnas, int filas,
                         int OrdenCasamiento)
/* Devuelve 0 si no puede asignar memoria temporal, distinto de
 * cero si puede puede analizar nivel */

{  
  /* Si orden de casamiento == 0 => ORDEN_ASCENDENTE
   *    Se busca un punto en el segmento superior que concuerde
   *    con el segmento inferior
   * Si orden de casamiento == 1 => ORDEN_DESCENDENTE
   *    Se busca un punto en el segmento inferior que concuerde
   *    con el segmento superior
   */
  if (!OrdenCasamiento) {   
    /* (OrdenCasamiento==0) => ORDEN ASCENDENTE*/
    /* El bucle de casamiento de  busca un punto equiva-
     * lente en el segmento pInferior que concuerde con
     * el punto del segmento pSuperior analizado
     */
    struct struct_capa* pAuxiliar = pSuperior;
    pSuperior = pInferior;
    pInferior = pAuxiliar;
  }

  /* Inicializacin de la matriz Segs con "nmero de segmento+1" y
   * "tita" con el angulo (0..256) para cada punto (se supone que
   * Segs esta rellenado con ceros, Tita no es necesario) */

  int columnas_1= columnas-1;
  int col_fil_1 = columnas*(filas-1)-1;
  int i= pInferior->nsegs;

  /* Inicializo numero de contorno y el ngulo en la imagen
   * de los contornos y de los ngulos */
  for ( ; i ; i--) {
    i--;
    int *offset = pInferior->seg[i].offset;
    byte *tita_grad = pInferior->seg[i].tita_grad;
    int j=pInferior->seg[i].nptos;
    i++; /* el numero de segmento 0 se reserva para indicar que en
          * ese punto no hay ningn segmento -> el numero de segmento
          * va de 1 a nsegs */
    for ( ; j ; ) {
      j--;
      /* Inicializo nmero de contorno */
      Segs[offset[j]]= i;
      /* Inicializo angulo del gradiente (0..255) */
      Tita[offset[j]]= tita_grad[j];
    }
  }

  /* Ubicacin de memoria temporal */
  u_int *ocurrencias = (u_int *) 
    busca_memoria ( (pInferior->nsegs + 1) 
                     * sizeof(u_int));
  if (!ocurrencias) return 1;/* memory allocation failure */

  u_int *PtosEnPadre = NULL;
  if (OrdenCasamiento) /* ORDEN_DESCENDENTE */ {
    /* Ubicacin de memoria temporal */
    PtosEnPadre = (u_int *) 
      busca_memoria ( (pInferior->nsegs + 1) 
                       * sizeof(u_int));
    if (!PtosEnPadre){
      free(ocurrencias);
      return 1;/* memory allocation failure */
    }
  }

  /* Para calcular la memoria necesaria para 
   * BufferSegmentosHijoEncontrados supongo que el caso ms
   * desfavorable (irreal) es un segmento padre cuyos hijos
   * son todos los segmentos del nivel inferior ...
   * (bueno... en este algoritmo se utiliza mucha memoria que
   * podra reducirse utilizando coeficientes sobre los reque-
   * rimientos tericos mximos, pero si el nmero de segmen-
   * tos se mantiene comedido, tampoco se solicita demasiada
   * memoria). La tabla PtosEnPadre se podra eliminar, slo
   * se utiliza para obtener una mnima optimizacin (evitar
   * la conversin de int a float y viceversa entre ocurrencias
   * y resultado->ptosComun.
   */
  int* BufferSegmentosHijoEncontrados = (int*) busca_memoria ( 
                       pInferior->nsegs*(sizeof(int)) );
  if (!BufferSegmentosHijoEncontrados) {
    free(ocurrencias);
    if (OrdenCasamiento) /* ORDEN_DESCENDENTE */ 
      free(PtosEnPadre);
    return NULL; /* memory allocation failure */
  }

  /* Inicializacion de la tabla de casamiento de segmentos */
  /* (por defecto los contornos no tienen contorno padre)  */
  if (OrdenCasamiento) {
    /* (OrdenCasamiento==1) => ORDEN_DESCENDENTE */ 
     memset((void*)resultado->segSup,0,(pInferior->nsegs+1)*sizeof(int));
     memset((void*)PtosEnPadre,0,(pInferior->nsegs+1)*sizeof(u_int));
  }
  /* Inicializacion de la tabla de ocurrencias*/
  memset((void*)ocurrencias,0,(pInferior->nsegs+1)*sizeof(u_int));

  /* ANALIZO TODOS LOS PUNTOS EN LOS SEGMENTOS SUPERIORES */
  pSEGMENTO SegmentoPadre = &pSuperior->seg[pSuperior->nsegs];
  for ( u_int numseg = pSuperior->nsegs; numseg ; numseg--) {
    
    SegmentoPadre--;
    /* A offset y tita_grad se le resta 1 en el bucle se
     * incrementan al principio del bucle.
     * rango: offset[0] hasta offset[pSuperior->nptos-1] */
    int numpto=SegmentoPadre->nptos;
    byte *tita_grad = SegmentoPadre->tita_grad + numpto;
    int *offset = SegmentoPadre->offset + numpto;
    /* Inicializo los segmentos encontrados a cero */
    int* UltimoSegmentoHijoEncontrado=BufferSegmentosHijoEncontrados;

    /* numpto contador sobre los puntos */    
    for ( ; numpto ; numpto-- ) {

      /* avanza los punteros al siguiente punto del segmento */
      offset--;
      tita_grad--;
        
      /* BUCLE QUE SE REALIZA PARA CADA PUNTO CONTORNO DE LA CAPA SUPERIOR*/
      int  despla=*offset, iteracion=max_iter;
      /* divido la circunferencia en ocho direcciones principales        */
      /* Nota: trato igual a tita que a 180+tita (direcciones opuestas)  */
      int *vecino_tita = vecino[(*tita_grad & 127)>>4];

      /* VALIDO COMO CANDIDATO A MAXIMO SI TIENE EL MODULO MAS GRANDE EN LA */
      /* DIRECCION DEL GRADIENTE Y ADEMAS CONSERVA EL ANGULO DEL PUNTO DE   */
      /* ESCALA SUPERIOR DENTRO DEL OCTANTE 45.	(en total, 8 octantes)    */
      /* 1 OCTANTE = 360/8 = 32 grados en escala (0..256)                   */

      /* El punto analizado y el original (sin el desplazamiento) deben tener
       * argumentos del gradiente similares para que formen parte del mismo 
       * gradiente
       */

      for (;;) {
        if ( Segs[despla] ) {
         /* El punto presenta un mximo local -> puede ser contorno */
         byte distancia_angular = (byte) (Tita[despla] - *tita_grad);
         if ( distancia_angular < HolguraAngular 
              || distancia_angular> ((byte) ~HolguraAngular)/*(256-HOLGURA_ANG)*/ ) {
   
           /*  EL PUNTO DE DIRECCION despla DEL SEGMENTO INFERIOR 
            *  pInferior->seg[Segs[despla]] CORRESPONDE AL SEGMENTO SUPERIOR
            *  SegmentoPadre  => AUMENTO LA OCURRENCIA DEL SEGMENTO */
           if (!ocurrencias[Segs[despla]]) {
             /* Para llevar un control de todos los segmentos en el nivel
              * inferior con puntos correspondientes en el nivel superior
              * los voy anotando en BufferSegmentosHijoEncontrados (una
              * sola vez cada segmento distinto encontrado).
              * El puntero UltimoSegmentoHijoEncontrado apunta al ltimo
              * segmento encontrado, de forma que puedo saber el nmero
              * total de segmentos hijos correspondientes al padre
              * analizado. De esta forma se acelera mucho el casamiento
              * de segmentos ya que slo se han de actualizar el estado
              * de la tabla de ocurrencias y casamientos de los segmentos
              * hijos en BufferSegmentosHijoEncontrados.
              */
             *UltimoSegmentoHijoEncontrado=Segs[despla];
             UltimoSegmentoHijoEncontrado++;
           }
           ocurrencias[Segs[despla]]++;
           /* Paso directamente a estudiar el proximo punto  */
           /* MODIFICACION: El algoritmo se comporta mejor si*/
           /* estudia todos los puntos posibles, ya que a veces*/
           /* no se encuentra el punto adecuado (sobretodo */
           /* si hay ruido cercano. Por tanto, se mira todo los */
           /* vecinos, y de esta forma tampoco afecta si se em */
           /* pieza a mirar antes un punto u otro */
           //break;     
          } /* Fin de condicin direccin de gradiente afn   */
        } /* Fin de posibilidad de que sea maximo             */

        if (!iteracion) break; /* control mximo de iteraciones */
        iteracion--;
  
        /* despla no es contorno => ANALIZAMOS OTRO PUNTO DEL ENTORNO      */
        despla = *offset + *vecino_tita;  
        /* actualizo el puntero al prximo vecino */
        vecino_tita++;

        /* MODIFICO LOS DESPLAZAMIENTOS PARA ESTUDIAR UN PUNTO CERCANO EN LA */
        /* DIRECCIN QUE MARCA EL GRADIENTE */

        // Check range y ( y != 0 || y != filas-1 )
        if (despla <= columnas || despla >= col_fil_1 )
          break;
        else {
          // Check range x ( x != 0 || x != columnas-1 )
          int aux = despla % columnas;
          if ( !aux || aux == columnas_1 ) 
            break;
        }
      } /* fin del bucle indefinido */
    } /* Fin del bucle de puntos */


    /* Asignacin del segmento padre si los puntos que tienen
     * en comn sobrepasa reclass */

    if (OrdenCasamiento) { /* (OrdenCasamiento==1) => ORDEN_DESCENDENTE */

      for (int j=UltimoSegmentoHijoEncontrado-BufferSegmentosHijoEncontrados;
           j; j--) {
        /* j = nmeo de segmentos distintos encontrados */
        /* en PtosEnPadre[num_segmento] guardo provisionalmente
         * el nmero de puntos en comn que tiene un segmento
         * con el padre
         */
        UltimoSegmentoHijoEncontrado--;
        if ( PtosEnPadre[*UltimoSegmentoHijoEncontrado] 
             < ocurrencias[*UltimoSegmentoHijoEncontrado] ) {
          // nota: *UltimoSegmentoHijoEncontrado es 
          // el ndice del segmento hijo
          PtosEnPadre[*UltimoSegmentoHijoEncontrado] = 
            ocurrencias[*UltimoSegmentoHijoEncontrado];
          /* el numero de segmento 0 se reserva para indicar que en
           * ese punto no hay ningn segmento -> el numero de segmento
           * va de 1 a nsegs */
          resultado->segSup[*UltimoSegmentoHijoEncontrado]
                       =numseg;
        }
        /* Inicializacion de la tabla de ocurrencias
         * (slo es necesario volver a inicializar los segmentos
         * hijos que han sido encontrados)*/
        ocurrencias[*UltimoSegmentoHijoEncontrado]=0;
      } /* Fin del bucle de segmentos hijo encontrados */

    } else {  /* (OrdenCasamiento==0) => ORDEN_ASCENDENTE */
      
      /* ojo: pInferior y pSuperior  que estn intercambiados
       * si OrdenCasamiento ==0 (orden ascendente)*/
      /* Puesto que en UltimoSegmentoHijoEncontrado, al contrario de lo que
       * expresa se guarda el segmento padre y slo puede haber uno, se esco-
       * ge aquel que tiene mayor nmero de puntos en comn */
      u_int SegmentoPadre = 0, PuntosEnComun = 0;
      for (int j=UltimoSegmentoHijoEncontrado-BufferSegmentosHijoEncontrados;
           j; j--) {
        /* j = nmeo de segmentos distintos encontrados */
        UltimoSegmentoHijoEncontrado--;
        if ( PuntosEnComun 
             < ocurrencias[*UltimoSegmentoHijoEncontrado] ) {
          // nota: *UltimoSegmentoHijoEncontrado es 
          // el ndice del segmento padre
          PuntosEnComun = ocurrencias[*UltimoSegmentoHijoEncontrado];
          /* el numero de segmento 0 se reserva para indicar que en
           * ese punto no hay ningn segmento -> el numero de segmento
           * va de 1 a nsegs */
          SegmentoPadre = *UltimoSegmentoHijoEncontrado;
        }
        /* Inicializacion de la tabla de ocurrencias
         * (slo es necesario volver a inicializar los segmentos
         * hijos que han sido encontrados)*/
        ocurrencias[*UltimoSegmentoHijoEncontrado]=0;
      } /* Fin del bucle de segmentos hijo encontrados */

      resultado->ptosComun[numseg] = (float) PuntosEnComun
         /(float) (MIN(pSuperior->seg[numseg-1].nptos,
                   pInferior->seg[SegmentoPadre-1].nptos));
      resultado->segSup[numseg] = SegmentoPadre;
      if (resultado->ptosComun[numseg] < reclass) {
        /* segmento padre no encontrado o no vlido */
        resultado->segSup[numseg] = 0;
        resultado->ptosComun[numseg] = (float) 0.0;

      }
    }
  } /* Fin del bucle de segmentos padre */


  if (OrdenCasamiento) { /* (OrdenCasamiento==1) => ORDEN_DESCENDENTE*/

    /* Para salir del bucle, en resultado->ptosComn[j] guardo 
     * el ratio (por unidad) de los puntos casados en 
     * el segmento superior partido por el nmero de
     * puntos del segmento inferior */
    /* analizo todos los contornos inferiores */
    for ( numseg = pInferior->nsegs; numseg ; numseg--) {
        /*resultado->ptosComun[numseg] = (float) PtosEnPadre[numseg]
           /(float) pInferior->seg[numseg-1].nptos;*/
        int nptos = pSuperior->seg[resultado->segSup[numseg]-1].nptos;;
        if (nptos > pInferior->seg[numseg-1].nptos)
          nptos = pInferior->seg[numseg-1].nptos; 
        resultado->ptosComun[numseg] = 
          (resultado->segSup[numseg]) ?
             ((float) PtosEnPadre[numseg]
              /(float) nptos)
            : 0;
        if (resultado->ptosComun[numseg] < reclass) {
          /* el segmento no ha sido casado con uno superior */
          resultado->segSup[numseg] = 0; /* segmento nulo */
          resultado->ptosComun[numseg] = (float) 0.0;
        }
      
    }
  }

  /* Libero memoria alojada y vuelvo */
  free(ocurrencias);
  if (OrdenCasamiento) 
    /* ORDEN_DESCENDENTE */ free(PtosEnPadre);
  free(BufferSegmentosHijoEncontrados);
  return 1; /* OK */

}  /* Fin del procedimiento analiza_punto */

/******************************************************************************/

static int InicializaCanyMultiNivel( struct struct_capa *CapaMultiNivel,
                              int columnas, int filas,int maxlevel)
/* Devuelve 0 si no puede asignar memoria, distinto de cero si 
 * puede inicializar CapaMultiNivel */
{
 /* Estimo la memoria necesaria para guardar todos los puntos borde de las
  * MAX_LEVEL por exceso como si entre todos los niveles existieran
  * tantos puntos como tiene la imagen.
  */
 CapaMultiNivel[0].nsegs=0;
 CapaMultiNivel[0].seg = (pSEGMENTO) busca_memoria (
                   (filas*columnas*maxlevel
                   *(sizeof(struct struct_segmento/*SEGMENTO*/)))
                   /(2*RATIO_MAX_SEGS));
 /* Nota: Tericamente, el mximo nmero de segmentos en una imagen es 
  * uno por cada 4 pixels (es decir, puntos aislados rodeados de vecinos
  * que no son contorno. Sin embargo, como el algoritmo de extraccin de
  * contornos utiliza el ngulo para extraer segmentos recorriendolos
  * en el sentido perpendicular al gradiente, el mximo nmero de 
  * segmentos prctico es un segmento por cada 2 pixels, correspondiente
  * al mximo nmero de puntos contorno.
  */
 if (!CapaMultiNivel[0].seg) {
    return 0; /* failure */
 }
 CapaMultiNivel[0].seg[0].offset = (int *) 
    busca_memoria ((filas*columnas*maxlevel
                   *(sizeof(float)+sizeof(byte)))
                   /(2*RATIO_PTOS_CONTORNO));
 if (!CapaMultiNivel[0].seg[0].offset) {
    free(CapaMultiNivel[0].seg[0].offset);
    return 0; /* failure */
 }
 /* Fin de busqueda de memoria */
 return 1; /* OK */
}

/******************************************************************************/

static int InicializaCasamientoCapa( struct casamiento_segs *CasamientoCapa,
                              struct struct_capa *CapaMultiNivel,
                              int maxlevel, int OrdenCasamiento)
/* Devuelve 0 si no puede asignar memoria, distinto de cero si 
 * puede inicializar CapaMultiNivel */
{
 /* Una vez que s el nmero de segmentos en cada capa, ubico la
  * memoria necesaria para el casamiento de segmentos
  * (nmero de casamientos [0..maxlevel-2])
  */
 /* Si orden de casamiento == 0 => ORDEN_ASCENDENTE
  *    Se busca un punto en el segmento superior que concuerde
  *    con el segmento inferior
  * Si orden de casamiento == 1 => ORDEN_DESCENDENTE
  *    Se busca un punto en el segmento inferior que concuerde
  *    con el segmento superior
  */
 for ( int nivel=0 ; nivel<maxlevel-1 ; nivel++) {
   CasamientoCapa[nivel].segSup = (u_int *) busca_memoria (
           (CapaMultiNivel[nivel/*+1-OrdenCasamiento^*/].nsegs+1)
           *(sizeof(u_int)) );
   if (!CasamientoCapa[nivel].segSup) {
      for ( ; nivel ; ) {
        --nivel;
        free(CasamientoCapa[nivel].segSup);
        free(CasamientoCapa[nivel].ptosComun);
      }
      return 0; /* failure */
   }
   CasamientoCapa[nivel].ptosComun = (float *) busca_memoria (
         (CapaMultiNivel[nivel/*+1-OrdenCasamiento*/].nsegs+1)
         *(sizeof(float)) );
   if (!CasamientoCapa[nivel].ptosComun) {
      free(CasamientoCapa[nivel].segSup);
      for ( ; nivel ; ) {
        --nivel;
        free(CasamientoCapa[nivel].segSup);
        free(CasamientoCapa[nivel].ptosComun);
      }
      return 0; /* failure */
   }
 }
 /* Fin de busqueda de memoria */
 return 1; /* OK */
}

/******************************************************************************/

static void LiberaCanyMultiNivel( struct struct_capa *CapaMultiNivel)
{
  free(CapaMultiNivel[0].seg[0].offset);
  free(CapaMultiNivel[0].seg);
}

/******************************************************************************/
static void LiberaCasamientoCapa( struct casamiento_segs *CasamientoCapa,
                           int maxlevel)
{
  for (int nivel = maxlevel-1; nivel ; ) {
    nivel--;
    free(CasamientoCapa[nivel].segSup);
    free(CasamientoCapa[nivel].ptosComun);
  }
}

/******************************************************************************/
