/******************************************************************************/

byte *ImgBasic::doFastMultires24  ( byte *pic24, int ncol, int nfil, int minlevel,
                                   int maxlevel, int umbral_grad, float reclass, int minlen,
                                   int seguimiento, int debug )
/* EN LAS BANDERAS GUARDO A!L PRINCIPIO LA CARACTERISRICA + - =, Y  */
/* LUEGO LAS COORDENADAS DEL PUNTO DETECTADO EN LA ESCALA MAS BAJA.*/
 /* MAX_LEVEL ha de ser menor o igual a 7 */
{
 int nivel;
 int *auxiliar;
 float  escala, *mod_grad;
 double factor=1.0;
 char mensaje[128];

 ProgressMeter(maxlevel+3, 0, 0, "Multirresolucin gausiana acelerada");

 /* Iniciacin dimensiones */
 columnas = ncol;
 filas = nfil;

 
 /* BUSCA LA MEMORIA NECESARIA */
 gradtita[0]=NULL;
 supresi[0]=NULL;
 auxiliar= (int*) malloc(filas*columnas*3*sizeof(int));
 if (!auxiliar) return NULL;
 mod_grad = (float*) malloc(filas*columnas*3*sizeof(float));
 if (!mod_grad) {
   free(auxiliar);
   return NULL;
 }

 for ( nivel=1 ; nivel<=maxlevel ; nivel++ )
   {
    gradtita[nivel]=busca_memoria(filas*columnas);
    if (!gradtita[nivel]) {
        for(;nivel--;) free(gradtita[nivel]);
        free(auxiliar);
        free(mod_grad);
        TRACE("Fallo requiriendo memoria en BEDEC");
        return NULL;
    }
    supresi[nivel]= busca_memoria ( (long)filas*(long)columnas );
    if (!supresi[nivel]) {
       // Fallo requiriendo memoria
       free(gradtita[nivel]);          
       for (; nivel-- ;) {
          free(supresi[nivel]);
          free(gradtita[nivel]);
       }
       free(auxiliar);
       free(mod_grad);
       return NULL;
    }
   }

 TRACE("Fin de busqueda de memoria\n");

 /* CALCULO EL GRADIENTE DE LA IMAGEN SUAVIZADA */ 
  for ( nivel=1 ; nivel<=maxlevel ; nivel++ ) {
    
    TRACE("SUAVIZADO ESCALA %d\n",nivel);
    if (nivel==1) {
      ConvByteToInt(pic24,auxiliar,columnas*filas*3);
    } else {
      FastBlur24  ( auxiliar, columnas, filas, BlurWindow[nivel]);
      /* Actualizo el valor de factor de escala de la imagen 
       * debido al filtro.
       * rango imagen original:  0..255 
       * rango imagen filtrada (contenida en auxiliar):  0..255*factor */
      factor *= FactorWindow[nivel];
    }

    FastGradient24(auxiliar,mod_grad,gradtita[nivel],columnas,filas);

    /* Para grandes ventanas, el valor mnimo del gradiente ha de ser bajo
     * -> utilizo el factor anterior de escala */
    if (debug) {
      TRACE("Supresion de puntos no maximos escala %d\n",nivel);
      escala  = MaximumFlt(mod_grad,filas*columnas*3);
      escala /= (float) 255.49;
    }

    supresion_puntos_32(mod_grad,gradtita[nivel],supresi[nivel],
		    columnas,filas,
        (float)((factor*ReduccionRuidoNivel[nivel]/EscalaGradient[nivel]*
                (double)(umbral_grad*COLOR_GRADIENT_FACTOR))),
        /*escala*/(float)(factor*COLOR_GRADIENT_FACTOR/EscalaGradient[nivel]));
    
    /*if (seguimiento > 1)
       FAST_UMBRALIZA_HISTERESIS(supresi[nivel],(byte*) mod_grad,
         (byte**)gradtita, umbral_grad, 2*umbral_grad,
         minlen,columnas,filas)  ; 
    else */

    if ( seguimiento > 1  && nivel <= MAX_LEVEL_ERASE_LONELY_POINTS )
        ERASE_LONELY_POINTS ( supresi[nivel], gradtita[nivel], columnas, filas);

    if (debug) {
      ConvIntToByte(auxiliar,(byte*)mod_grad,columnas*filas*3, (int) factor );
      sprintf(mensaje,"Imagen filtrada %d (ventana %d)",
              nivel,BlurWindow[nivel]);
      ShowImage((byte*)mod_grad,F_FULLCOLOR,mensaje);

      sprintf(mensaje,"supresion de puntos\nno maximos %d\nmaximo: %f sobre %f",
                 nivel, escala, double(COLOR_GRADIENT_FACTOR)*factor/EscalaGradient[nivel]);
      ShowImage(supresi[nivel],F_GREYSCALE,mensaje);
    }

    /* Dont excess integer 32 bits representation -> decrease factor */
    if (factor*FactorWindow[nivel+1] > MAX_FACTOR_COLOR) {
        register int *ptr=auxiliar, i=filas*columnas*3;
        int escala = (int) (log(factor)/log(2.0)) - 6;  
        /* Minimum resolution: 1/(1<<6)= 1/64 (to avoid loss precision)*/
        if (escala >1 ) { 
          factor /= (double) (1<<escala);
          for(;i;i--, ptr++)
            *ptr = *ptr>>escala; 
        }
    }
    ProgressMeter(maxlevel+3, -1, nivel, "Multirresolucin gausiana acelerada");
 } /* Fin del bucle de procesamiento de un nivel */

 TRACE("Fin de gradiente y gausianas\n");
  
 /* Slo necesito auxiliar para guardar la imagen suavizada o 
    la supresion de puntos*/
 free(auxiliar);
 free(mod_grad);

 /* Con las supresiones a cada nivel, realiza anlisis descendiente */
 return AnalisisDescendente(minlevel,maxlevel, 
          (seguimiento) ? reclass : (float) 0.0, minlen,debug);
}
 
/*********************************************************************************/
  
byte *ImgBasic::doFastMultiresGrey  ( byte *picGrey, int ncol, int nfil, int minlevel,
                                       int maxlevel, int umbral_grad, float reclass, int minlen,
                                       int seguimiento, int debug )
/* EN LAS BANDERAS GUARDO A!L PRINCIPIO LA CARACTERISRICA + - =, Y  */
/* LUEGO LAS COORDENADAS DEL PUNTO DETECTADO EN LA ESCALA MAS BAJA.*/
 /* MAX_LEVEL ha de ser menor o igual a 7 */
{
 int *auxiliar, nivel;
 float  escala,*mod_grad;
 double factor=1.0;
 char mensaje[128];

 ProgressMeter(maxlevel+3, 0, 0,  "Multirresolucin gausiana acelerada");

 /* Iniciacin dimensiones */
 columnas = ncol;
 filas = nfil;

 /* BUSCA LA MEMORIA NECESARIA */
 gradtita[0]=NULL;
 supresi[0]=NULL;
 auxiliar= (int*) malloc(filas*columnas*sizeof(int));
 if (!auxiliar) return NULL;
 mod_grad=(float*) malloc(filas*columnas*sizeof(float));
 if (!mod_grad) {
   free(auxiliar);
   return NULL;
 }

 for ( nivel=1 ; nivel<=maxlevel ; nivel++ )
   {
    gradtita[nivel]=busca_memoria(filas*columnas);
    if (!gradtita[nivel]) {
        for(;nivel--;) free(gradtita[nivel]);
        free(auxiliar);
        free(mod_grad);
        TRACE("Fallo requiriendo memoria en BEDEC");
        return NULL;
    }
    supresi[nivel]= busca_memoria ( (long)filas*(long)columnas );
    if (!supresi[nivel]) {
       // Fallo requiriendo memoria
       free(gradtita[nivel]);          
       for (; nivel-- ;) {
          free(supresi[nivel]);
          free(gradtita[nivel]);
       }
       free(auxiliar);
       free(mod_grad);
       return NULL;
    }
   }

 TRACE("Fin de busqueda de memoria\n");
 
 /* CALCULO EL GRADIENTE DE LA IMAGEN SUAVIZADA */ 
  for ( nivel=1 ; nivel<=maxlevel ; nivel++ ) {
    
    TRACE("SUAVIZADO ESCALA %d\n",nivel);
    if (nivel==1) {
      ConvByteToInt (picGrey,auxiliar,columnas*filas);
    } else {
      FastBlurGrey  ( auxiliar, columnas, filas, BlurWindow[nivel]);
      /* Actualizo el valor de factor  imagen = auxiliar ./ factor
       * rango imagen:  0..255 
       * rango auxiliar:  0..255*factor */
      factor *= FactorWindow[nivel];
    }

    FastGradientGrey(auxiliar,mod_grad,gradtita[nivel],columnas,filas);      

    /* Para grandes ventanas, el valor mnimo del gradiente ha de ser bajo
     * -> utilizo el factor anterior de escala */
    TRACE("Supresion de puntos no maximos escala %d\n",nivel);
    if (debug) {
      ConvIntToByte(auxiliar,supresi[nivel],columnas*filas,(int) factor);
      sprintf(mensaje,"Imagen filtrada %d (ventana %d)",
              nivel,BlurWindow[nivel]);
      ShowImage(supresi[nivel],F_GREYSCALE,mensaje);

      sprintf(mensaje,"Orientacin del gradiente %d",
              nivel);
      ShowImage(gradtita[nivel],F_GREYSCALE,mensaje);
       
      escala=MaximumFlt(mod_grad,filas*columnas);
      escala /= (float) 255;
      FitFltInByte(mod_grad,supresi[nivel], columnas*filas, (float)4);
      ShowImage(supresi[nivel],F_GREYSCALE,"Gradiente");
    }

    supresion_puntos_32(mod_grad,gradtita[nivel],supresi[nivel],
		    columnas,filas,
        (float)((factor*ReduccionRuidoNivel[nivel]/EscalaGradient[nivel]
        *(double)umbral_grad)),
        /*escala*/(float)(MAX(factor/EscalaGradient[nivel],1)));
 
    /* if (seguimiento > 1)
         FAST_UMBRALIZA_HISTERESIS(supresi[nivel],(byte*) mod_grad,
           (byte**)floatSupresi, umbral_grad, 2*umbral_grad,
           minlen,columnas,filas);
    else */

    if ( (seguimiento > 1)  && (nivel <= MAX_LEVEL_ERASE_LONELY_POINTS))
      ERASE_LONELY_POINTS ( supresi[nivel], gradtita[nivel], columnas, filas);
    
    if (debug) {
      sprintf(mensaje,"supresion de puntos no maximos\nnivel %d\nmaximo: %f sobre %f",
        nivel, escala, factor/EscalaGradient[nivel]);
      ShowImage(supresi[nivel],F_GREYSCALE,mensaje);
    }

    /* Dont excess integer 32 bits representation -> decrease factor */
    if ((factor*FactorWindow[nivel+1]) > MAX_FACTOR) {
        int escala = (int) (log(factor)/log(2.0)) - 7;  
        /* Minimum resolution: 1/(1<<7)= 1/128 (to avoid loss precision)*/
        if (escala >1 ) { 
          factor /= (double) (1<<escala);
          for(int *ptr=auxiliar, i=filas*columnas; i ; i--, ptr++)
            *ptr = *ptr>>escala; 
        }
     }
    ProgressMeter(maxlevel+3, -1, nivel,  "Multirresolucin gausiana acelerada");
  } /* Fin del bucle de procesamiento de un nivel */

 TRACE("Fin de gradiente y gausianas\n");  
 
 /* Slo necesito auxiliar para guardar la imagen suavizada o 
    la supresion de puntos*/
 free(auxiliar);
 free(mod_grad);

 return AnalisisDescendente(minlevel,maxlevel, 
          (seguimiento) ? reclass : (float) 0.0, minlen,debug);
}