/*********************************************************************************\
*     File:                  G A M . C P P                                        *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/* 
 * gam.cpp
 *
 * callable functions:
 *   <too many to list>
 */

#define __MAIN__
#include <stdlib.h>
#include <math.h>
#include "ImgBasic.h"

#define DEG2RAD (3.14159 / 180.0)
#define RAD2DEG (180.0 / 3.14159)




/*************************************************************************/
byte *ImgBasic::FSDither(byte *inpic, int intype, int w, int h, byte *fsgamcr, 
                           byte *rmap, byte *gmap, byte *bmap )
{
  /* takes an input pic of size w*h, and type 'intype' (PIC8 or PIC24),
   *                (if PIC8, colormap specified by rmap,gmap,bmap)
   * and does the floyd-steinberg dithering algorithm on it.
   * generates (mallocs) a w*h 1-byte-per-pixel 'outpic', using 'BLACK_COLOR'
   * and 'WHITE_COLOR' as the 'black' and 'white' pixel values, respectively
   */
  #define WHITE_COLOR 1
  #define BLACK_COLOR 0

  int    i, j, err, w1, h1;
  byte  *pp, *outpic, rgb[256];
  int   *thisline, *nextline, *thisptr, *nextptr, *tmpptr;


  outpic = (byte *) malloc((size_t) (w * h));
  if (!outpic) return outpic;
    

  if (intype == PIC8) {       /* monoify colormap */
    for (i=0; i<256; i++)
      rgb[i] = MONO(rmap[i], gmap[i], bmap[i]);
  }


  thisline = (int *) malloc(w * sizeof(int));
  nextline = (int *) malloc(w * sizeof(int));
  if (!thisline || !nextline) 
    FatalError("ran out of memory in FSDither()\n");


  w1 = w-1;  h1 = h-1;

  /* load up first line of picture */
  pp = inpic;
  if (intype == PIC24) {
    for (j=0, tmpptr=nextline; j<w; j++, pp+=3)
      *tmpptr++ = fsgamcr[MONO(pp[0], pp[1], pp[2])];
  }
  else {
    for (j=0, tmpptr=nextline; j<w; j++, pp++)
      *tmpptr++ = fsgamcr[rgb[*pp]];
  }


  for (i=0; i<h; i++) {
    if ((i&0x3f) == 0) WaitCursor();

    /* get next line of picture */
    tmpptr = thisline;  thisline = nextline;  nextline = tmpptr;  /* swap */
    if (i!=h1) {
      if (intype == PIC24) {
	pp = inpic + (i+1) * w * 3;
	for (j=0, tmpptr=nextline; j<w; j++, pp+=3)
	  *tmpptr++ = fsgamcr[MONO(pp[0], pp[1], pp[2])];
      }
      else {
	pp = inpic + (i+1) * w;
	for (j=0, tmpptr = nextline; j<w; j++, pp++)
	  *tmpptr++ = fsgamcr[rgb[*pp]];
      }
    }

    pp  = outpic + i * w;
    thisptr = thisline;  nextptr = nextline;

    if (i&1 == 0) {  /* go right */
      for (j=0; j<w; j++, pp++, thisptr++, nextptr++) {
	if (*thisptr<128) { err = *thisptr;     *pp = (byte) BLACK_COLOR; }
	             else { err = *thisptr-255; *pp = (byte) WHITE_COLOR; }
	
	if (j<w1) thisptr[1] += ((err*7)/16);
	
	if (i<h1) {
	  nextptr[0] += ((err*5)/16);
	  if (j>0)  nextptr[-1] += ((err*3)/16);
	  if (j<w1) nextptr[ 1] += (err/16);
	}
      }
    }

    else {   /* go left */
      pp += (w-1);  thisptr += (w-1);  nextptr += (w-1);
      for (j=w-1; j>=0; j--, pp--, thisptr--, nextptr--) {
	if (*thisptr<128) { err = *thisptr;     *pp = (byte) BLACK_COLOR; }
	             else { err = *thisptr-255; *pp = (byte) WHITE_COLOR; }
	
	if (j>0) thisptr[-1] += ((err*7)/16);
	
	if (i<h1) {
	  nextptr[0] += ((err*5)/16);
	  if (j>0)  nextptr[-1] += (err/16);
	  if (j<w1) nextptr[ 1] += ((err*3)/16);
	}
      }
    }
  }

  free(thisline);  free(nextline);
  return outpic;

  #undef WHITE_COLOR
  #undef BLACK_COLOR
}

/*************************************************************************/

void GenerateFSGamma(int x[],int y[],int n, byte fsgamcr[])
{
  /* this function generates the Floyd-Steinberg gamma curve (fsgamcr)

     This function generates a 'n' point spline curve to be used as a 
     non-linear grey 'colormap'.  The first and the last point are nailed 
     down at 0,0 and 255,255, and can't be changed.  You specify the other two.
     If you specify points on the line (0,0 - 255,255), you'll get the normal
     linear reponse curve.  If you specify points of 50,0 and 200,255, you'll
     get grey values of 0-50 to map to black (0), and grey values of 200-255
     to map to white (255) (roughly).  Values between 50 and 200 will cover
     the output range 0-255.  The reponse curve will be slightly 's' shaped.
     range values: x[0..n-1] and y[0..n-1], ordered with x from low to high */

  int i,j;
  //static int x[4] = {0,16,240,255};
  //static int y[4] = {0, 0,255,255};
  //double yf[4];
  double* yf = (double*)malloc(sizeof(double) * n);
  if (yf == NULL) return;

  InitSpline(x, y, 4, yf);
  
  for (i=0; i<256; i++) {
    j = (int) EvalSpline(x, y, yf, 4, (double) i);
    fsgamcr[i] = (j<0) ? 0 : (j>255) ? 255 : j;
  }

  free(yf);
}

/*********************/

byte *GammifyPic24(byte *pic24, int wide, int high, byte RedRemap[/*256*/],
                   byte GreenRemap[/*256*/], byte BlueRemap[/*256*/], 
                   int Saturation, int ValueRemap[/*256*/], int HueRemap[/*360*/] )
{
  /* applies HSV/RGB modifications to each pixel in given 24-bit image.
     creates and returns a new picture, or NULL on failure.
     Also, checks to see if the result will be the same as the input, and
     if so, also returns NULL, as a time-saving maneuver */

  byte *pp, *op;
  int   i,j;
  int   rv, gv, bv;
  byte *outpic = (byte *) NULL;
  int   min, max, del, h, s, v;
  int   f, p, q, t, vs100, vsf10000;
  int   hsvmod = 0, rgbmod = 0;

  /* check for HSV/RGB control linearity */
  
  /* check HUE remapping */
  if (HueRemap != NULL) {
    for (i=360; i ; ) {
      i--;
      if (HueRemap[i] != i) break;
    }
    if (i) hsvmod++; /* not linear hsv */
  }
 
  /* Check saturation offset */
  if (Saturation != 0) hsvmod++;

  /* check intensity remapping */
  if (ValueRemap != NULL) {
    for (i=256; i ; ) {
      i--;
      if (ValueRemap[i] != i) break;
    }
    if (i) hsvmod++; /* not linear hsv */
  }

  /* check R,G,B grafs simultaneously */
  if (RedRemap != NULL && GreenRemap != NULL && BlueRemap != NULL) {
    for (i=256; i; ) {
      i--;
      if (RedRemap[i] != (byte) i ||
	        GreenRemap[i] != (byte) i ||
	        BlueRemap[i] != (byte) i) break;
    }
    if (i) rgbmod++; /* not linear rgb */
  }
 
  if (!hsvmod && !rgbmod) return outpic;  /* apparently, it's linear */

  //WaitCursor();

  outpic = (byte *) malloc((size_t) wide * high * 3);
  if (!outpic) return outpic;

  pp = pic24;  op = outpic;
  for (i=wide * high; i; i--) {

    //if ((i&0x7fff)==0) WaitCursor();

    rv = *pp++;  gv = *pp++;  bv = *pp++;

    #define NOHUE -1

    if (hsvmod) {
      /* convert RGB to HSV */
      /* the HSV computed will be int's ranging -1..359, 0..100, 0..255 */

      max = (rv>gv) ? rv : gv;      /* compute maximum of rv,gv,bv */
      if (max<bv) max = bv;

      min = (rv<gv) ? rv : gv;      /* compute minimum of rd,gd,bd */
      if (min>bv) min=bv;

      del = max - min;
      v = max;
      if (max != 0) s = (del * 100) / max;
               else s = 0;

      h = NOHUE;
      if (s) {
	      if      (rv==max) h =       ((gv - bv) * 100) / del;
	      else if (gv==max) h = 200 + ((bv - rv) * 100) / del;
	      else if (bv==max) h = 400 + ((rv - gv) * 100) / del;


	      /* h is in range -100..500  (= -1.0 .. 5.0) */
	      if (h<0) h += 600;          /* h is in range 000..600  (0.0 .. 6.0) */
	      h = (h * 60) / 100;         /* h is in range 0..360 */
	      if (h>=360) h -= 360;
      }

      /* apply HSV mods */
      
      /* map near-black to black to avoid weird effects */
      if (v <= 16) s = 0;
      
      /* apply intGraf.func[] function to 'v' (the intensity) */
      v = ValueRemap[v];

      /* do Hue remapping */
      if (h>=0) h = HueRemap[h];
      else {  /* NOHUE */
	    }

      /* apply satDial value to s */
      s = s + Saturation;
      if (s<  0) s =   0;
      if (s>100) s = 100;


      /* convert HSV back to RGB */

      if (h==NOHUE || !s) { rv = gv = bv = v; }
      else {
	      if (h==360) h = 0;
	      
	      h        = (h*100) / 60;    /* h is in range 000..599 (0.0 - 5.99) */
	      j        = h - (h%100);     /* j = 000, 100, 200, 300, 400, 500 */
	      f        = h - j;           /* 'fractional' part of h (00..99) */
	      vs100    = (v*s)/100;
	      vsf10000 = (v*s*f)/10000;

	      p = v - vs100;
	      q = v - vsf10000;
	      t = v - vs100 + vsf10000;
	      
	      switch (j) {
	          case 000:  rv = v;  gv = t;  bv = p;  break;
	          case 100:  rv = q;  gv = v;  bv = p;  break;
	          case 200:  rv = p;  gv = v;  bv = t;  break;
	          case 300:  rv = p;  gv = q;  bv = v;  break;
	          case 400:  rv = t;  gv = p;  bv = v;  break;
	          case 500:  rv = v;  gv = p;  bv = q;  break;
	          default:   rv = gv = bv = 0;  /* never happens */
        }
      }
    }   /* if hsvmod */
    #undef NOHUE

    *op++ = RedRemap[rv];  
    *op++ = GreenRemap[gv];
    *op++ = BlueRemap[bv]; 
  }

  return outpic;
}


/*********************/
#ifndef DBL_MAX
  #define DBL_MAX 1.7976931348623158e+308
#endif

/**************************************/
double EvalSpline(int xa[],int ya[],double y2a[],int n, double x)
{
  int klo,khi,k;
  double h,b,a;

  klo = 0; /*binary search */
  khi = n-1;
  while (khi-klo > 1) {
    k = (khi+klo) >> 1;
    if (xa[k] > x) khi = k;
    else klo = k;
  }
  h = xa[khi] - xa[klo];
  
  //if (h==0.0) FatalError("bad xvalues in splint\n");
  if (h==0.0) return DBL_MAX;

  a = (xa[khi]-x)/h;
  b = (x-xa[klo])/h;
  return (a*ya[klo] + b*ya[khi] + ((a*a*a-a)*y2a[klo] +(b*b*b-b)*y2a[khi]) 
	  * (h*h) / 6.0);
}

/**************************************/
void InitSpline( int *x, int *y, int n, double *y2)
{
  /* given arrays of data points x[0..n-1] and y[0..n-1], computes the
     values of the second derivative at each of the data points
     y2[0..n-1] for use in the splint function */

  int i,k;
  double p,qn,sig,un,u[MAX_GHANDS];

  y2[0] = u[0] = 0.0;

  for (i=1; i<n-1; i++) {
    sig = ((double) x[i]-x[i-1]) / ((double) x[i+1] - x[i-1]);
    p = sig * y2[i-1] + 2.0;
    y2[i] = (sig-1.0) / p;
    u[i] = (((double) y[i+1]-y[i]) / (x[i+1]-x[i])) - 
           (((double) y[i]-y[i-1]) / (x[i]-x[i-1]));
    u[i] = (6.0 * u[i]/(x[i+1]-x[i-1]) - sig*u[i-1]) / p;
  }
  qn = un = 0.0;

  y2[n-1] = (un-qn*u[n-2]) / (qn*y2[n-2]+1.0);
  for (k=n-2; k>=0; k--)
    y2[k] = y2[k]*y2[k+1]+u[k];
}

/*************************************************************************/

void intsort(int *a, int n)
{
  /* uses the shell-sort algorithm.  for the relatively small data sets 
     we'll be using, should be quicker than qsort() because of all the
     function calling overhead associated with qsort(). */

  for (int d=n>>1;  d>0;  d=d>>1) {
    for (int i=d; i<n; i++) {
      for (int j=i-d;  j>=0 && a[j]>a[j+d];  j-=d) {
	      int t = a[j];  a[j] = a[j+d];  a[j+d] = t;
      }
    }
  }
}

/*************************************************************************/
void SortSpline(int x[], int y[], int n)
{
  /* uses the shell-sort algorithm.  for the relatively small data sets 
     we'll be using, should be quicker than qsort() because of all the
     function calling overhead associated with qsort(). */

  for (int d=n>>1;  d>0;  d=d>>1) {
    for (int i=d; i<n; i++) {
      for (int j=i-d;  j>=0 && x[j]>x[j+d];  j-=d) {
        /* swap x */
        int t = x[j];  x[j] = x[j+d];  x[j+d] = t;
        /* swap y */
        t = y[j];  y[j] = y[j+d];  y[j+d] = t; 
      }
    }
  }
}


/*************************************************************************/
    
        

