/*********************************************************************************\
*     File:                  M U L T I R E S. C P P                               *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/
/*
 *  multires.cpp: multiresolution border detection
 *
 *  Contains:
 *         void Multires();
 */

#include "ImgBasic.h"
#include "..\miguelan\miguelan.h"
#include "multires.h"
#include <math.h>


/*********************************************************************************/

static int  __fastcall fast_analiza_punto (int offset, int contador);
static int  __fastcall fast_clasifica_punto();
static void __fastcall calcula_vecinos    ();
static void __fastcall calcula_gaps       ();
static void __fastcall FillGaps (byte *solucion,  byte *supresion,
                                 byte* grad_tita, int grad_min );

/* Variables locales usadas para reducir el nmero de parametros en las llamadas
 * a las funciones
 */
static struct tp_bandera  banderas;
static byte *gradtita[MAX_LEVEL];
static byte *supresi[MAX_LEVEL];
static int columnas,filas,columnas_1,col_fil_1;

/* Use a once-calculated offset: 
 * vecino = vecino_x + vecino_Y*columnas;
 */
static int vecino[8][16];


/* Valor de sigma que se utiliza en el algoritmo GaussMultires para realizar el
 * suavizado de la imagen.
 */
const static double sigma[MAX_LEVEL+1]= {0,0.3,0.6,1.2,2.4,4,7,0};

/*********************************************************************************/
void ImgBasic::CannyMRes (int type)
{
  Multires(ALG_CANYMRES);
}

/*********************************************************************************/
void ImgBasic::Multires (int MultiresType)
{
 /* This funcition calls doRobGreadient24 if the image is coloured  *
   * or doRobGreadientGrey if the image is monocrome                 */

  byte *thepic;         /* Copy of the formatted original image */
  byte *results;         /* Buffer where store transformed image */
  static char buf1[20] = "10";         /* Default threshold */
  static char buf2[20] = "3";          /* Default min_level */
  static char buf3[20] = "0.3";        /* Default reclass   */
  static char buf4[20] = "10";         /* Default min_len   */
  int i, threshold = UMBRAL_DETECCION_MINIMO;
  int min_len=10, min_level=MIN_LEVEL;
  float reclass = (float) 0.3;
  int pfree, typePic;
  u_int NewNumCols;
  byte *rp, *gp, *bp;


  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Umbral mnimo del gradiente en los contornos:",
                  buf1, 20,"0123456789 ", 1);
  
  /* Check for right values                                      */
  if (strlen(buf1)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  threshold = atoi(buf1);

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Escala mnima para la bsqueda descendente:",
                  buf2, 20, "0123456789 ", 1);
  
  /* Check for right values                                      */
  if (strlen(buf2)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  min_level = atoi(buf2);
  if ((min_level<=1)||(min_level>=MAX_LEVEL)) min_level=MIN_LEVEL;

  /* Ask the value of the parameter to the user                 *
   * Only alow numbers and decimal point: "0123456789."         */
  i = InputDialog("Qu tanto por uno de reclasificacin quieres utilizar (0 para no usarla)",
                  buf3, 20, "0123456789.", 1);
  
  /* Check for right values                                      */
  if (strlen(buf3)==0) {
    Error("El formato del parametro no es correcto");
    return;
  }
  reclass = (float) atof(buf3);

  if (reclass > 0) {
    i = InputDialog("Longitud mnima de los contornos:",
                    buf4, 20, "0123456789 ", 1);
  
    /* Check for right values                                      */
    if (strlen(buf4)==0) {
      Error("El formato del parametro no es correcto");
      return;
    }  
    min_len = atoi(buf4);
  }

  
  /* Call a function depending of the color information */

  if ( mono ) { // MONOCHROME IMAGE

    /* Get a standard greyscale version of the real image    */
    thepic = GetImage(&typePic, F_GREYSCALE, &pfree,
                        &NewNumCols, &rp, &gp, &bp,256);
    /* Failed to get the greyscale image     */
    if (!thepic) return;  //NULL PIC

    /* Call the actual algorithm for greyscale images        */ 
    switch ( MultiresType ) {
      case ALG_MULTIRES:
          results = doGaussMultiresGrey(thepic, pWIDE,
                        pHIGH, min_level, 6 /*maxlevel*/,
                        threshold,reclass,min_len);
          break;
      case ALG_FASTMRES:
          results = doFastMultiresGrey (thepic, pWIDE, pHIGH,
                                        min_level, 6 /*maxlevel*/, threshold,
                                        reclass, min_len);
          break;
      case ALG_CANYMRES:
          results = doCannyMultiresGrey(thepic, pWIDE, pHIGH,
                                        min_level, 6 /*maxlevel*/, threshold,
                                        threshold/RATIO_UMBRAL_MAX_MIN,
                                        reclass,min_len);
          break;
    }
    
  } else {  // COLOR IMAGE

    /* Get a standard RGB color version of the real image    */
    thepic = GetImage(&typePic, F_FULLCOLOR, &pfree,
                        &NewNumCols, &rp, &gp, &bp,256);
    /* Failed to get the greyscale image     */
    if (!thepic) return;  //NULL PIC

    /* Call the actual algorithm for color images            */ 
    switch ( MultiresType ) {
      case ALG_MULTIRES:
          results = doGaussMultires24 (thepic, pWIDE, pHIGH,
                                       min_level,6 /*maxlevel*/,
                                       threshold,reclass,min_len);
          break;
      case ALG_FASTMRES:
          results = doFastMultires24  (thepic, pWIDE, pHIGH,
                                       min_level, 6 /*maxlevel*/, threshold,
                                       reclass,min_len);
          break;
      case ALG_CANYMRES:
          results = doCannyMultires24 (thepic, pWIDE, pHIGH,
                                        min_level, 6 /*maxlevel*/, threshold,
                                        threshold/RATIO_UMBRAL_MAX_MIN,
                                        reclass,min_len);
          break;
    }
  }

  /* free thepic if needed */
  if (pfree) free(thepic);

  if (!results) return;

  /* Store new image, allow 'undo', uptade display, etc.   */
  StoreNewPic(results,PIC8,TRUE);
  
  if ( MultiresType != ALG_CANYMRES ) {
     /* image is not really grayscaled */
     mono=FALSE;
     // Puntos gap o de rotura de contorno: naranja
     gMap[GAP]=bMap[GAP]=32;
     rMap[GAP]=128;
     // Puntos indeterminados: gris oscuro
     rMap[INDETERMINADO]=
       gMap[INDETERMINADO]=
         bMap[INDETERMINADO]=100;
     // Puntos escalera: azul claro
     rMap[ESCALERA]=gMap[ESCALERA]=128;
     bMap[ESCALERA]=255;
     // Puntos rampa: verde claro
     rMap[RAMPA]=bMap[RAMPA]=128;
     gMap[RAMPA]=255;
     // Puntos pulso: rojo intenso
     gMap[PULSO]=bMap[PULSO]=0;
     rMap[PULSO]=255;
     // Puntos escalon: color blanco (por defecto)
     gMap[PULSO]=bMap[PULSO]=
       rMap[PULSO]=255;
  }

  RedrawPic(); // Display new image
}

/*********************************************************************************/
byte *ImgBasic::AnalisisDescendente(int minlevel, int maxlevel, float reclass, int minlen, int debug)
{
 /* Alojar memoria para guardar la solucin */                           
 int contador;
 byte *solucion = (byte *) calloc(filas*columnas, 1);
 if (!solucion) {
    for ( contador=1 ; contador<= maxlevel; contador++) {
      free(supresi[contador]);
      free(gradtita[contador]);
    }
    return solucion;
 }
 
 TRACE("Fin de alojamiento de la solucin\n");
 
 /* ANLISIS DESCENDENTE, PARTIENDO DE LA ESCALA contador HASTA LA CORRESPONDIENTE
  * A LA ESCALA SIN SUAVIZAR. (PARA EVITAR RUIDO, SE RECONOCE UN PUNTO DE LA 
  * IMAGEN EN UNA ESCALA INFERIOR A minlevel TIENE QUE CORRESPONDERLE AL MENOS UN
  * PUNTO EN LAS ESCALAS SUPERIORES A minlevel -MAS SUAVIZADAS Y POR TANTO CON
  * MENOS RUIDO */
 calcula_vecinos();

 byte *picLevel;
 int y;
 char mensaje[128];

 for (contador=maxlevel; contador >= minlevel ; contador-- ) {
    picLevel = supresi[contador];
    for ( y=filas-1 ; --y ; ) {
      register int x=columnas-1;
      register byte *thepic = picLevel + y*columnas + x;
      for (; --x ; ) {
        thepic--;
        if ( *thepic ) { //supresi[contador][x+y*columnas]         
          
          banderas.c[0]=banderas.c[1]=banderas.c[2]=banderas.c[3]=
                      banderas.c[4]=INDET;
          /* clasifica punto y deja en banderas.x               *
           * la direccin inicial del punto                     */
          /* almacena en su lugar, el tipo de punto encontrado  */
          if (fast_analiza_punto  (thepic-picLevel,contador))
            solucion[ banderas.x ] = fast_clasifica_punto();     
        } // fin del if (*thepic)
      }
    }
    if  ( debug ) {
      sprintf(mensaje,"Proceso busqueda e identif. escala %d\n",contador);
      ShowImage(solucion,F_GREYSCALE,mensaje);
    }
    TRACE("Proceso busqueda e identif. escala %d\n",contador);    
    ProgressMeter(maxlevel+3, -1, maxlevel+1, "Multirresolucin gausiana");
  }

  TRACE("Fin de la busqueda descendente\n");

  /* ACABADO EL PROCESO LIBERO LA MEMORIA DE LOS PUNTEROS USADOS */
  for ( contador=1 ; contador<=maxlevel ; contador++) {
    free(supresi[contador]);
    free(gradtita[contador]);
  }

  ProgressMeter(maxlevel+3, -1, maxlevel+2, "Multirresolucin gausiana");
  
  if (reclass > 0.0) 
    /* RECALIFICACIN DE TIPOS  */
    RECONVERSION_TIPOS_TOTAL(solucion,reclass,
                             minlen,columnas,filas);

  ProgressMeter(maxlevel+3, -1, maxlevel+3, "Multirresolucin gausiana");
  
  TRACE ("Fin de liberar memorias y vuelvo");
  return solucion;
}

/******************************************************************************/

static void __fastcall calcula_vecinos()
{
  columnas_1= columnas-1;
  col_fil_1 = columnas*(filas-1)-1;
  int i=8*16;
  for(; i ; ) {
    i--;
    vecino[0][i] = vecino_x[0][i] + columnas*vecino_y[0][i];
  }
}

/******************************************************************************/

static int  __fastcall fast_analiza_punto  (int offset, int contador)
{
  int  despla=offset, iteracion=max_iteracion[contador-2];
  /* divido la circunferencia en ocho direcciones principales        */
  /* Nota: trato igual a tita que a 180+tita (direcciones opuestas)  */
  int *vecino_tita = vecino[(gradtita[contador][offset] & 127)>>4];

  /* Go to next level */
  contador--;

  while( 1 ) { // this fuction exits with return  

  /* VALIDO COMO CANDIDATO A MAXIMO SI TIENE EL MODULO MAS GRANDE EN LA */
  /* DIRECCION DEL GRADIENTE Y ADEMAS CONSERVA EL ANGULO DEL PUNTO DE   */
  /* ESCALA SUPERIOR DENTRO DEL OCTANTE 45.	(en total, 8 octantes)    */
  /* 1 OCTANTE = 360/8 = 32 grados en escala (0..256)                   */

  /* El punto analizado y el original (sin el desplazamiento) deben tener
   * argumentos del gradiente similares para que formen parte del mismo gradiente
   */
    
  if ( supresi[contador][despla] ) {
   /* El punto presenta un mximo local -> puede ser contorno */
   byte distancia_angular = (byte) (gradtita[contador][despla]-
                                    gradtita[contador+1][offset]);
   if ( distancia_angular < HOLGURA_ANG  
        || distancia_angular>(256-HOLGURA_ANG) ) {
    // distancia_octantes = distancia MOD 8 == distancia & 7
    
     if (contador==1) {
        /* SI LA ESCALA ES LA 1 GUARDO LAS COORDENADAS DEL PUNTO A MARCAR     */
        banderas.x = despla;
     } else { // contador > 1
       /* PUNTO EN LA ESCALA contador ES MAXIMO. ANALISIS A ESCALA contador-1 */
       if ( !fast_analiza_punto(despla,contador) )
         break;
     }   
     
     /* PUNTO EN LA ESCALA contador ES MAXIMO ==> ES UN CONTORNO         */
     float cociente = (float) supresi[contador][despla] /
                      (float) supresi[contador+1][offset] ;
     
     banderas.c[contador-1] = (cociente > (1.0 + MIN_PERCENTAGE)) ? MAS :
                              (cociente < (1.0 - MIN_PERCENTAGE)) ? MENOS : IGUAL ;     

    /* borramos el punto actual como no contorno para no volverlo a estudiar
       y devolvemos falso (OJO, no borro el punto en contador+1 porque necesito
       ese valor para calcular banderas.c[contador] en escala superior). */
     supresi[contador][despla]=0;

     return TRUE;     
    } /* Fin de condicin direccin de gradiente afn */
  } /* Fin de posibilidad de que sea maximo                          */

  /* PUNTO NO MAXIMO => ANALIZAMOS OTRO PUNTO DEL ENTORNO              */
  if ( iteracion ) {
    despla = offset + *vecino_tita;
    vecino_tita++;
    iteracion--;
    
    /* MODIFICO LOS DESPLAZAMIENTOS PARA ESTUDIAR UN PUNTO CERCANO EN LA */
    /* DIRECCIN QUE MARCA EL GRADIENTE */
    
    // Check range y ( y != 0 || y != filas-1 )
    if (despla <= columnas || despla >= col_fil_1 )
      break;
    else {
      // Check range x ( x != 0 || x != columnas-1 )
      int aux = despla % columnas;
      if ( !aux || aux == columnas_1 ) 
        break;
    }    
  } 
  /* NO SE HA ENCONTRADO MAXIMO EN PUNTOS DEL ENTORNO => NO ES CONTORNO */
  /* borramos el punto actual como no contorno para no volverlo a estudiar
   * y devolvemos falso*/
  else break; 

  }/* Fin del bucle indefinido */  

  /* HEMOS SALIDO DEL BUCLE CON BREAK.
   * Borramos el punto SUPERIOR como no contorno para no volverlo a estudiar
   * y devolvemos falso */
  supresi[contador+1][offset]=0;
  return (FALSE);

}  /* Fin del procedimiento analiza_punto */

/******************************************************************************/

static int __fastcall fast_clasifica_punto()
{
    /* COMPRUEBO TIPO DE CONTORNO ESCALON */
    if ( banderas.c[0] & IGUAL  
        && banderas.c[1] & (IGUAL | INDET) 
          && banderas.c[2] & (IGUAL | INDET) 
            && banderas.c[3] & (IGUAL | INDET)
              && banderas.c[4] & (IGUAL | INDET) )
      return RAMPA;
  
    /* COMPRUEBO TIPO DE CONTORNO RAMPA */
    if ( banderas.c[0] & MENOS 
         &&	banderas.c[1] & (MENOS | IGUAL)
           && banderas.c[2] & (MENOS | IGUAL | INDET) 
	           && banderas.c[3] & (MENOS | IGUAL | INDET)
                && banderas.c[4] & (MENOS | IGUAL | INDET) )
               /* && ( (contador>3) || ( (contador<=3) && 
                  (fabs(( actual-superior)/superior)>MIN_PERCENTAGE )))*/   
      return ESCALON;
  
    /* COMPRUEBO TIPO DE CONTORNO ESCALERA */
    if ( banderas.c[0] & IGUAL
        && banderas.c[1] & (IGUAL | MENOS | INDET )
          && banderas.c[2] & (MENOS | INDET)
            && banderas.c[3] & (MENOS | IGUAL | INDET)
              && banderas.c[4] & (MENOS | IGUAL | INDET) )
              /* && ( (contador>3) || ( (contador<=3) && 
                  (fabs(( actual-superior)/superior)>MIN_PERCENTAGE ) ) )*/
      return ESCALERA;
  
    /* COMPRUEBO TIPO DE CONTORNO PULSO */
    if ( banderas.c[0] & (IGUAL | MAS) 
         &&	banderas.c[1] & (IGUAL | MAS /*| INDET*/)
           && banderas.c[2] & (IGUAL | MAS /*| INDET*/)
            && banderas.c[3] & (MAS | INDET) 
              && banderas.c[4] & (MAS | INDET) )
      return PULSO;
    
    /* TIPO DESCONOCIDO */
    return INDETERMINADO;
}

/******************************************************************************/

static void __fastcall FillGaps (byte *solucion,  byte *supresion, byte* grad_tita, int grad_min )
{
  int x,y,despla, iteracion, offset, max_offset = columnas*filas-1;
  int npuntos,gap[MAX_GAP+1];

  for ( y=filas-2 ; y ; y-- ) {
    x=columnas-1;
    offset = x + y*columnas;
    for (  ; --x ; ) {
      offset--;
      // a seed in solution is needed to start finding gaps
      if ( !solucion[offset] ) continue;
      gap[0]=despla=offset;
      supresion [offset]=npuntos=iteracion=0;
      while (iteracion < MAX_GAP) {
        /* divido la circunferencia en ocho direcciones principales        */
        /* Nota: trato igual a tita que a 180+tita (direcciones opuestas)  */
        /* Al sumar 64 a gradtita, calculo el ngulo perpendicular ( == sumar 90)*/
        /* divido 0..180 en ocho partes */
        int aux = ( grad_tita[despla] & 127)>>4;  
        despla  = offset + vecino[aux][iteracion];
        iteracion++;
        
        /* MODIFICO LOS DESPLAZAMIENTOS PARA ESTUDIAR UN PUNTO CERCANO EN LA */
        /* DIRECCIN QUE MARCA EL GRADIENTE */
  
        // Check x range
        if (despla < 0 || despla > max_offset )
          /* borramos el punto actual como no contorno para no volverlo a estudiar
             y devolvemos falso */      
          break;
  
        // Check  y range
        aux = despla % columnas;
        if ( !aux || aux == columnas_1 ) 
          /* borramos el punto actual como no contorno para no volverlo a estudiar
             y devolvemos falso */      
          break;
        
        /* El punto siguiente es un punto marcado como solucin ->
             nueva iteracin */
        if ( solucion[despla] ) continue;

        /* El punto bajo estudio es un gap */
        if (supresion[despla]>grad_min)
        {
          npuntos++;
          gap[npuntos] = despla;
        }        
      } /* Fin del bucle iterativo */
      
      if (solucion[despla])
        for (int i=npuntos+1;i;) {
          i--;
          despla = gap[i];
          supresion[despla] = 0;
          solucion[despla] = GAP;          
        }
    }
  }
}  /* Fin del procedimiento FillGaps */

/******************************************************************************/
static void __fastcall calcula_gaps ()
{
  columnas_1= columnas-1;
  col_fil_1 = columnas*(filas-1)-1;
  int i=8*4;
  for(; i ; ) {
    i--;
    vecino[0][i] = vecino_x[0][i] + columnas*vecino_y[0][i];
  }
}


/******************************************************************************/

#include "GSMRes.cpp"

/******************************************************************************/

#include "FastMRes.cpp"

/******************************************************************************/

#include "CanyMRes.cpp"

/******************************************************************************/
