
#ifndef TRACE
#define TRACE //
/*#define TRACE( )*/
#endif

#define INDET 0X1
#define MAS   0x2
#define MENOS 0X4
#define IGUAL 0X8

/* limit factor to avoid overpass integer precision (2^24)*/
#define MAX_FACTOR 16777216
/* for color images, MAX_FACTOR must be tree times smaller
 * at least (or 96 times smaller if FastGradient24 computes
 * gradient using shifts).
 */
#define COLOR_GRADIENT_FACTOR 3
#define MAX_FACTOR_COLOR MAX_FACTOR/COLOR_GRADIENT_FACTOR
/* limit max gap filling in function FillGaps */
#define MAX_GAP 5

/* In order to speed up fast multires, dont look lonely point at higher
 * scales since the image is very smooth and it's very strange to find
 * lonely points. */
#define MAX_LEVEL_ERASE_LONELY_POINTS 0

/* Factor para ahorrar memoria para algoritmo CannyMRes
 * Relacin entre nmero mximo de puntos contorno en un nivel
 * y el nmero mximo de puntos contorno terico
 * (valores normales: 2-8). Valor seguro 1 */
#define RATIO_PTOS_CONTORNO 1

/* Factor para ahorrar memoria para algoritmo CannyMRes
 * Relacin entre nmero mximo de contornos en un nivel
 * y el nmero mximo de contorno terico
 * contorno = agrupacin de pixels contiguos en la direccin
 * perpendicular al gradiente = "segmentos curvos"
 * (valores normales: 2-8). Valor seguro 1 */
#define RATIO_MAX_SEGS 1

/* NOTA: El mximo nmero de puntos contorno y contornos
 * que hay en una imagen depende mucho del ruido de la
 * imagen y sobretodo del umbral del mdulo mximo del
 * gradiente para incluir el contorno el la lista 
 * de contornos. Si este nivel baja por debajo de 5 
 * (referido a un mximo de 255), es muy probable que
 * haya algn error de proteccin debido al sobrepasamiento
 * de la memoria asignada por defecto (no se controla
 * el sobrepasamiento de la memoria asignada). Adems,
 * valores iguales o mayores a un umbral del contorno 10
 * sobre 255, suelen ser ms que suficientes si la imagen
 * est suficientemente contrastada.
 */

#define RATIO_UMBRAL_MAX_MIN 4
/* Es el valor por defecto del nivel medio del ruido. Este valor se utiliza si
 * no se indica otra cosa en CannyMRes para desechar como contornos los puntos
 * con mdulos del gradiente menores que este valor
 */

/* Estructura vecino: 1er indice: nmero de direccin (entre 8 posibles)
 * 2  indice: iteracin i (n de intento de encontrar un contorno.
 * valor devuelto: desplazamiento (en x o y) respecto del punto de anlisis
 * del candidato i a ser contorno.
 * Nota: 2 indice aumentado a una potencia de 2 para aumentar velocidad
 * acceso al elemento.
 */
#ifdef _TABLA_VECINOS_INICIAL
const static int vecino_x[8][16] =
                             {{ 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 3,-3, 3,-3,0,0,0,0},
                              { 0, 0, 1,-1, 2,-2, 2,-2, 3,-3, 3,-3,0,0,0,0},
                              { 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1,0,0,0,0},
                              { 0, 0, 0, 0, 0, 0, 0, 0,-1, 1,-1, 1,0,0,0,0},
                              { 0, 0,-1, 1,-1, 1,-2, 2,-3, 3,-3, 3,0,0,0,0},
                              {-1, 1,-1, 1,-2, 2,-2, 2,-3, 3,-3, 3,0,0,0,0},
                              {-1, 1,-2, 2,-3, 3,-4, 4,-5, 5,-6, 6,0,0,0,0}};

const static int vecino_y[8][16] = 
                             {{ 0, 0, 0, 0, 0, 0, 0, 0, 1,-1, 1,-1,0,0,0,0},
                              { 0, 0, 1,-1, 1,-1, 2,-2, 2,-2, 3,-3,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 3,-3, 4,-4, 5,-5,0,0,0,0},
                              { 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6,0,0,0,0},
                              { 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 2,-2, 3,-3,0,0,0,0},
                              { 1,-1, 0, 0, 1,-1, 2,-2, 2,-2, 3,-3,0,0,0,0},
                              { 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1,0,0,0,0}};

#else // _TABLA_VECINOS_INICIAL
// Tabla terica calculada a partir de los senos y cosenos                   
const static int vecino_x[8][16] =  
                             {{ 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6,0,0,0,0},
                              { 1,-1, 2,-2, 2,-2, 3,-3, 4,-4, 5,-5,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 3,-3, 3,-3,0,0,0,0},
                              { 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1,0,0,0,0},
                              { 0, 0, 0, 0, 0, 0,-1, 1,-1, 1,-1, 1,0,0,0,0},
                              {-1, 1,-1, 1,-2, 2,-2, 2,-3, 3,-3, 3,0,0,0,0},
                              {-1, 1,-2, 2,-2, 2,-3, 3,-4, 4,-5, 5,0,0,0,0},
                              {-1, 1,-2, 2,-3, 3,-4, 4,-5, 5,-6, 6,0,0,0,0}};

const static int vecino_y[8][16] =  
                             {{ 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 3,-3, 3,-3,0,0,0,0},
                              { 1,-1, 2,-2, 2,-2, 3,-3, 4,-4, 5,-5,0,0,0,0},
                              { 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6,0,0,0,0},
                              { 1,-1, 2,-2, 3,-3, 4,-4, 5,-5, 6,-6,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 2,-2, 3,-3,0,0,0,0},
                              { 1,-1, 1,-1, 2,-2, 2,-2, 3,-3, 3,-3,0,0,0,0},
                              { 0, 0, 0, 0, 0, 0, 1,-1, 1,-1, 1,-1,0,0,0,0}};
#endif // _TABLA_VECINOS_INICIAL                   

/* Dimensin de las ventanas de suavizado utilizadas en los algoritmos 
 * FastMultires y CannyMultitres
 */

const static int BlurWindow [MAX_LEVEL] = {0,1,2,3,5,9,17};
const static int FactorWindow[MAX_LEVEL+1]={0,1,16,9,25,81,289,0};

/*FactorGradient relacin entre el gradiente mximo y el factor de
 * escala de la imagen 
 * (factor gradiente = factor imagen / EscalaGradient) */
const double EscalaGradient[MAX_LEVEL+1]=
             {0.0,0.666666,1.33333,2,3.33333,6,11.33333,0};

/* ReduccionRuidoNivel es un factor de correccin del nivel de 
 * ruido para cada escala de la imagen. En la escala base, el
 * nivel del ruido en el gradiente es el introducido por el
 * usuario, pero conforme aumenta el suavizado, el ruido
 * es menor y este factor se tiene en cuenta con
 * ReduccinRuidoNivel
 */
const float ReduccionRuidoNivel[MAX_LEVEL+1] = 
              {0.0,1.0,0.8,0.7,0.5,0.4,0.3,0.0};

/* Maximo nivel de iteracin buscando un contorno cercano en la escala inferior */
#ifdef LONG_MATCHES
const static int max_iteracion[MAX_LEVEL] = {4 /*Nivel 1-2*/,6/*Nivel 2-3*/,
                                 8/*Nivel 3-4*/,10/*Nivel 4-5*/,12/*Nivel 5-6*/ };
#else
const static int max_iteracion[MAX_LEVEL] = {2 /*Nivel 1-2*/,2/*Nivel 2-3*/,
                       4/*Nivel 3-4*/,6/*Nivel 4-5*/,10/*Nivel 5-6*/ };
#endif


/* En el algoritmo Canny Multirresolucin, se puede utilizar la opcin de bajar
 * an ms la holgura en el casamiento, o lo que es lo mismo, el nmero
 * de iteraciones para su bsqueda 
 */

const static int max_iteracion2[MAX_LEVEL] = {0 /*Nivel 1-2*/,2/*Nivel 2-3*/,
                       2/*Nivel 3-4*/,4/*Nivel 4-5*/,6/*Nivel 5-6*/ };
