/*
 *  common.h  -  definitions used in IOfile class
 * 
 */

#ifndef _COMMON_H_
#define _COMMON_H_

#ifndef __cplusplus
#define __cplusplus
#endif

/* The BSD typedefs are used throughout.
 * If your system doesn't have them in <sys/types.h>,
 * then define BSDTYPES in your Makefile.
 */
#ifndef u_char
    typedef unsigned char  u_char;
#endif
#ifndef u_short
    typedef unsigned short u_short;
#endif
#ifndef u_int
    typedef unsigned int   u_int;
#endif
#ifndef u_long
    typedef unsigned long  u_long;
#endif
#ifndef INT32
    typedef long INT32;
#endif

#define DEBUG_LEVEL 0

#ifndef MAXPATHLEN
#  define MAXPATHLEN 127
#endif


#ifndef boolean
typedef unsigned char boolean;
#endif


#ifndef FALSE      /* in case these macros already exist */
    #define FALSE  0    /* values of boolean */
#endif

#ifndef TRUE
    #define TRUE  1
#endif

#ifndef M_PI
#  define M_PI (3.1415926535897932385)
#endif


#define ERRSTR(x) sys_errlist[x]

#define random rand
#define srandom srand
#define index  strchr
#define rindex strrchr
#define debug_printf printf
#define  VERSTR "v 0.09 Nov-1996"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* lots of things don't have <malloc.h> */
/* A/UX systems include it from stdlib, from Xos.h */
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "X.h"
#include "Xmd.h"


#define USE_GETCWD          /* use 'getcwd()' instead of 'getwd()' */
#define HAVE_PDS
#define HAVE_TIFF
#define HAVE_JPEG



/*****************************/
/* END OF CONFIGURATION INFO */
/*****************************/

/* strings in the INFOBOX (used in SetISTR and GetISTR) */
#define NISTR         9    /* number of ISTRs */
#define ISTR_INFO     0
#define ISTR_WARNING  1
#define ISTR_FILENAME 2
#define ISTR_FORMAT   3
#define ISTR_RES      4
#define ISTR_CROP     5
#define ISTR_EXPAND   6
#define ISTR_SELECT   7
#define ISTR_COLOR    8

/* potential values of 'infomode', used in info box drawing routines */
#define INF_NONE 0    /* empty box */
#define INF_STR  1    /* just ISTR_INFO */
#define INF_PART 2    /* filename, format, size and infostr */
#define INF_FULL 3    /* INF_PART + clipping, expansion, colorinfo */


/* values 'epicMode' can take */
#define EM_RAW    0
#define EM_DITH   1
#define EM_SMOOTH 2



/* possible values of 'rootMode' */
#define RM_NORMAL  0     /* default X tiling */
#define RM_TILE    1     /* integer tiling */
#define RM_MIRROR  2     /* mirror tiling */
#define RM_IMIRROR 3     /* integer mirror tiling */
#define RM_CENTER  4     /* modes >= RM_CENTER centered on some sort of bg */
#define RM_CENTILE 4     /* centered and tiled.  NOTE: equals RM_CENTER */
#define RM_CSOLID  5     /* centered on a solid bg */
#define RM_CWARP   6     /* centered on a 'warp-effect' bg */
#define RM_CBRICK  7     /* centered on a 'brick' bg */
#define RM_ECENTER 8     /* symmetrical tiled */
#define RM_ECMIRR  9     /* symmetrical mirror tiled */
#define RM_MAX     RM_ECMIRR


/* values of colorMapMode */
#define CM_NORMAL    0        /* normal RO or RW color allocation */
#define CM_PERFECT   1        /* install own cmap if necessary */
#define CM_OWNCMAP   2        /* install own cmap always */
#define CM_STDCMAP   3        /* use stdcmap */


/* indicies into conv24MB */
#define CONV24_8BIT  0
#define CONV24_24BIT 1
#define CONV24_SEP1  2
#define CONV24_LOCK  3
#define CONV24_SEP2  4
#define CONV24_FAST  5
#define CONV24_SLOW  6
#define CONV24_BEST  7
#define CONV24_MAX   8

/* values 'picType' can take */
#define PIC8  CONV24_8BIT
#define PIC24 CONV24_24BIT

/* constants for setting default 'save mode' in dirW */
#define F_COLORS    0
#define F_FORMAT    1

/* the following list give indicies into saveColors[] array in xvdir.c */
#define F_FULLCOLOR 0
#define F_GREYSCALE 1
#define F_BWDITHER  2
#define F_REDUCED   3
#define F_MAXCOLORS 4   /* length of saveColors[] array */

/* things EventLoop() can return (0 and above reserved for 'goto pic#') */
#define QUIT      -1   /* exit immediately  */
#define NEXTPIC   -2   /* goto next picture */
#define PREVPIC   -3   /* goto prev picture */
#define NEXTQUIT  -4   /* goto next picture, quit if none (used by 'wait') */
#define LOADPIC   -5   /* load 'named' pic (from directory box) */
#define NEXTLOOP  -6   /* load next pic, loop if we're at end */
#define DFLTPIC   -7   /* load the default image */
//#define DELETE    -8   /* just deleted pic.  load 'right' thing */
#define GRABBED   -9   /* just grabbed a pic.  'load' it up */
#define POLLED    -10  /* polling, and image file has changed... */
#define RELOAD    -11  /* 'reload' interrupt came. be happier about errors */
#define THISNEXT  -12  /* load 'current' selection, Next until success */
#define OP_PAGEUP -13  /* load previous page of multi-page document */
#define OP_PAGEDN -14  /* load next page of multi-page document */
#define PADDED    -15  /* just grabbed a pic.  'load' it up */

/* Codes for savePic -rewrite the file/save with new name- */
#define SAVEPIC     0
#define REWRITEPIC -1

/* following list gives indicies into 'saveFormats[]' array in xvdir.c
   note that JPEG and TIFF entries may or may not exist, and following
   constants have to be adjusted accordingly.  Also, don't worry about 
   duplicate cases if JPGINC or TIFINC = 0.  All code that references
   F_JPEG or F_TIFF is #ifdef'd, so it won't be a problem */


/* Black and White colors for 24 bit display modes */
#define whtRGB 0xFFFFFFL
#define blkRGB 0L

/* Black and White colors for 1 bit display modes */
#define white 1
#define black 0


#ifdef HAVE_JPEG
#define F_JPGINC  1
#else
#define F_JPGINC  0
#endif

#ifdef HAVE_TIFF
#define F_TIFINC  1
#else
#define F_TIFINC  0
#endif


#define F_GIF         0
#define F_JPEG      ( 0 + F_JPGINC)
#define F_TIFF      ( 0 + F_JPGINC + F_TIFINC)
#define F_PS        ( 1 + F_JPGINC + F_TIFINC)
#define F_PBMRAW    ( 2 + F_JPGINC + F_TIFINC)
#define F_PBMASCII  ( 3 + F_JPGINC + F_TIFINC)
#define F_XBM       ( 4 + F_JPGINC + F_TIFINC)
#define F_XPM       ( 5 + F_JPGINC + F_TIFINC)
#define F_BMP       ( 6 + F_JPGINC + F_TIFINC)
#define F_SUNRAS    ( 7 + F_JPGINC + F_TIFINC)
#define F_IRIS      ( 8 + F_JPGINC + F_TIFINC)
#define F_TARGA     ( 9 + F_JPGINC + F_TIFINC)
#define F_FITS      (10 + F_JPGINC + F_TIFINC)
#define F_PM        (11 + F_JPGINC + F_TIFINC)
#define F_DELIM1    (12 + F_JPGINC + F_TIFINC)     /* ----- */
#define F_FILELIST  (13 + F_JPGINC + F_TIFINC)
#define F_RAW       (14 + F_JPGINC + F_TIFINC)     /* 16, normally */
#define F_MAXFMTS   (15 + F_JPGINC + F_TIFINC)     

/* return values from ReadFileType()
 * positive values are *definitely* readable formats (HAVE_*** is defined)
 * negative values are random files that XV can't read, but display as
 *   different icons in the visual browser 
 */
#define RFT_ERROR    -1    /* couldn't open file, or whatever... */
#define RFT_UNKNOWN   0
#define RFT_GIF       1
#define RFT_PM        2
#define RFT_PBM       3
#define RFT_XBM       4
#define RFT_SUNRAS    5
#define RFT_BMP       6
#define RFT_UTAHRLE   7
#define RFT_IRIS      8
#define RFT_PCX       9
#define RFT_JFIF     10
#define RFT_TIFF     11
#define RFT_PDSVICAR 12
#define RFT_COMPRESS 13
#define RFT_PS       14
#define RFT_IFF      15
#define RFT_TARGA    16
#define RFT_XPM      17
#define RFT_XWD      18
#define RFT_FITS     19
#define RFT_RAW      20
#define RFT_IMG      21



/* values used in Draw3dRect() */
#define R3D_OUT 0  /* rect sticks 'out' from screen */
#define R3D_IN  1  /* rect goes 'in' screen */


/* values 'GetSelType()' (in xvcut.c) can return */
#define SEL_RECT 0

/* mode values for PadPopUp() */
#define PAD_SOLID 0
#define PAD_BGGEN 1
#define PAD_LOAD  2

#define PAD_ORGB  0
#define PAD_OINT  1
#define PAD_OHUE  2
#define PAD_OSAT  3
#define PAD_OMAX  4


/* random string-placing definitions */
#define SPACING 3      /* vertical space between strings */
#define ASCENT   (mfinfo->ascent)
#define DESCENT  (mfinfo->descent)
#define CHIGH    (ASCENT + DESCENT)
#define LINEHIGH (CHIGH + SPACING)


#define MAX_ALLOCABLE_COLS 256     /* max number of colors alocated in GetImage */
#define MIN_ALLOCABLE_COLS 216     /* Maximum number of palette colors without
                                      reordening in GetImage*/

#define STDINSTR "<stdin>"

typedef unsigned char byte;


/* MACROS */
#define CENTERX(f,x,str) ((x)-XTextWidth(f,str, (int) strlen(str))/2)
#define CENTERY(f,y) ((y)-((f->ascent+f->descent)/2)+f->ascent)

/* RANGE forces a to be in the range b..c (inclusive) */
#define RANGE(a,b,c) { if (a < b) a = b;  if (a > c) a = c; }

/* PTINRECT returns '1' if x,y is in rect */
#define PTINRECT(x,y,rx,ry,rw,rh) \
           ((x)>=(rx) && (y)>=(ry) && (x)<((rx)+(rw)) && (y)<((ry)+(rh)))

/* MONO returns total intensity of r,g,b triple (i = .33R + .5G + .17B) */
#define MONO(rd,gn,bl) ( ((int)(rd)*11 + (int)(gn)*16 + (int)(bl)*5) >> 5)

/* ISPIPE returns true if the passed in character is considered the
   start of a 'load-from-pipe' or 'save-to-pipe' string */
#define ISPIPE(c) ((c)=='!' || (c)=='|')

/* CMAPVIS returns true if the visual class has a modifyable cmap */
#define CMAPVIS(v) (((v)->class == PseudoColor) || ((v)->class == GrayScale))

/* program name for printf's */
const char cmd[]="VISIO v0.11 by JOAKO" ;

#ifdef __MAIN__
    #define WHERE
#else
    #define WHERE extern 
#endif


/* info structure filled in by the LoadXXX() image reading routines */
typedef struct { byte *pic;                  /* image data */
     int   w, h;                 /* pic size */
   int   type;                 /* PIC8 or PIC24 */
    
     byte  r[256],g[256],b[256]; /* colormap, if PIC8 */

   int   normw, normh;         /* 'normal size' of image file (normally eq. w,h, 
                                    except when doing 'quick' load for icons */

  int   frmType;              /* def. Format type to save in */
  int   colType;              /* def. Color type to save in */
  char  fullInfo[128];        /* Format: field in info box */
  char  shrtInfo[128];        /* short format info */
  char *comment;              /* comment text */

  int   numpages;             /* # of page files, if >1 */
  char  pagebname[64];        /* basename of page files */
  } PICINFO;

// structure used to read IRIS files.
typedef struct {
    u_short  imagic;    /* stuff saved on disk . . */
    u_short   type;
    u_short   dim;
    u_short   xsize;
    u_short   ysize;
    u_short   zsize;
    u_long   min;
    u_long   max;
    u_long  wastebytes;  
    char   name[80];
    u_long  colormap;

    long   file;    /* stuff used in core only */
    u_short   flags;
    short  dorev;
    short  x;
    short  y;
    short  z;
    short  cnt;
    u_short  *ptr;
    u_short  *base;
    u_short  *tmpbuf;
    u_long  offset;
    u_long  rleend;    /* for rle images */
    u_long  *rowstart;  /* for rle images */
    long  *rowsize;  /* for rle images */
} IRIS_IMAGE;

// Structure used for reading FITS files.
typedef struct {
  FILE     *fp;          /* file pointer */
  int       bitpix;      /* number of bits per pixel */
  int       size;        /* Size of each pixel, in bytes */
  int       naxis;       /* number of axes */
  long int  axes[3];     /* size of each axis */
  long int  ndata;       /* number of elements in data */
  long int  cpos;        /* current position in data file */
  char     *comment;     /* malloc'ed comment string, or NULL if none */
} FITS;

/* Color histogram stuff. */
typedef unsigned char pixval;
typedef struct { pixval r, g, b; } pixel;
typedef struct chist_item* chist_vec;
typedef unsigned char pixval;
struct chist_item { pixel color;
      int value;
          };
typedef struct chist_list_item* chist_list;
struct chist_list_item { struct chist_item ch;
           chist_list next;
         };

#define MAX_GHANDS 16   /* maximum # of GRAF handles */

#define N_GFB 6
#define GFB_SPLINE 0
#define GFB_LINE   1
#define GFB_ADDH   2
#define GFB_DELH   3
#define GFB_RESET  4
#define GFB_GAMMA  5

#define GVMAX 8


typedef chist_list* chash_table;

// miscelanea functions
/*************************** MISCELA.C ***************************/
char *BaseName              ( char *);
void xvbzero                (char *, size_t);
int  xvbcmp                 (char *, char *, size_t);
char *inv_strstr            (char *string, char *substring);

#ifdef USE_OWN_MEMCPY_ROUTINES
  void xvbcopy                (char *, char *, size_t);
  char *xv_strstr             (char *, char *);
  void Quit                   ( int i);
  FILE *xv_fopen              (char *, char *);
#else // USE_OWN_MEMCPY_ROUTINES
  #define xvbcopy(src,dst,len)   memmove((dst),(src),(len))
  #define xv_strstr(a,b)         strstr((a),(b))
  #define Quit(i)                exit((i))
  #define xv_fopen(a,b)          fopen((a),(b))
#endif // USE_OWN_MEMCPY_ROUTINES

/* *************** HSV RGB stuff  in PAD.CPP ************** */
void hsv2rgb(double h, double s, double v, int *rr, int *gr, int *br);
void rgb2hsv(int r,int g,int b, double *hr, double *sr, double *vr);

/* max length of an Info String */
#define ISTRLEN 80

/* values 'popUp' can take */
#define ISPOPUP  1
#define ISALERT  2
#define ISGETSTR 3
#define ISGRAB   4
#define ISPAD    5

#ifdef HAVE_TIFF
#include "..\\tiff\\tiffio.h"
/* has to be after common.h, as it needs varargs/stdarg */
#endif

#endif // #define _COMMON_H_
