/*********************************************************************************\
*     File:               I O f i l e . C P P                                           *
*                                                                                 *
*     Defines the IOfile class for data manipulation of an image               *
*     Standard   C++                                                              *
\*********************************************************************************/

#include "IOfile.h"

/* include files */
#include <math.h>
#include <ctype.h>
#include <stddef.h>
#include <errno.h>

  
int IOfile::DEBUG = DEBUG_LEVEL;       /* debug level */
int IOfile::conv24 = CONV24_SLOW;      /* use 'SLOW' algorithm by default 
                                             (medium speed)*/
int IOfile::SaveMem = FALSE;           /* if not SaveMem, undoes are supported*/

//------------------------------------------------------------------------------
// Constructor of IOfile
//------------------------------------------------------------------------------
IOfile::IOfile(): pic(NULL),            /* none picture in memory */
                        pWIDE(0),             /* size of 'pic' */
                        pHIGH(0),
                        numPages(1),          /* default: no multipage image*/
                        curPage(0),
                        picType(PIC8),        /* gets set once file is loaded */
                        mono(0),               /* color default image */
                        picComments(NULL),    /* None comments in pointer */
                        numcols(0),           /* # of desired colors in picture */
                        origPic(NULL),        /* no undo option */
                        origPicType(PIC8),
                        origMono(0),
                        origWIDE(0),
                        origHIGH(0),
                        origNumcols(0)

{ /*****************************************************/
  /*** variable Initialization                       ***/
  /*****************************************************/
  pageBaseName[0] = '\0';/* None aux file path for multipage images */
  filename[0] = '\0';     /* None file loaded */
  formatStr[0] = '\0';   /* None comments in array */

  for (int i=0; i<NISTR  ;i++) istrs[i][0]='\0';
  for (i=0; i<256 ; i++)
          { rMap [i]= gMap [i]= bMap [i] = i; }

} //End of constructor of IOfile()

//------------------------------------------------------------------------------
// Destructor of IOfile
//------------------------------------------------------------------------------

IOfile::~IOfile()
{
   // Kill temporaly files
   KillPageFiles(pageBaseName, numPages);
   // Kill dinamic allocated memory ( pic, origpic & picComments)
   KillOldPics();
}




/********************************/ 
void IOfile::KillPageFiles(char *bname, int numpages)
{
  /* deletes any page files (numbered 1 through numpages) that might exist */
  char tmp[128];
  int  i;

  if (strlen(bname) == 0) return;   /* no page files */

  for (i=1; i<=numpages; i++) {
    sprintf(tmp, "%s%d", bname, i);
    unlink(tmp);
  }

}

/************************************/
void IOfile::KillOldPics(void)
{
  /* throw away all previous images and dinamic-alocated memory*/

  if (pic) free(pic);
  pic = (byte*) NULL;

  pWIDE=0;
  pHIGH=0;
  numcols = 0;
  
  KillOrigPic();

  if (picComments) free(picComments);
  picComments = (char *) NULL;

}

/************************************************************/
void IOfile::KillOrigPic()
{
  /* called whenver an image file is loaded.  disposes of origPic 
     if neccessary, and points it to null */

  if (origPic!=NULL) free(origPic);
  origPic = (byte *) NULL;

  origWIDE = 0;
  origHIGH = 0;
  origNumcols = 0;

}


/**************************************/
void IOfile::StoreNewPic()
{
  /* saves original picture into origPic, if it hasn't already been done.
   * This allows us to undo algorithms...  
   */
  if (origPic!=NULL) {
    free(origPic); // kill former oldpic, as well
    origPic = (byte*) NULL;
  }

  // if we are short of memory, don't save backup pics
  if (SaveMemory()) {
    origWIDE = 0;
    origHIGH = 0;
    origNumcols = 0;
    return; 
  }

  /* We do use the orig picture to support undo  */
  origPic = pic;
  pic = NULL; /* disposes the pic to be overwriten */
  origPicType = picType;
  origMono = mono;
  origWIDE = pWIDE;
  origHIGH = pHIGH;
  origNumcols = numcols;

  if (picType==PIC8) { //Copy colormap
     xvbcopy((char *)rMap, (char *)rorg,origNumcols);
     xvbcopy((char *)gMap, (char *)gorg,origNumcols);      
     xvbcopy((char *)bMap, (char *)borg,origNumcols);
  }
}

/**************************************/
void IOfile::StoreNewPic(byte *NewPic,  int NewPicType, int  NewMono,
                            u_int w, u_int h, u_int  NewNumCols,
                            byte *rp, byte *gp, byte *bp)
{
  /* saves original picture into origPic, if it hasn't already been done.
   * This allows us to undo algorithms...
   * colormap may be NULL or even IOfile's original. No problem!
   * If no size is given, it's suposed to be the same than before (except
   * if null image is given ).
   */

  if (NewPic!=pic) // kill former oldpic, if a actually new one has come
     StoreNewPic(); 
  
  pic =     NewPic;
  picType = NewPicType;
  mono =    (NewPicType==PIC8) ? NewMono : FALSE;
  numcols = (NewPicType==PIC8) ? NewNumCols : 0;
  if ( pic || !( pic || w || h )) {
    // Check if pic is NULL. If  pic is NULL, w and h must be zero
    
    // 
    if ( w && pic ) pWIDE = w;  else if (pic==NULL) pWIDE = 0;
    if ( h && pic ) pHIGH = h;  else if (pic==NULL) pHIGH = 0;
  
    if ( NewPicType==PIC8 && pic != NULL ) {
      if ( mono && NewNumCols > 2) {
        // INSTALL greyscale (it isn't a black and white image)
        register int i;
        for (i=256; i; ) {
          i--;
          rMap[i]=gMap[i]=bMap[i] = (byte) i;
        }
      } else if (mono) {
        // INSTALL black and white
          rMap[0]=gMap[0]=bMap[0] = 0;
          rMap[1]=gMap[1]=bMap[1] = 255;
      } else if (  rp != NULL && gp != NULL && bp != NULL  
          && rp != rMap && gp != gMap && bp != bMap) {
        xvbcopy((char *)rp, (char *)rMap, NewNumCols);
        xvbcopy((char *)gp, (char *)gMap, NewNumCols);      
        xvbcopy((char *)bp, (char *)bMap, NewNumCols);
      }
    } // ( NewPicType!=PIC8 || pic == NULL )
  }
}

/**************************************/
void IOfile::RestorePic()
{
  int sig;   /* Undo last action */
  u_int unsig;
  byte *auxptr;

  /* restore original picture */
  if (origPic==NULL) {
    Error("RestoreOldPic: imposible restaurar antigua imagen porque"
            " no ha sido guardada");
    return;  /* none to restore */
  }

  WaitCursor();

  /* don't toss the old pic away, save for a re-undo */

  sig= picType;      /* Interchange Type and origType */
  picType = origPicType;
  origPicType = sig;

  sig=mono;
  mono=origMono;
  origMono=sig;

  unsig=numcols;
  numcols = origNumcols;
  origNumcols=unsig;

  unsig=pWIDE;
  pWIDE=origWIDE;
  origWIDE=unsig;

  unsig=pHIGH;
  pHIGH=origHIGH;
  origHIGH=unsig;

  auxptr = pic;    /* Interchange pic and origPic */
  pic = origPic;
  origPic = auxptr;

  if (picType == PIC8) {
    register byte j;        /* Interchange color Map and original color Map */
    register int i;
    for (i=0; i<256; i++) {
      j=rMap[i]; rMap[i] = rorg[i]; rorg[i] = j;
      j=gMap[i]; gMap[i] = gorg[i]; gorg[i] = j;
      j=bMap[i]; bMap[i] = borg[i]; borg[i] = j;
    }
  }
  RedrawPic();
  ArrowCursor();
}

/*********************************************************************************/
void IOfile::ShowImage(byte *image,  int colType, char* menssage,
                                 int w, int h, 
                                 byte *rp, byte *gp, byte *bp)
{
  byte *FormerPic;
  int formerPicType, formerMono;
  u_int formerNumcols, formerWIDE, formerHIGH;
  byte formerRmap[256], formerGmap[256], formerBmap[256];

  if (!image) return;
  
  // Back-up former values
  FormerPic=pic;
  formerPicType = picType;
  formerMono = mono;
  formerNumcols = numcols;
  formerWIDE = pWIDE;
  formerHIGH = pHIGH;

  // Store provisional values
  pic=image;
  picType = ( colType == F_FULLCOLOR ) ? PIC24 : PIC8;
  mono = ( (colType == F_GREYSCALE) || (colType == F_BWDITHER) )? TRUE : FALSE;
  numcols = ( colType == F_FULLCOLOR ) ?   0 :
            (colType == F_BWDITHER)    ?   2 :
            (colType == F_GREYSCALE)   ? 256 :
            ( rp && gp && bp )         ? 256 : numcols; 
  
  if ( w>0 && h>0 && pic!=NULL ) {
    // Check if default values has been passed to this function
    // (if pic exists, then w and h must be non-zero)
    pWIDE = (u_int) w;
    pHIGH = (u_int) h;
  }
  
  // Store provisional colormap
  if (  picType==PIC8 && ( colType==F_GREYSCALE || colType==F_BWDITHER ||
       (colType==F_REDUCED && rp != NULL && gp != NULL &&  
        bp != NULL && rp != rMap && gp != gMap && bp != bMap) ) ) {
    // Back-up original colormap
      xvbcopy((char *)rMap, (char *)formerRmap, 256);
      xvbcopy((char *)gMap, (char *)formerGmap, 256);
      xvbcopy((char *)bMap, (char *)formerBmap, 256);

    if ( colType==F_GREYSCALE ) 
        // Grey image
        for (register int i=255; i; ){
          i--;
          rMap [i] = gMap [i] = bMap [i] = i;
        }
    else if ( colType==F_BWDITHER ) {
        // Black & White image
        rMap[0]=gMap[0]=bMap[0]=0;     // black
        rMap[1]=gMap[1]=bMap[1]=255;   // white
    }
    else  {
        // Colormapped image
        xvbcopy((char *)rp, (char *)rMap, numcols);
        xvbcopy((char *)gp, (char *)gMap, numcols);      
        xvbcopy((char *)bp, (char *)bMap, numcols);
    }
  }
  
  //Show pic
  RedrawPic();

  // Show message if exist
  if (menssage)
     Warning( menssage);

  // Restore original colormap if needed
  if (  picType==PIC8 && ( colType==F_GREYSCALE || colType==F_BWDITHER ||
       (colType==F_REDUCED && rp != NULL && gp != NULL &&  
        bp != NULL && rp != rMap && gp != gMap && bp != bMap) ) ) {
    xvbcopy((char *)formerRmap, (char *)rMap, 256);
    xvbcopy((char *)formerGmap, (char *)gMap, 256);
    xvbcopy((char *)formerBmap, (char *)bMap, 256);
  }

  // RESTORE former values
  pic=FormerPic;
  picType=formerPicType;
  mono = formerMono;
  numcols = formerNumcols;
  pWIDE = formerWIDE;
  pHIGH = formerHIGH;
}

static clock_t alg_start=0;

/***********************************************/
int IOfile::start24bitAlg(byte **pic24, byte **tmpPic)
{
  /* generates a 24-bit version of 'pic', if neccessary, and also mallocs
   * a pWIDE*pHIGH*3 24-bit output pic.  
   *
   * Returns '1' if there's some sort of screwup, '0' if cool
   */


  if (picType == PIC8) {
    *pic24 = Conv8to24(pic, pWIDE, pHIGH, rMap, gMap, bMap);
    if (!*pic24) return 1; 
  }
  else *pic24 = pic;

  /* need to create another w*h*3 pic to hold results */
  *tmpPic = (byte *) calloc((size_t) (pWIDE * pHIGH * 3), (size_t) 1);
  if (!(*tmpPic)) {
    Error("Imposible malloc() imagen temporal 24-bit en start24bitAlg()");
    if (picType == PIC8) free(*pic24);
    return 1;
  }

  /* Cronometer start*/
  alg_start = clock();
  return 0;
}



/***********************************************/
void IOfile::end24bitAlg(byte *pic24, byte *outPic)
{
  /* given pic24, and outPic, which has the new 24-bit image, installs it */
  
  /* Stop cronometer!*/
  clock_t alg_finish = clock();

  StoreNewPic();  /* also kills pic, NOT pic24 */


  if (picType == PIC8 && mono) {
    /* Grayscale or B/W */
      free(pic24);
      pic = (byte *) malloc((size_t) pWIDE * pHIGH);
      if (!pic)
        FatalError("Ha fallado el alojamiento de la imagen en escala de grises (end24bit)");
      else {
        int i; byte *p, *ep;
        for (i=pWIDE * pHIGH,p=pic,ep=outPic; i; i--, p++, ep+=3)
          *p = MONO(ep[0],ep[1],ep[2]);
      }
      numcols=256;
      mono=1;
      for (int i=256; i; ) {
        i--;
        rMap [i] = gMap [i] = bMap [i] = i;
      }
      free(outPic);
  }
  else if (picType == PIC8){
    /* colormapped -> copy results from pic24 */
    free(pic24);
    pic = Conv24to8(outPic, pWIDE, pHIGH, rMap,gMap,bMap, MAX_ALLOCABLE_COLS);
    free(outPic);
    if (!pic) { 
      ArrowCursor();
      Error("Algn fallo ocurri en la conversin 24to8\n");
      RestorePic(); 
      return;
    }
    numcols=MAX_ALLOCABLE_COLS;
  }
  else  { //if ( pic != NULL &&  PicType==PIC24)    
    /* Also, frees all pics, (except 'pic', if we're in PIC24 mode) */
    //pic24 cant be freed since origPic == pic24 in 'SaveOrigPic'
    pic = outPic;
    mono=0;
  } 

  RedrawPic();  // Redraw image.

  printf( "Tiempo discurrido durante el algoritmo: %2.1f segundos\n",
  (double)(alg_finish - alg_start) / CLOCKS_PER_SEC);
}

/***********************************************/
int IOfile::startGreyAlg(byte **picGrey, byte **tmpPic)
{
  /* generates a gray version of 'pic', if neccessary, and also mallocs
   * a pWIDE*pHIGH 256 greyscale output pic.  
   *
   * Returns '1' if there's some sort of screwup, '0' if cool
   */

  int ptype,pfree,colType=F_GREYSCALE;
  unsigned nc;
  byte *rp, *gp, *bp;

  if (picType == PIC8 && mono && numcols>2 ) {
    // pic is greyscaled
    *picGrey = pic;
  } else {
    // get a greyscaled version of pic
    *picGrey = GetImage(&ptype, colType, &pfree, &nc, &rp, &gp, &bp,
                      MAX_ALLOCABLE_COLS);
    if (!(*picGrey)) {
      Error("Imposible obtener imagen en startGreyAlg");
      return 1; 
    }
  }
  
  /* need to create another w*h pic to hold results */
  *tmpPic = (byte *) calloc((size_t) (pWIDE * pHIGH), (size_t) 1);
  if (!(*tmpPic)) {
    Error("Imposible malloc() imagen temporal en startGreyAlg()");
    if (pfree) free(*picGrey);
    return 1;
  }
  /* Cronometer start*/
  alg_start = clock();
  return 0;
}



/***********************************************/
void IOfile::endGreyAlg(byte *picGrey, byte *outPic)
{
  /* given pic24, and outPic, which has the new 24-bit image, installs it */
  
  /* Stop cronometer!*/
  clock_t alg_finish = clock();

  // Free grey initial image if is not the one of IOfile
  if ( picGrey!=pic && picGrey!=origPic && picGrey) free(picGrey);

  StoreNewPic();  /* do able to restore */

  pic=outPic;
  register int i=255;
  for (; i; i--)
    rMap [i] = gMap [i] = bMap [i] = i;
  rMap [0] = gMap [0] = bMap [0]= 0;
  picType=PIC8;
  numcols=256;
  mono=1; 
  RedrawPic();  // Redraw image.

  printf( "Tiempo discurrido durante el algoritmo: %2.1f segundos\n",
    (double)(alg_finish - alg_start) / CLOCKS_PER_SEC);
}

/***********************************************/
void IOfile::ElapsedTime( char* str, clock_t start)
{
   /* Measure the duration of an event. */
   double  duration;
   clock_t finish;
   finish = clock();
   duration = (double)(finish - start) / CLOCKS_PER_SEC;
   printf( "Tiempo discurrido en \"%s\": %2.1f segundos\n", str, duration );
}




