/*****************************************************************************
  Programa 	: itex.cpp 
  Proposito	: Leer una imagen de formato IM (itex 151)
  Autor    	: Luis Mario Cascante Nasarre de Letosa
  Fecha    	: 1-10-1997
  Ultima modif. : 22-10-1997 by JOAKO
*****************************************************************************/  

/*****************************************************************************
*                              I N C L U D E S                               *
******************************************************************************/

#include "common.h"
#include "IOfile.h"
#include "Itex.h"
#include "..\miguelan\miguelan.h"


/*****************************************************************************
*                                                                            *
*                              F U N C T I O N S                             *
*                                                                            *
******************************************************************************/
 


/*****************************************************************************
	Funciones para leer imagenes adquiridas con el procesador de imagenes 151
    Empiezan por IM. Formato en el Programmer's Manual 2 (aptdo. C-2) 
*****************************************************************************/	  

/****************************************************************************/
__inline int read_word (FILE *f)
/*****************************************************************************
 Lee dos bytes consecutivos de la imagen IM.
 *****************************************************************************/ 
{
#ifdef WIN32
  /* Los bytes en plataforma PC estn en el orden
     correcto */
#else
  /* Leer los bytes de forma genrica */
#endif
  register int retval = getc (f);
  return retval + (getc (f)<<8);
}

/****************************************************************************/
char* comprueba_flag (FILE *f)

/*****************************************************************************
 Coincide con los bytes 12-13 de la imagen IM.
 Esta al reves! 
 *****************************************************************************/ 
{
  int flag;  
  flag = read_word(f);
  if (flag == DIECISEIS_BITS)
     return "Imagen de 16-bits. No implementado";
  if (flag == COMPRIMIDA)
     return "Imagen comprimida. No implementado";
  /* si pasa esto es que es una imagen de 8-bits */
  return NULL;
}

/****************************************************************************/
void salta_zona_reservada (FILE *f)

/*****************************************************************************
 Coincide con los bytes 14-63 de la imagen IM.
 Se saltan sin mas! 
 *****************************************************************************/ 
{
  int i;   
  
  for (i=14; i<=63; i++)
      getc(f);

  return;
}


/****************************************************************************/
void salta_comentario (FILE *f, int long_comentario)
/*****************************************************************************
 * ATENCION: PROVISIONAL!!!
 * Coincide con los bytes 64-nnn donde nnn es 64+long. del comentario.
 * Acaba con un /0 (al menos eso dice el manual). Como la imagen ha sido obtenida
 * por el interprete POR AHORA simplemente saltaremos este campo... 
 *****************************************************************************/ 
{
  int i;   
  for (i=long_comentario; i; i--)
      getc(f);
  return;
}



/****************************************************************************/
#define BLOCK_SIZE 4096
#define NOT_DEBUGING


byte* lee_raw (FILE *fp, int tamano)

/*****************************************************************************
  Lee la Foto propiamente dicha: bytes del nnn (64+long_comentario)+1 hasta el
  final de la imagen.  Esta ordenado por filas con 1 byte por pixel. (8-bit)

*****************************************************************************/ 
{
  byte *picGrey = (byte *) malloc((size_t) tamano);
  if (!picGrey) {
    return NULL;
  }

#ifdef NOT_DEBUGING

   /* Read whole file once */
  fread( (void *) picGrey,1,tamano, fp); 

#else // NOT_DEBUGING

  /* PROBLEMS: Use better */
  byte *ptr = picGrey;
 /* Escribe el fichero */
  do {
     int nleidos, nbytes;
     nbytes = MIN(tamano,BLOCK_SIZE);   
     nleidos = fread((void *) ptr, 1 , nbytes ,fp);
     if ( nleidos != nbytes ) {
            break;
     }
     /* update pointer to next value */
     ptr += nleidos;
     tamano -= nleidos;
  } while (tamano>0);
#endif // NOT_DEBUGING
  return picGrey;
}


/****************************************************************************/
int IOfile::LoadIMG (char *nombre, PICINFO *pinfo)

/*****************************************************************************
  Intenta leer un fichero de formato pgm 
  devuelve un valor distinto de 0 en caso de xito
*****************************************************************************/
{
  
  /* returns '1' on success */

  FILE  *fp;
  int    c, c1, long_comentario;
  char   message[128],*bname,*msg;

  bname = BaseName(nombre);

  pinfo->pic     = (byte *) NULL;
  pinfo->comment = (char *) NULL;


  /* open the file */
  fp = xv_fopen(nombre,"rb");
  if (!fp) {
     sprintf(message,"No puedo abrir %s",bname);
     Warning(message);
     goto END;
  }

  /* Intenta leer el IM de la cabecera (bytes 0-1) */
  c = getc(fp);  c1 = getc(fp);
  if (c!='I' || c1!='M') {
    Warning ("lee_imagen_IM: La cabecera no es IM");
    goto END;
  }
         
  /* read in header information */

  long_comentario = read_word(fp);
  pinfo->h = read_word(fp);
  pinfo->w = read_word(fp);
  pinfo->normw = pinfo->w;   pinfo->normh = pinfo->h;
  /* coordenada_x = */ read_word(fp);
  /* coordenada_y = */ read_word(fp);
  msg = comprueba_flag (fp);
  if (msg) {
      Warning(msg);
      goto END;
  }
  salta_zona_reservada(fp);
  salta_comentario (fp,long_comentario);

  pinfo->frmType = F_RAW;
  pinfo->colType = F_GREYSCALE;

  sprintf(pinfo->fullInfo, "%dx%d Imaging Technology, 8-bits greyscale image",
    pinfo->w,pinfo->w);

  sprintf(pinfo->shrtInfo, "%dx%d IMG.", pinfo->w, pinfo->h);

  mono=TRUE; /* RAW images are allways greyscalled */
  numcols=256;

 
  /* Ahora llegamos a los (pinfo->h*pinfo->w) bytes
     equivalentes a la luminosidad de la imagen */

#ifdef NOT_DEBUGING
  pinfo->pic = lee_raw (fp, (pinfo->h)*(pinfo->w));
#else
{
  int tamano = (pinfo->h)*(pinfo->w);
  byte *picGrey = (byte *) malloc((size_t) tamano);
  if (!picGrey) {
    return NULL;
  }

   /* Read whole file once 
  fread( (void *) picGrey,1,contador, fp); */

  /* PROBLEMS: Use better */
  byte *ptr = picGrey;
  int error=0;
 /* Escribe el fichero */
  do {
     int nleidos, nbytes;
     nbytes = MIN(tamano,BLOCK_SIZE);   
     nleidos = fread((void *) ptr, 1 , nbytes ,fp);
     printf("LoadIMG: bytes ledos %6d de %6d (faltan %6d)\n",
       nleidos,nbytes,tamano);
     if ( nleidos != nbytes ) {
            break;
     }
     /* update pointer to next value */
     ptr += nleidos;
     tamano -= nleidos;
  } while (tamano>0);
 
  pinfo->pic = picGrey;
}
#endif

  if (!pinfo->pic) {
     sprintf(message,"No puedo leer %s por falta de memoria",bname);
     Warning(message);
  }

END:
  fclose(fp);
  if (!pinfo->pic) {
    /* error on reading IMG file */
    return pinfo->h = pinfo->w = 0;
  }

  /* Set Greyscale colormap */
  for (int i=0; i<256; i++)
    pinfo->b[i] = pinfo->g[i] = pinfo->r[i] = (byte) i;

  return 1;
}


