/*********************************************************************************\

*     File:                  J P E G . C P P                                      *
\*********************************************************************************/


/*
 * jpeg.cpp - i/o routines for 'jpeg' format pictures
 */

/* based in part on 'example.c' from the IJG JPEG distribution */

#include "IOfile.h"

#ifdef HAVE_JPEG


#ifdef _USE_SET_JUMP
#include <setjmp.h>
#endif //_USE_SET_JUMP

#include "..\\jpeg\\jpeglib.h"
#include "..\\jpeg\\jerror.h"

#define CREATOR_STR "CREATOR: "
#define REVDATE "8-11-96"

#if BITS_IN_JSAMPLE != 8
  Sorry, this code only copes with 8-bit JSAMPLEs. /* deliberate syntax err */
#endif

/* minimum size compression when doing a 'quick' image load.  (of course, if
   the image *is* smaller than this, you'll get whatever size it actually is.
   This is currently hardcoded to be twice the size of a schnauzer icon, as
   the schnauzer's the only thing that does a quick load... */

#define QUICKWIDE 160    
#define QUICKHIGH 120


struct my_error_mgr {
  struct jpeg_error_mgr pub;
#ifdef _USE_SET_JUMP
  jmp_buf               setjmp_buffer;
#endif //_USE_SET_JUMP
};

typedef struct my_error_mgr *my_error_ptr;


/*** local functions ***/
#ifdef __cplusplus
extern "C" {
#endif
METHODDEF void         xv_error_exit      (j_common_ptr);
METHODDEF void         xv_error_output    (j_common_ptr);
METHODDEF void         xv_prog_meter      (j_common_ptr);
METHODDEF boolean      xv_process_comment (j_decompress_ptr);
static    unsigned int j_getc             (j_decompress_ptr);
//static    int          writeJFIF          (FILE *, byte *, int,int,int);
#ifdef __cplusplus
}
#endif


/*** local variables ***/
static char *filename;
static char *fbasename;
static char *comment;

static int JPEGquality=100; // Default: highest quality
static int JPEGsmooth=0;    // Default: no smooth



/***************************************************************************/
/******** JPEG SAVE ROUTINES ***********************************************/
/***************************************************************************/

/*******************************************/
void IOfile::SetJPEGquality(int value)
{
  JPEGquality = value;
}
void IOfile::SetJPEGsmooth(int value)
{
  JPEGquality = value;
}
int IOfile::GetJPEGquality(void)
{
  return JPEGquality;
}
int IOfile::GetJPEGsmooth(void)
{
  return JPEGquality;
}
                                   
/*******************************************/
void IOfile::writeJPEG(int colorType)
{
  FILE          *fp;
  u_int          nc,i;
  int            rv, w, h, ptype, pfree;
  register byte *ip;
  byte          *inpix, *rmap, *gmap, *bmap;

  /* get the XV image into a format that the JPEG software can grok on.
     Also, open the output file, so we don't waste time doing this format
     conversion if we won't be able to write it out */


  fp = OpenOutFile(filename);
  if (!fp) return;

  fbasename = BaseName(filename);

  WaitCursor();

  /* monocity: see if the image is mono, save it that way to save space */
  if (colorType != F_GREYSCALE) {
    if (ptype == PIC8) {
      for (i=0; i<nc && rmap[i]==gmap[i] && rmap[i]==bmap[i]; i++);
      if (i==nc)  /* made it all the way through */
        colorType = F_GREYSCALE;
      else
        colorType = F_FULLCOLOR;
    }
    else {  /* PIC24 */
      for (i=pWIDE*pHIGH,ip=pic; i && ip[0]==ip[1] && ip[1]==ip[2]; i--,ip+=3);
      if (i==0)  /* all the way through */
        colorType = F_GREYSCALE;
      else
        colorType = F_FULLCOLOR;
    }
  }

  inpix = GetImage(&ptype, colorType, &pfree, &nc, &rmap, &gmap, &bmap,
                     MAX_ALLOCABLE_COLS);
  w=pWIDE; h=pHIGH;

  /* in any event, we've got some valid image.  Do the JPEG Thing */
  rv = writeJFIF(fp, inpix, w, h, colorType);
  
  if (pfree) free(inpix);
  
  CloseOutFile(fp, filename, rv);
  ArrowCursor();
}


/***************************************************************************/
/* JPEG INTERFACE ROUTINES *************************************************/
/***************************************************************************/
extern "C" {
/**************************************************/

METHODDEF void xv_error_exit( j_common_ptr cinfo) 
{
  my_error_ptr myerr;

  myerr = (my_error_ptr) cinfo->err;
  (*cinfo->err->output_message)(cinfo);     /* display error message */
#ifdef _USE_SET_JUMP
  longjmp(myerr->setjmp_buffer, 1);         /* return from error */
#endif //_USE_SET_JUMP
}


/**************************************************/
METHODDEF void xv_error_output(j_common_ptr cinfo) 
{
  my_error_ptr myerr;
  char         buffer[JMSG_LENGTH_MAX];

  myerr = (my_error_ptr) cinfo->err;
  (*cinfo->err->format_message)(cinfo, buffer);

  /* We can't use object function to display errors since we call this function
     by means of a generic pointer of function */
  //SetISTR(ISTR_WARNING, "%s: %s", fbasename, buffer);
   fprintf(stderr,"%s: %s", fbasename, buffer);
}


/**************************************************/
METHODDEF void xv_prog_meter(j_common_ptr cinfo)
{
  struct jpeg_progress_mgr *prog;

  prog = cinfo->progress;

//  if ((prog->pass_counter & 0x3f)==0) WaitCursor();

#ifdef FOO
  fprintf(stderr,"xv_prog_meter: cnt=%ld, maxcnt=%ld, pass=%d, maxpass=%d\n",
    prog->pass_counter, prog->pass_limit, prog->completed_passes,
    prog->total_passes);
#endif
}
/************************************/
} //extern "C" 


/***************************************************************************/
/* LOAD ROUTINES ***********************************************************/
/***************************************************************************/

/*******************************************/
int IOfile::LoadJFIF(char *fname, PICINFO *pinfo, int quick)
{
  /* returns '1' on success, '0' on failure */

  struct jpeg_decompress_struct    cinfo;
  struct jpeg_progress_mgr         prog;
  struct my_error_mgr              jerr;
  JSAMPROW                         rowptr[1];
  FILE                            *fp;
  static byte                     *pic;
  long                             filesize;
  int                              i,w,h,bperpix;


  fbasename = BaseName(fname);
  pic       = (byte *) NULL;
  comment   = (char *) NULL;

  pinfo->type  = PIC8;

  if ((fp = xv_fopen(fname, "rb")) == NULL) return 0;

  fseek(fp, 0L, 2);
  filesize = ftell(fp);
  fseek(fp, 0L, 0);


  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit     = xv_error_exit;
  jerr.pub.output_message = xv_error_output;

  #ifdef _USE_SET_JUMP
  /* Set up C exceptions handler */
  if (setjmp(jerr.setjmp_buffer)) {
    // if we're here, it blowed up... 
    jpeg_destroy_decompress(&cinfo);
    fclose(fp);
    if (pic)     { free(pic); pic=(byte *) NULL; }
    if (comment) { free(comment); comment = (char *) NULL; }
    if (picComments) { free(picComments); picComments = (char *) NULL; }
    IOfile();
    return 0;
  }
  #endif _USE_SET_JUMP

  jpeg_create_decompress(&cinfo);
  jpeg_set_marker_processor(&cinfo, JPEG_COM, xv_process_comment);

  /* hook up progress meter */
  prog.progress_monitor = xv_prog_meter;
  cinfo.progress = &prog;

  jpeg_stdio_src(&cinfo, fp);
  (void) jpeg_read_header(&cinfo, TRUE);



  /* do various cleverness regarding decompression parameters & such... */



  jpeg_calc_output_dimensions(&cinfo);
  w = cinfo.output_width;
  h = cinfo.output_height;
  pinfo->normw = w;  pinfo->normh = h;

  if (quick) {
    int wfac, hfac, fac;
    wfac = w / QUICKWIDE;
    hfac = h / QUICKHIGH;

    fac = wfac;  if (fac > hfac) fac = hfac;
    if      (fac > 8) fac = 8;
    else if (fac > 4) fac = 4;
    else if (fac > 2) fac = 2;
    else fac = 1;

    cinfo.scale_num   = 1;
    cinfo.scale_denom = fac;
    cinfo.dct_method = JDCT_FASTEST;
    cinfo.do_fancy_upsampling = FALSE;

    pinfo->normw = w;  pinfo->normh = h;

    jpeg_calc_output_dimensions(&cinfo);
    w = cinfo.output_width;
    h = cinfo.output_height;
  }


  if (cinfo.jpeg_color_space == JCS_GRAYSCALE) {
    cinfo.out_color_space = JCS_GRAYSCALE;
    cinfo.quantize_colors = FALSE;
    
    SetISTR(ISTR_INFO,"Loading %dx%d Greyscale JPEG (%ld bytes)...",
      w,h,filesize);
    
    for (i=0; i<256; i++) pinfo->r[i] = pinfo->g[i] = pinfo->b[i] = i;
  }
  else {
    cinfo.out_color_space = JCS_RGB;
    cinfo.quantize_colors = FALSE;     /* default: give 24-bit image to XV */
        
    SetISTR(ISTR_INFO,"Loading %dx%d Color JPEG (%ld bytes)...",
      w,h,filesize);
  }
  
  jpeg_calc_output_dimensions(&cinfo);   /* note colorspace changes... */
    

  if (cinfo.output_components != 1 && cinfo.output_components != 3) {
    SetISTR(ISTR_WARNING, "%s:  can't read %d-plane JPEG file!",
      fbasename, cinfo.output_components);
    jpeg_destroy_decompress(&cinfo);
    fclose(fp);
    if (comment) free(comment);
    return 0;
  }


  bperpix = cinfo.output_components;
  pinfo->type = (bperpix == 1) ? PIC8 : PIC24;

  pic = (byte *) malloc((size_t) (w * h * bperpix));
  if (!pic) {
    SetISTR(ISTR_WARNING, "%s:  can't read JPEG file - out of memory",
      fbasename);
    jpeg_destroy_decompress(&cinfo);
    fclose(fp);
    if (comment) free(comment);
    return 0;
  }
  
  jpeg_start_decompress(&cinfo);

  while (cinfo.output_scanline < cinfo.output_height) {
    rowptr[0] = (JSAMPROW) &pic[cinfo.output_scanline * w * bperpix];
    (void) jpeg_read_scanlines(&cinfo, rowptr, (JDIMENSION) 1);
  }

  

  /* return 'PICINFO' structure to XV */

  pinfo->pic = pic;
  pinfo->w = w;
  pinfo->h = h;
  pinfo->frmType = F_JPEG;

  if (cinfo.out_color_space == JCS_GRAYSCALE) {
    sprintf(pinfo->fullInfo, "Greyscale JPEG. (%ld bytes)", filesize);
    pinfo->colType = F_GREYSCALE;
    
    for (i=0; i<256; i++) pinfo->r[i] = pinfo->g[i] = pinfo->b[i] = i;
  }
  else {
    sprintf(pinfo->fullInfo, "Color JPEG. (%ld bytes)", filesize);
    pinfo->colType = F_FULLCOLOR;

    if (cinfo.quantize_colors) {
      for (i=0; i<cinfo.actual_number_of_colors; i++) {
  pinfo->r[i] = cinfo.colormap[0][i];
  pinfo->g[i] = cinfo.colormap[1][i];
  pinfo->b[i] = cinfo.colormap[2][i];
      }
    }
  }
  
  sprintf(pinfo->shrtInfo, "%dx%d %s JPEG. ", w,h, 
    (cinfo.out_color_space == JCS_GRAYSCALE) ? "Greyscale " : "Color ");
  
  pinfo->comment = comment;

  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);
  fclose(fp);

  comment = (char *) NULL;
  return 1;
}
  

/**************************************************/
extern "C" {
/**************************************************/
static unsigned int j_getc( j_decompress_ptr cinfo)
{
  struct jpeg_source_mgr *datasrc = cinfo->src;
  
  if (datasrc->bytes_in_buffer == 0) {
    if (! (*datasrc->fill_input_buffer) (cinfo))
      ERREXIT(cinfo, JERR_CANT_SUSPEND);
  }
  datasrc->bytes_in_buffer--;
  return GETJOCTET(*datasrc->next_input_byte++);
}


/**************************************************/
METHODDEF boolean xv_process_comment(j_decompress_ptr cinfo)
{
  int          length, hasnull;
  unsigned int ch;
  char         *oldsp, *sp;

  length  = j_getc(cinfo) << 8;
  length += j_getc(cinfo);
  length -= 2;                  /* discount the length word itself */

  if (!comment) {
    comment = (char *) malloc((size_t) length + 1);
    if (comment) comment[0] = '\0';
  }
  else comment = (char *) realloc(comment, strlen(comment) + length + 1);
  /* We can't use object function to display errors since we call this function
     by means of a generic pointer of function */
  //if (!comment) FatalError("out of memory in xv_process_comment");
  if (!comment) return FALSE; 
  
  oldsp = sp = comment + strlen(comment);
  hasnull = 0;

  while (length-- > 0) {
    ch = j_getc(cinfo);
    *sp++ = (char) ch;
    if (ch == 0) hasnull = 1;
  }

  if (hasnull) sp = oldsp;       /* swallow comment blocks that have nulls */
  *sp++ = '\0';

  return TRUE;
}


/**************************************************/
} //extern "C"


/***************************************************************************/
/* WRITE ROUTINES **********************************************************/
/***************************************************************************/
int IOfile::writeJFIF(FILE *fp, byte *pic, int w, int h, int coltype)
{
  struct     jpeg_compress_struct cinfo;
  struct     jpeg_progress_mgr    prog;
  struct     my_error_mgr         jerr;
  JSAMPROW                        rowptr[1];
  int                             i, bperpix;
  char                            xvcmt[256];

  comment = (char *) NULL;

  cinfo.err               = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit     = xv_error_exit;
  jerr.pub.output_message = xv_error_output;

#ifdef _USE_SET_JUMP
  /* Set up C exception handler */
  if (setjmp(jerr.setjmp_buffer)) {
    /* if we're here, it blowed up... */
    jpeg_destroy_compress(&cinfo);
    if (pic)     { free(pic); pic=(byte *) NULL; }
    if (comment) { free(comment); comment = (char *) NULL; }
    if (picComments) { free(picComments); picComments = (char *) NULL; }
    IOfile();
    return 1;
  } 
#endif //_USE_SET_JUMP

  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, fp);

  cinfo.image_width  = w;
  cinfo.image_height = h;
  if (coltype == F_GREYSCALE) {
    cinfo.input_components = 1;
    cinfo.in_color_space = JCS_GRAYSCALE;
  }
  else {
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_RGB;
  }

  bperpix = cinfo.input_components;


  prog.progress_monitor = xv_prog_meter;
  cinfo.progress = &prog;


  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, JPEGquality, TRUE);
  cinfo.smoothing_factor = JPEGsmooth;


  jpeg_start_compress(&cinfo, TRUE);


  /*** COMMENT HANDLING ***/

  sprintf(xvcmt, "%sDiblook %s  Quality = %d, Smoothing = %d\n",
    CREATOR_STR, REVDATE, JPEGquality, JPEGsmooth);
  
  if (picComments) {   /* append XV comment */
    char *sp, *sp1;  int done;

    i   = strlen(picComments);
    comment = (char *) malloc(i + strlen(xvcmt) + 2 + 1);
    if (!comment) FatalError("out of memory in writeJFIF()");
    
    strcpy(comment, picComments);
    
    /* see if there's a line that starts with 'CREATOR: ' in the
       comments.  If there is, rip it out. */
    
    sp = comment;  done = 0;
    while (!done && *sp) {
      if (strncmp(sp, CREATOR_STR, strlen(CREATOR_STR)) == 0) {
        sp1 = sp;
        while (*sp1 && *sp1 != '\n') sp1++;    /* find end of this line */
        if (*sp1 == '\n') sp1++;               /* move past \n */

        /* move comments from sp1 and on down to sp */
        xvbcopy(sp1, sp, strlen(sp1) + 1);   /* +1 to copy the trailing \0 */

        done = 1;
      }
      else {   /* skip ahead to next line */
        while (*sp && *sp != '\n') sp++;
        if (*sp == '\n') sp++;
      }
    }

    /* count # of \n's at end of comment.  
       If none, add 2.   If one, add 1.  If two or more, add none. */

    sp = comment + strlen(comment);
    for (i=0; i<3 && ((size_t) i < strlen(comment)); i++) {
      sp--;
      if (*sp != '\n') break;
    }

    for ( ; i<2; i++) strcat(comment, "\n");
    strcat(comment, xvcmt);
  }
  else comment = xvcmt;
  
  
  jpeg_write_marker(&cinfo,JPEG_COM,(byte *) comment,(u_int) strlen(comment));
  
  while (cinfo.next_scanline < cinfo.image_height) {
    rowptr[0] = (JSAMPROW) &pic[cinfo.next_scanline * w * bperpix];
    (void) jpeg_write_scanlines(&cinfo, rowptr, (JDIMENSION) 1);
  }
  
  jpeg_finish_compress(&cinfo);
  jpeg_destroy_compress(&cinfo);
  return 0;
}




#endif  /* HAVE_JPEG */

