/*********************************************************************************\
*     File:                  M I S C E L A  . C P P                                   *
*                                                                                 *
*     random 'handy' routines                                                     *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/


/*
 * miscela.c - random 'handy' routines
 *
 *  Contains:
 *     char  *BaseName(char *)
 *     void   xvbcopy(src, dst, length)
 *     int    xvbcmp (s1,  s2,  length)
 *     void   xvbzero(s, length)
 *     char  *xv_strstr(s1, s2)
 *     char  *inv_strstr(s1, s2)
 *     void   ProgressMeter(min, max, val, str);
 *     void   CropRect2Rect (xp,yp,wp,hp,cx,cy,cw,ch);
 *               crops rect xp,yp,wp,hp to be entirely within bounds of cx,cy,cw,ch
 *
 */

#include "IOfile.h"

/***********************************/
//------------------------------------------------------------------------------
// Shows a message window.
//------------------------------------------------------------------------------
/* SetISTR( ISTR, format, arg1, arg2, ...)     */
// We use the standard C form for multiple parameters
void IOfile::SetISTR(int stnum, ...)
{
  va_list args;
  char     *fmt;
  
  va_start(args, stnum);

  if (stnum>0 && stnum < NISTR) {
    fmt = va_arg(args, char *);    
    if (fmt) {
      vsprintf(istrs[stnum], fmt, args);
      for(char *ptr=istrs[stnum];*ptr && !isalnum(*ptr);ptr++);
      if (*ptr) {//Don't print empty strings
        printf(istrs[stnum]);
        printf("\n");
      }
    }
    else istrs[stnum][0] = '\0';
  }
  va_end(args);
}


/***************************************************/

//------------------------------------------------------------------------------
// Returns message info string.
//------------------------------------------------------------------------------
char *IOfile::GetISTR( int stnum)
{
  /* returns pointer to ISTR string */
  if (stnum < 0 || stnum>=NISTR) return(NULL);
  return (istrs[stnum]);
}

/***************************************************/
void IOfile::FilterString( char *buf, char *gsFilter, int allow)
{
  /* handle characters typed at GetStrPopUp window.  Button accel. keys
     have already been checked for elsewhere.  Practical upshot is that
     we don't have to do anything with ESC or Return (as these will normally
     be Cancel and Ok buttons) 
 
     Normally returns '0'.  Returns '1' if character wasn't accepted, for
     whatever reason. */

  int i,position=0;
  char* ptr = buf;

  if (!buf || !gsFilter) return;

  for(; *ptr ; ptr++)
  {                                               
    if (*ptr>=' ' && *ptr<'\177') {               /* 'NORMAL' CHARACTER */
      if (allow){   /* check filter string */
          for (i=0; gsFilter[i] && *ptr!=gsFilter[i]; i++);
      
          if (gsFilter[i]) 
            buf[position++]=*ptr; /* character in filter */
          else
            buf[position++]=' '; /* character not in filter -> space */
      }
      else buf[position++]=*ptr;
    }

    else if (*ptr=='\010' || *ptr=='\177') {    /* BS or DEL */
      if (position==0) break;                     /* at beginning of str */
      position++;
    }

    else if (*ptr=='\025') {                 /* ^U: clear entire line */
      position=0;
    }

    else if (*ptr=='\013') {                 /* ^K: clear to end of line */
      buf[position++]='\0';
    }

    else if (*ptr=='\001') {                 /* ^A: move to beginning */
      position = 0;
    }

    else if (*ptr=='\005') {                 /* ^E: move to end */
      break;
    }

    else if (*ptr=='\004') {                 /* ^D: delete character at gsCurPos */
    }

    else if (*ptr=='\002') {                 /* ^B: move backwards char */
      if (position==0) break;
      position--;
    }

    else if (*ptr=='\006') {                 /* ^F: move forwards char */
      if (!*(ptr+1)) break;
      buf[position++]=' ';
    }

    else break;                        /* unhandled character */
  }
  buf[position]='\0';
  return ;
}

/***************************************************/
/*
void IOfile::TextView       (char *str) 
{
    char command[MAXPATHLEN + 10];
    sprintf(command,"notepad %s",str);
    _flushall(); // fflush all opened streams before the system call
    system(command);
}
*/

/***************************************************/
char *BaseName(char *fname)
{
  char *basname;

  /* given a complete path name ('c:\foo\bar\weenie.gif'), returns just the 
     'simple' name ('weenie.gif').  Note that it does not make a copy of
     the name, so don't be modifying it... */

  basname = (char *) strrchr(fname, '\\');
  if (!basname) basname = fname;
  else basname++;

  return basname;
}


/***************************************************
/* void ProgressMeter(int min, int max, int val, char *str)
{
     called during 'long' operations (algorithms, smoothing, etc.) to
     give some indication that the program will ever finish.  Draws a 
     temperature gauge in either mainW (if not useRoot), or ctrlW.
     Tries to be clever:  only draws gauge if it looks like the operation is
     going to take more than a few seconds.  Calling with val == max removes
     the gauge. 
}*/

/***************************************************/
int xvbcmp (char *s1, char *s2, size_t len)
{
  for ( ; len ; len--, s1++, s2++) {
    if      (*s1 < *s2) return -1;
    else if (*s1 > *s2) return 1;
  }
  return 0;
}

/***************************************************/
void xvbzero( char *s, size_t len)
{
  for ( ; len ; len--) *s++ = 0;
}

/***********************************
void WaitCursor(void)
{
  // Do nothing in text only version
}

/***********************************
void ArrowCursor(void)
{
  // Do nothing in text only version
}

/***********************************/
void IOfile::CropRect2Rect( int *xp, int *yp, int *wp, int *hp,
                               int cx, int cy, int cw, int ch)
{
  /* crops rect xp,yp,wp,hp to be entirely within bounds of cx,cy,cw,ch */

  int x1,y1,x2,y2;

  x1 = *xp;            y1 = *yp;
  x2 = *xp + *wp - 1;  y2 = *yp + *hp - 1;
  RANGE(x1, cx, cx+cw-1);
  RANGE(y1, cy, cy+ch-1);
  RANGE(x2, cx, cx+cw-1);
  RANGE(y2, cy, cy+ch-1);

  if (x2<x1) x2=x1;
  if (y2<y1) y2=y1;

  *xp = x1;           *yp = y1;
  *wp = (x2 - x1)+1;  *hp = (y2 - y1)+1;
}

/**********************************************************************/  
  void IOfile::GetSelRCoords( int *xp, int *yp, int *wp, int *hp)
{
  /* returns selection rectangle x,y,w,h in pic coordinates */

  /* NOTE: At this no selection is implemented, so returns the values 
      corresponding to the hole image. */

  *xp = 0;  *yp = 0;  
  *wp = pWIDE;  *hp = pHIGH;
}

/**********************************************************************/  
  void IOfile::ProgressMeter(int min, int max, int val, char *str)
{
  /* called during 'long' operations (algorithms, smoothing, etc.) to
     give some indication that the program will ever finish.  Draws a 
     temperature gauge in either mainW (if not useRoot), or ctrlW.
     Tries to be clever:  only draws gauge if it looks like the operation is
     going to take more than a few seconds.  Calling with val == max removes
     the gauge. */

  static int    waiting  = 0;
  static time_t lastTime = 0;
  time_t        nowTime;
  double        doneness;

  if (min >= max) return;
  if (val==max && lastTime==0) return;  /* already erased meter */

  doneness = ((double) (val-min)) / ((double) (max - min));

  time(&nowTime);

  if (lastTime == 0) {   /* first time we're calling for this meter */
    lastTime = nowTime;
    waiting = 1;                           
    printf("\n%s :",str);
  }
  else if (waiting) {    /* haven't drawn meter yet... */
    if ((nowTime - lastTime >= 2) && doneness < .5) waiting = 0;
  }

  if (!waiting) {     /* not waiting (or not waiting any longer) */
    if (nowTime == lastTime && val<max) return;  /* max one draw per second */
    lastTime = nowTime;
    if (val >= max) {            /* remove temp gauge */
      printf("Hecho!\n\n");
      lastTime = 0;
    }
  }
}

/**********************************************************************/  
/* Ignore at the moment */
int IOfile::UncompressFile    (char *, char *) {return 0; };

/**********************************************************************/  
/* Ignore at the moment */
int IOfile::HaveSelection  (void) {return 0; }; /* default ignore */

/**********************************************************************/  


#ifdef USE_OWN_MEMCPY_ROUTINES

/***********************************/
void Quit( int i)
{ 
  /* called when the program exits.  frees everything explictly created
     EXCEPT allocated colors.  This is used when 'useroot' is in operation,
     as we have to keep the alloc'd colors around, but we don't want anything
     else to stay */
  exit(i);
}

/***************************************************/
void xvbcopy(char *src, char *dst, size_t  len)
{
  /* Modern OS's (Solaris, etc.) frown upon the use of bcopy(),
   * and only want you to use memcpy().  However, memcpy() is broken,
   * in the sense that it doesn't properly handle overlapped regions
   * of memory.  This routine does, and also has its arguments in
   * the same order as bcopy() did, without using bcopy().
   */

  /* determine if the regions overlap
   *
   * 3 cases:  src=dst, src<dst, src>dst
   *
   * if src=dst, they overlap completely, but nothing needs to be moved
   * if src<dst and src+len>dst then they overlap
   * if src>dst and src<dst+len then they overlap
   */

  if (src==dst || len<=0) return;    /* nothin' to do */
  
  if (src<dst && src+len>dst) {  /* do a backward copy */
    src = src + len - 1;
    dst = dst + len - 1;
    for ( ; len>0; len--, src--, dst--) *dst = *src;
  }

  else {  /* they either overlap (src>dst) or they don't overlap */
    /* do a forward copy */
    for ( ; len>0; len--, src++, dst++) *dst = *src;
  }
}
    
/***************************************************/
char *inv_strstr(char *string, char *substring)
//     char *string;          /* String to search. */
//     char *substring;    /* Substring to try to find in string. */
// The inv_strstr function returns a pointer to the last occurrence
// of string2 in string1. The search does not include terminating null characters.
// Do the same than strstr but starting to find from the end to the start
{
  register char *a, *b;
  char *end_string=string + strlen(substring)-1;
  char *end_substring = substring+strlen(substring)-1;

  /* First scan quickly through the two strings looking for a
   * single-character match.  When it's found, then compare the
   * rest of the substring.
   */

  b = end_substring;
  if (b == substring) return string;

  for ( ; string != end_string; end_string -= 1) {
    if (*end_string != *b) continue;

    a = end_string;
    while (1) {
      if (b == substring) return a;
      if (*a-- != *b--) break;
    }
    b = end_substring;
  }
  return (char *) 0;
}




/***************************************************/

/* 
 *  Source code for the "strstr" library routine.
 *
 * Copyright 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

char *xv_strstr(char *string, char *substring)
//     char *string;          /* String to search. */
//     char *substring;    /* Substring to try to find in string. */
{
  register char *a, *b;

  /* First scan quickly through the two strings looking for a
   * single-character match.  When it's found, then compare the
   * rest of the substring.
   */

  b = substring;
  if (*b == 0) return string;

  for ( ; *string != 0; string += 1) {
    if (*string != *b) continue;

    a = string;
    while (1) {
      if (*b == 0) return string;
      if (*a++ != *b++) break;
    }
    b = substring;
  }
  return (char *) 0;
}

/***************************************************/

FILE *xv_fopen(char *fname, char *mode)
{
  return fopen(fname,mode);
}

/***************************************************/

#endif   //USE_OWN_MEMCPY_ROUTINES

