/*********************************************************************************\
*     File:                  O P E N P I C . C P P                                *
*                                                                                 *
*     Standard   C++                                                              *
*                                                                                 *
\*********************************************************************************/


#include "IOfile.h"

/* include files */
#include <io.h>
#include <ctype.h>
#include <stddef.h>
#include <errno.h>



/***********************************/
int IOfile::OpenPic(int filenum)
{
  /* tries to load file filename
   * returns 0 on failure (cleans up after itself)
   * returns '-1' if caller should display DFLTPIC  (shown as text)
   * if successful, returns 1
   *
   * By the way, I'd just like to point out that this procedure has gotten
   * *way* out of hand...
   * Now, the only valid values of filenum are DFLTPIC, LOADPIC, RELOAD
   * and for multipage documents OP_PAGEDN and OP_PAGEUP.
   */

  PICINFO pinfo;
  int   i,filetype, killpage;
  char *tmp;
  char   basefname[MAXPATHLEN];   /* just the current fname, no path */
  char   AuxFileName[MAXPATHLEN];    /* file to load */
  /* set default field of PICINFO struct */
  xvbzero((char *) &pinfo, sizeof(PICINFO));

  /* init important fields of pinfo */
  pinfo.pic = (byte *) NULL;
  pinfo.comment = (char *) NULL;
  pinfo.numpages = 1;
  pinfo.pagebname[0] = '\0';

  StoreNewPic();
 
  /* set default values of IOfile */
  killpage = 0;

  WaitCursor();

  SetISTR(ISTR_INFO,"");
  SetISTR(ISTR_WARNING,"");

  /* MULTIPAGE DOCUMENTS */ 
  /* if we're not loading next or prev page in a multi-page doc, kill off
     page files*/ 
  if (strlen(pageBaseName) && filenum!=OP_PAGEDN && filenum!=OP_PAGEUP) 
    killpage = 1;

  if (strlen(pageBaseName) && (filenum==OP_PAGEDN || filenum==OP_PAGEUP)) {
    if      (filenum==OP_PAGEUP && curPage>0)          curPage--;
    else if (filenum==OP_PAGEDN && curPage<numPages-1) curPage++;
    else    {
      ArrowCursor();
      return 0;
    }

    sprintf(AuxFileName, "%s%d", pageBaseName, curPage+1);
    strcpy(filename, AuxFileName);
    goto HAVE_FILENAME;
  }


  if (filenum == DFLTPIC) {
    filename[0] = '\0';  basefname[0] = '\0';  filename[0] = '\0';
    //LoadDfltPic(&pinfo);

    if (killpage) {      /* kill old page files, if any */
      KillPageFiles(pageBaseName, numPages);
      pageBaseName[0] = '\0';
      numPages = 1;
      curPage = 0;
    }

    goto GOTIMAGE;
  }
  
  if ((filenum != LOADPIC) && (filenum != RELOAD)) {
    goto FAILED;   
  }

  /* set up filename and basefname if a new pic has to be loaded */
  if (filenum == LOADPIC)  strncpy(filename,GetFileName( TRUE ), MAXPATHLEN);

  /* if > character is at the begining of the filename, it is a redirection of stdout
       * rather than a image file, so return with code sucessfull */
  /* if > character is at the begining of the filename, discard rest of string */
  if (filenum == LOADPIC) {
    char* ptr= strchr(filename,'>');
    if (ptr != NULL) {
      *ptr='\0';
      for (; (ptr > filename);) {
        ptr--;
        if (isalnum(*ptr)) break;
      }
      if (ptr == filename) return 1;
    }
  }

#ifdef GUNZIP
  /* zip COMPRESSED FILENAMES */
  /* chop off trailing ".Z", ".z", or ".gz" from displayed basefname, if any */
  if (strlen(basefname) > (size_t) 2     && 
      strcmp(basefname+strlen(basefname)-2,".Z")==0)
    basefname[strlen(basefname)-2]='\0';
  else {
    if (strlen(basefname)>2 && strcmp(basefname+strlen(basefname)-2,".z")==0)
      basefname[strlen(basefname)-2]='\0';

    else if (strlen(basefname)>3 && 
       strcmp(basefname+strlen(basefname)-3,".gz")==0)
      basefname[strlen(basefname)-3]='\0';
  }
#endif /* GUNZIP */

  /* if reload, use the previous file name */
  strcpy(AuxFileName, filename);
  tmp = BaseName(filename);
  strcpy(basefname, tmp);

  /*
    if (filenum == LOADPIC && ISPIPE(filename[0])) {
    if we're reading from a pipe, 'filename' will have the /tmp/xvXXXXXX
    filename, and we can skip a lot of stuff:  (such as prepending 
    'initdir' to relative paths, dealing with reading from stdin, etc.
    at this point, filename = "! do some commands", filename = "/tmp/xv123456",
    and basefname= "xv123456" 
  } */  
  {  /* NOT reading from a PIPE */
    FILE *fp;
    if (strcmp(filename,STDINSTR)==0) {
      /* if the file is STDIN, write it out to a temp file */
      sprintf(AuxFileName, "xvXXXXXX");
      mktemp(AuxFileName);

      clearerr(stdin);
      fp = fopen(AuxFileName,"wb");
      if (!fp) FatalError("OpenPic(): No puedo escribir en un fichero temporal");
    
      while ( (i=getchar()) != EOF) putc(i,fp);
      fclose(fp);

      /* and remove it from list, since we can never reload from stdin */
    }
    else {
      /******** see if said file exists ********/
    fp = fopen(AuxFileName, "rb");
    if (fp) {                          /* filename file exists */
        fclose(fp);
        }
    }
  }


 HAVE_FILENAME:

  /******* AT THIS POINT 'AuxFileName' is the name of an actual data file
    (no pipes or stdin, though it could be compressed) to be loaded */
  filetype = ReadFileType(AuxFileName);

  if (filetype == RFT_COMPRESS) {   /* a compressed file.  uncompress it */
    char tmpname[128];

    if ( UncompressFile(AuxFileName, tmpname) ) {

      filetype = ReadFileType(tmpname);    /* and try again */
      
      /* if we made a /tmp file (from stdin, etc.) won't need it any more */
      if (strcmp(filename,AuxFileName)!=0) unlink(AuxFileName);

      strcpy(AuxFileName, tmpname);
    }
    else filetype = RFT_ERROR;

    WaitCursor();
  }


  if (filetype == RFT_ERROR) {
    char  foostr[512];
    sprintf(foostr,"No puedo abrir el fichero '%s'\n\n  %s.",filename, ERRSTR(errno));

    //if (!polling) Error(foostr);
    Error(foostr);

    goto FAILED;  /* couldn't get magic#, screw it! */
  }


  if (filetype == RFT_UNKNOWN) {
    /* view as a text/hex file */
    SetISTR(ISTR_INFO,"El fichero '%s' tiene un formato irreconocible.", basefname);
    /* Warning();  */
    goto SHOWN_AS_TEXT;
  }

  if (filetype < RFT_ERROR) {
    SetISTR(ISTR_INFO,"El formato de '%s' no se puede leer.", basefname);
    Warning();
    goto FAILED;
  }


  /****** AT THIS POINT: the filetype is a known, readable format */

  /* kill old page files, if any */
  if (killpage) {
    KillPageFiles(pageBaseName, numPages);
    pageBaseName[0] = '\0';
    numPages = 1;
    curPage = 0;
  }


  SetISTR(ISTR_INFO,"Cargando...");

  i = ReadPicFile(AuxFileName, filetype, &pinfo, 0);

  if (filetype == RFT_XBM && (!i || pinfo.w==0 || pinfo.h==0)) {
    /* probably just a '.h' file or something... */
    SetISTR(ISTR_INFO," ");
    goto SHOWN_AS_TEXT;
  }

  if (!i) {
    SetISTR(ISTR_INFO,"No puedo leer el fichero '%s'.",AuxFileName);
    Warning();
    goto FAILED;
  }

  WaitCursor();

  if (pinfo.w==0 || pinfo.h==0) {  /* shouldn't happen, but let's be safe */
    SetISTR(ISTR_INFO,"Las imgenes de tamao '0x0' no estn permitidas.");
    Warning();
    if (pinfo.pic)     free(pinfo.pic);      pinfo.pic     = (byte *) NULL;
    if (pinfo.comment) free(pinfo.comment);  pinfo.comment = (char *) NULL;
    goto FAILED;
  }


  /**************/
  /* SUCCESS!!! */
  /**************/
    

 GOTIMAGE:
  /* successfully read this picture.  No failures from here on out
     (well, once the pic has been converted if we're locked in a mode) */


  /*
  // Uncomment the following lines if wanted a PIC24 image everytime a image
  // is loaded
  if (pinfo.type!=PIC24 && picType==PIC24) {    // 8 -> 24 bit 
      byte *pic24;
      pic24 = Conv8to24(pinfo.pic, pinfo.w, pinfo.h, 
      pinfo.r, pinfo.g, pinfo.b);
      free(pinfo.pic);
      pinfo.pic  = pic24;
      pinfo.type = PIC24;
  }
  else {    // not locked.  switch 8/24 mode 
    picType = pinfo.type;
  }
  */

  // Set internal format equal to the original file format 
  picType = pinfo.type;

  if (!pinfo.pic) {  // must've failed in the 8-24 or 24-8 conversion
    SetISTR(ISTR_INFO,"Couldn't do %s conversion.",
      (picType==PIC24) ? "8->24" : "24->8");
    if (pinfo.comment) free(pinfo.comment);  pinfo.comment = (char *) NULL;
    Warning();
    goto FAILED;
  }
  // ABSOLUTELY no failures from here on out... 

  if (strlen(pinfo.pagebname)) {
    strcpy(pageBaseName, pinfo.pagebname);
    numPages = pinfo.numpages;
    curPage = 0;
  }

  /* kill off OLD picture, now that we've succesfully loaded a new one */
  KillOldPics();

  /* get info out of the PICINFO struct */
  pic   = pinfo.pic;
  pWIDE = pinfo.w;
  pHIGH = pinfo.h;
  
  SetISTR(ISTR_FORMAT, pinfo.fullInfo);
  strcpy(formatStr, pinfo.shrtInfo);
  picComments = pinfo.comment;
  //ChangeCommentText();

  // copy colormap of image into IOfile
  xvbcopy((char *)pinfo.r, (char *)rMap, 256);
  xvbcopy((char *)pinfo.g, (char *)gMap, 256);      
  xvbcopy((char *)pinfo.b, (char *)bMap, 256);

  /* if we read a /tmp file, delete it.  won't be needing it any more */
  if (filename && strcmp(filename,AuxFileName)!=0) unlink(AuxFileName);

  SetISTR(ISTR_INFO,formatStr);
  
  //SetInfoMode(INF_PART);
  SetISTR(ISTR_FILENAME, 
    (filenum==DFLTPIC || filenum==GRABBED )
    ? "<none>" : basefname);

  SetISTR(ISTR_RES,"%d x %d",pWIDE,pHIGH);
  SetISTR(ISTR_COLOR, "");

  //ActivePrevNext();

  /* handle various 'auto-whatever' command line options
     Note that if 'frompoll' is set, things that have to do with 
     setting the expansion factor are skipped, as we'll want it to
     display in the (already-existing) window at the same size */
    
  tmp = GetISTR(ISTR_COLOR);
  SetISTR(ISTR_INFO,"%s  %s  %s", formatStr,
    (picType==PIC8) ? "8-bit mode." : "24-bit mode.",
    tmp);
  

  WaitCursor();
     
  /* See if image is monocromic and set direct gray / bwdither in case */
  if (picType == PIC8) {
      /* Is the image black & white or greyscaled ? */
#ifdef  WIN32
      /* Use 32 bit uint to speed up checking */
      u_int *ptr = (u_int*) pic;
      for (i=(pWIDE*pHIGH)/4; i && !(*ptr&0xfefefefeL) ; i--,ptr++);
#else
      byte* ptr=pic;
      for (i=pWIDE*pHIGH; i && !(*ptr&0xfe) ; i--,ptr++);
#endif
      if (i) {     /* may be a grayscale or a colormapped image */
        numcols=i=256;
        for (; i ;) {
          i--;
          if ( gMap[i]!=rMap[i] || gMap[i]!=bMap[i])
            break;
        }
        if (i )  /* made it all the way through */
          mono = 0;      /* colormapped */
        else
          mono = 1;      /* grayscale */
      }else {  /* there are only two colors -> b/w */
        mono = 1;
        numcols=2;
      }
    }
    else {  /* PIC24 */
      byte *ip;
      for (i=pWIDE*pHIGH,ip=pic; i && ip[0]==ip[1] && ip[1]==ip[2]; i--,ip+=3);
      if (i)  /* all the way through */
        mono = 0;
      else
        mono = 1;
      numcols = 0;
    }

  if (mono) { // construct a direct grayscale or black & white image
    byte *thepic, *rp, *gp, *bp;
    int colType,pfree,NewPicType = PIC8;
    u_int NewNumcols = numcols;
    
    // Get the image in suitable format 
    if (numcols == 2 ) 
      colType = F_BWDITHER;
    else
      colType = F_GREYSCALE;
    thepic = GetImage(&NewPicType, colType, &pfree, &NewNumcols, &rp, &gp, &bp,
                        MAX_ALLOCABLE_COLS);
  
    // Change pic if a new one has returned 
    if (pfree && thepic) 
       StoreNewPic(thepic,NewPicType,1 /* mono */,pWIDE, pHIGH,NewNumcols,rp,gp,bp);
    else
      numcols=NewNumcols; /* update real number of colors */
       // copy new colormap in IOfile.
      xvbcopy((char *)rp,  (char *)rMap, NewNumcols);
      xvbcopy((char *)gp, (char *)gMap, NewNumcols);      
      xvbcopy((char *)bp, (char *)bMap, NewNumcols);
  }
  

  /* force a redraw of the whole window, as I can't quite trust Config's
     to generate the correct exposes (particularly with 'BitGravity' turned
     on */
  //RedrawPic();
  //MFC does the update. Comented to avoid repeating initialitation
  return 0;
  
 FAILED:
  ArrowCursor();
  KillPageFiles(pinfo.pagebname, pinfo.numpages);

  if (filename && strcmp(filename,AuxFileName)!=0) 
    unlink(AuxFileName);   // kill /tmp file
 
  return -1;


 SHOWN_AS_TEXT:    /* file wasn't in recognized format... */
  ArrowCursor();

  if (strcmp(filename,AuxFileName)!=0) unlink(AuxFileName);   /* kill /tmp file */
  TextView(AuxFileName);
  
  return 1;       /* we've displayed the file 'ok' */
}                                 


/********************************/
int IOfile::ReadFileType(char *fname)
{
  /* opens fname (which *better* be an actual file by this point!) and
     reads the first couple o' bytes.  Figures out what the file's likely
     to be, and returns the appropriate RFT_*** code */


  FILE *fp;
  byte  magicno[30];    /* first 30 bytes of file */
  int   rv, n=0;
  char *suffix;

  if (!fname) return RFT_ERROR;   /* shouldn't happen */

  /* RAW FILES DON'T CONTAIN MAGIC KEYS TO IDENTIFY FILE TYPE,
     so we use extension .RAW to identify them */
  if ( suffix = strrchr(fname, '.') ) {
    if (strlen(suffix)<30) 
      strcpy((char*) magicno,suffix+1);
    else {
      strncpy((char*) magicno,suffix+1,29);
      magicno[30]=0;
    }
    for(n=0; n<3 ;n++)
      magicno[n] = toupper((char)magicno[n]);
    if ( magicno[0]=='R' && magicno[1]=='A' && magicno[2]=='W' && magicno[3]=='\0' )
      return RFT_RAW;
  }

  fp = xv_fopen(fname, "rb");
  if (!fp) return RFT_ERROR;

  n = fread(magicno, (size_t) 1, (size_t) 30, fp);  
  fclose(fp);

  if (n<30) return RFT_UNKNOWN;    /* files less than 30 bytes long... */

  rv = RFT_UNKNOWN;

  if (strncmp((char *) magicno,"GIF87a", (size_t) 6)==0 ||
      strncmp((char *) magicno,"GIF89a", (size_t) 6)==0)        rv = RFT_GIF;

  else if (strncmp((char *) magicno,"VIEW", (size_t) 4)==0 ||
     strncmp((char *) magicno,"WEIV", (size_t) 4)==0)     rv = RFT_PM;

  else if (magicno[0] == 'P' && magicno[1]>='1' && 
     magicno[1]<='6')                             rv = RFT_PBM;

  /* note: have to check XPM before XBM, as first 2 chars are the same */
  else if (strncmp((char *) magicno, "/* XPM */", (size_t) 9)==0) rv = RFT_XPM;

  else if (strncmp((char *) magicno,"#define", (size_t) 7)==0 ||
     (magicno[0] == '/' && magicno[1] == '*'))    rv = RFT_XBM;

  else if (magicno[0]==0x59 && (magicno[1]&0x7f)==0x26 &&
     magicno[2]==0x6a && (magicno[3]&0x7f)==0x15) rv = RFT_SUNRAS;

  else if (magicno[0] == 'B' && magicno[1] == 'M')      rv = RFT_BMP;

  else if (magicno[0]==0x52 && magicno[1]==0xcc)        rv = RFT_UTAHRLE;

  else if ((magicno[0]==0x01 && magicno[1]==0xda) ||
     (magicno[0]==0xda && magicno[1]==0x01))      rv = RFT_IRIS;

  else if (magicno[0]==0x1f && magicno[1]==0x9d)        rv = RFT_COMPRESS;

#ifdef GUNZIP
  else if (magicno[0]==0x1f && magicno[1]==0x8b)        rv = RFT_COMPRESS;
#endif

  else if (magicno[0]==0x0a && magicno[1] <= 5)         rv = RFT_PCX;

  else if (strncmp((char *) magicno,   "FORM", (size_t) 4)==0 && 
     strncmp((char *) magicno+8, "ILBM", (size_t) 4)==0)   rv = RFT_IFF;

  else if (magicno[0]==0 && magicno[1]==0 &&
     magicno[2]==2 && magicno[3]==0 &&
     magicno[4]==0 && magicno[5]==0 &&
     magicno[6]==0 && magicno[7]==0)              rv = RFT_TARGA;

  else if (magicno[4]==0x00 && magicno[5]==0x00 &&
     magicno[6]==0x00 && magicno[7]==0x07)        rv = RFT_XWD;

  else if (strncmp((char *) magicno,"SIMPLE  ", (size_t) 8)==0 && 
     magicno[29] == 'T')                          rv = RFT_FITS;
  

#ifdef HAVE_JPEG
  else if (magicno[0]==0xff && magicno[1]==0xd8 && 
     magicno[2]==0xff)                            rv = RFT_JFIF;
#endif

#ifdef HAVE_TIFF
  else if ((magicno[0]=='M' && magicno[1]=='M') ||
     (magicno[0]=='I' && magicno[1]=='I'))        rv = RFT_TIFF;
#endif

#ifdef HAVE_PDS
  else if (strncmp((char *) magicno,  "NJPL1I00", (size_t) 8)==0 ||
     strncmp((char *) magicno+2,"NJPL1I",   (size_t) 6)==0 ||
           strncmp((char *) magicno,  "CCSD3ZF",  (size_t) 7)==0 ||
     strncmp((char *) magicno+2,"CCSD3Z",   (size_t) 6)==0 ||
     strncmp((char *) magicno,  "LBLSIZE=", (size_t) 8)==0)
      rv = RFT_PDSVICAR;
#endif

#ifdef GS_PATH
  else if (strncmp((char *) magicno, "%!",     (size_t) 2)==0 ||
     strncmp((char *) magicno, "\004%!", (size_t) 3)==0)   rv = RFT_PS;
#endif
  else if (strncmp((char *) magicno, "IM",     (size_t) 2)==0 )
     rv = RFT_IMG;

  return rv;
}

/********************************/
//------------------------------------------------------------------------------
// Set IOfile with the image in the file fname
//------------------------------------------------------------------------------
int IOfile::ReadPicFile(char *fname, int ftype, PICINFO *pinfo, int quick )     
{
  /* if quick is set, we're being called to generate icons, or something
     like that.  We should load the image as quickly as possible.  Currently,
     this only affects the LoadPS routine, which, if quick is set, only
     generates the page file for the first page of the document */

  int rv = 0;

  /* by default, most formats aren't multi-page */
  pinfo->numpages = 1;
  pinfo->pagebname[0] = '\0';

  switch (ftype) {
  case RFT_GIF:     rv = LoadGIF   (fname, pinfo);         break;
  case RFT_PM:      rv = LoadPM    (fname, pinfo);         break;
  case RFT_PBM:     rv = LoadPBM   (fname, pinfo);         break;
  case RFT_XBM:     rv = LoadXBM   (fname, pinfo);         break;
  case RFT_SUNRAS:  rv = LoadSunRas(fname, pinfo);         break;
  case RFT_BMP:     rv = LoadBMP   (fname, pinfo);         break;
  case RFT_UTAHRLE: rv = LoadRLE   (fname, pinfo);         break;
  case RFT_IRIS:    rv = LoadIRIS  (fname, pinfo);         break;
  case RFT_PCX:     rv = LoadPCX   (fname, pinfo);         break;
  case RFT_IFF:     rv = LoadIFF   (fname, pinfo);         break;
  case RFT_TARGA:   rv = LoadTarga (fname, pinfo);         break;
  case RFT_XPM:     rv = LoadXPM   (fname, pinfo);         break;
  case RFT_XWD:     rv = LoadXWD   (fname, pinfo);         break;
  case RFT_FITS:    rv = LoadFITS  (fname, pinfo, quick);  break;
  case RFT_RAW:     rv = LoadRAW   (fname, pinfo);         break;

#ifdef HAVE_JPEG
  case RFT_JFIF:    rv = LoadJFIF  (fname, pinfo, quick);  break;
#endif

#ifdef HAVE_TIFF
  case RFT_TIFF:    rv = LoadTIFF  (fname, pinfo);         break;
#endif

#ifdef HAVE_PDS
  case RFT_PDSVICAR: rv = LoadPDS  (fname, pinfo);         break;
#endif

#ifdef GS_PATH
  case RFT_PS:      rv = LoadPS    (fname, pinfo, quick);   break;
#endif
  case RFT_IMG:     rv = LoadIMG   (fname, pinfo);          break;

  }
  return rv;
}

/***************************************************/
char *IOfile::GetFileName( int bOpenFileDialog )
// bOpenFileDialog: TRUE for FileOpen, FALSE for FileSaveAs
{
  /* pide el nombre de un fichero */
  static char aux_str[MAXPATHLEN];
  InputDialog("Introduce el nombre del fichero:\n", aux_str,
                            MAXPATHLEN, "", 0);
  return aux_str;
}

/***************************************************/
