/*********************************************************************************\
*     File:                  R A W  . C P P                                       *
*                                                                                 *
*     i/o routines for .RAW files (plain greyscaled imaged with 256 grey levels)  *
*     Standard   C++                                                              *
*                                                                                 *
*     IOfile::LoadRAW(fname, numcols)                                          *
*     IOfile::WriteRAW(fp, pic);                                               *
*                                                                                 *
\*********************************************************************************/

#include <math.h>

#include "IOfile.h"

#define BLOCK_SIZE 4096
#define MAX(a,b)   (((a)<(b))?(b):(a))
#define MIN(a,b)   (((a)<(b))?(a):(b))

/* Comments on error handling:
   a truncated file is not considered a Major Error.  The file is loaded, the
   rest of the pic is filled with 0's.
   not being able to malloc is a Fatal Error.  The program is aborted. */


int IOfile::LoadRAW (char* fname, PICINFO * pinfo)
{
  /* returns 0 if error, 1 if success */
  FILE *fp;
  int filesize,Width,Height,i;
  byte *picGrey;
  char buf2[128], buf[64], *bname;
  static char *labels[] = { "Ok" }; /* Button for user input    */
  
  
  bname = BaseName(fname);
  fp = fopen(fname, "rb");
  if (!fp) {
    sprintf(buf2,"LoadRAW: no se puede abrir el archivo %s",bname);
    Error(buf2);
    return 1;
  }
  
  if (DEBUG)
    printf("LoadXPM(): Loading RAW from %s\n", fname);

  /* COMPUTE LENGTH */
  fseek(fp, 0L, SEEK_END);
  filesize = ftell(fp);
  Width = (int) sqrt((double) filesize);
  Height = filesize/Width;
  fseek(fp, 0L, SEEK_SET);
  
  sprintf(buf,"%d %d", Width, Height);
  sprintf(buf2,"Introduce la dimensin anchura x altura de la imagen"
          " contenida en %s (%d bytes)",
          bname,filesize);
  InputDialog(buf2,buf, 64, "0123456789 ", 1);
  
  /* Check for right values                                      */
  if (sscanf(buf,"%d %d",&Width,&Height) != 2) {
    Error("El formato del parametro no es correcto");
    return 0;
  }
  
  
  picGrey = (byte *) calloc((size_t) Width * Height, (size_t) 1);
  if (!picGrey) {
    Error("No puedo alojar memoria en LoadRAW");
    return 0;
  }

   /* Read whole file once */
  fread( (void *) picGrey, 1,(size_t) MAX((Width * Height),filesize), fp);
 
  /* Cierra el fichero de lectura y vuelve */
  fclose( fp );


  for (i=0; i<256; i++)
      pinfo->b[i] = pinfo->g[i] = pinfo->r[i] = (byte) i;
  pinfo->pic  = picGrey;
  pinfo->type = PIC8;
  pinfo->comment = (char *) NULL;
  pinfo->w = Width;  pinfo->h = Height;
  pinfo->normw = pinfo->w;   pinfo->normh = pinfo->h;
  pinfo->frmType = F_RAW;
  pinfo->colType = F_GREYSCALE;

  sprintf(pinfo->fullInfo, "%dx%d RAW greyscale image,"
    " 8 bits per pixel (%ld bytes).",
    pinfo->w,pinfo->w, filesize);

  sprintf(pinfo->shrtInfo, "%dx%d RAW.", pinfo->w, pinfo->h);
  pinfo->comment = (char *) NULL;

  mono=TRUE; /* RAW images are allways greyscalled */
  numcols=256;

  return 1;
}


/*******************************************/

int IOfile::WriteRAW(FILE *fp, byte *pic824, int ptype, int w, int h,
                        byte *rmap, byte *gmap,byte *bmap )
{
  int i;
  byte *graypic=NULL, *sp, *dp;
  
  if (ptype == PIC24 ) {
    /* generate a faked 8-bit per pixel image with a grayscale cmap,
       so that it can just fall through existing 8-bit code */

    graypic = (byte *) malloc((size_t) w*h);
    if (!graypic) FatalError("unable to malloc in WriteRAW()");

    for (i=0,sp=pic824,dp=graypic; i<w*h; i++,sp+=3, dp++) {
      *dp = (byte) MONO(sp[0],sp[1],sp[2]);
    }

    pic824 = graypic;
  }
  else if ( rmap && gmap && bmap ) {  // PIC8 and colormap

    /* check direct GRAYSCALE */
    for (i=255; i && rMap [i]==i && gMap [i]==i && bMap [i] == i; i--);
      /* if i!=0, or color[0] != black, the colormap is not lineal */
    
    if ( i || rMap[0] || gMap[0] || bMap[0]) {
    /*ptype == PIC8 and not direct GREYSCALE*/
    /* create a Greyscale direct mapped image*/

      byte graytable[256];

      graypic = (byte *) malloc((size_t) w*h);
      if (!graypic) FatalError("unable to malloc in WriteRAW()");
    
      for (i=256; i; ) {
        i--;
        graytable[i] = MONO(rmap[i],gmap[i],bmap[i]);
      }

      for (i=w*h; i; ) {
        i--;
        graypic[i] = graytable[pic824[i]];
      }
    
      pic824 = graypic;
    }
  }

  // VOLCADO DE LA IMAGEN EN TONOS DE GRISES A DISCO.
  {
     int ngrabados, nbytes, tamano = w*h;

   /* Abrir fichero solo ESCRITURA (ya abierto antes de entrar en 
    * la funcin)
   if ((input_file = fopen( nombre,"wb" )) == (FILE*) -1) {
     return -1;
   } */

   /* Escribe el fichero */
    do {
       nbytes = MIN(tamano,BLOCK_SIZE);   
       ngrabados = fwrite((void *) pic824, 1 , nbytes ,fp);
       // check number of bytes written 
       if ( ngrabados != nbytes ) break;
       /* update pointer to next value */
       pic824 += ngrabados;
       tamano -= ngrabados;
    } while (tamano>0);
 
    if (tamano) Warning
      ("La escritura del fichero en formato RAW no se ha completado con"
       " xito debido a un fallo en fwrite");

    if (graypic) free(graypic);

    /* La funcin que llama a WriteRaw se encarga de cerrar
     * el fichero de lectura */
    return tamano;
  }
}
