/*=====================================================================*/
/*                 RUTINAS DE VISION POR ORDENADOR CHANI32.C           */
/*                      POR MIGUEL ANGEL MARTINEZ ESTEBAN              */
/*=====================================================================*/

/***********************************************************************/
/* Modificado por Joaqun Mur ( abril 1997) jmur@robot3.cps.unizar.es  */
/***********************************************************************/

#define IMPLEMENTATION_MIGUELAN
#include "miguelan.h"

/*=====================================================================*/
/*                      Cuerpo del paquete CHANI32.C                   */
/*=====================================================================*/

const float factor_escala[] = { (float) 1, (float) 0.94088, (float) 1.0289,
                                           (float) 0.9780, (float) 0.9621,(float) 0.9621 };

float factor_gaus(int escala)
{
 /* Usando el valor del area de la gausiana */
 /* Corregido con factores para que sobrepase el escalon en cada escala. */
 return (float) (factor_escala[escala]*sqrt(2.0*M_PI)*0.3*(1<<escala));
}

void suavizado_gausiana_32(unsigned char far *picGrey,
			   float far *gausiana,
			   int columnas, int filas,float sigma)
{
  /* Blur grey imagen with gaussian weight: f=exp(-xx/2ss) */  
  float far *aux;      /* Auxiliar image for separating components */
  int indice1,indice2,ancho,ancho_max;
  register int posicion1,posicion2;
  float valor_1,factor;
  float gaus[MAX_SIGMA];

  /* Allocate memory */
  aux = (float*) malloc( (long)filas*(long)columnas*sizeof(float) );
  if (!aux) {
    TRACE("Fallo al requerir memoria");
    return;
  }

  /* width of gaussian window */
  ancho_max=(int)(((float) 2)*sigma)+1;
  ancho_max=MIN(ancho_max,MAX_SIGMA);
  
  /* gaussian weights (only half window because of simetry) */
  gaus[0]=(float) 1;
  for ( indice1=1 ; indice1<ancho_max ; indice1++ )
     gaus[indice1]= (float) exp(-indice1*indice1/(2*sigma*sigma));

  /* Scale factor */
  factor=gaus[0];
  for ( ancho=1 ; ancho<ancho_max ; ancho++)
       factor += 2*gaus[ancho];
  factor *= factor; //factor^2 => saves fil*col multipic

  /* adjust ancho_max to ancho_max-1 to fit array index limit */
  ancho_max--;

  /* BLUR X (calculated with a "mirror image" beyond the border )*/
  for ( indice1=filas ; indice1 ;  ) {
     indice1--;
     for ( indice2=columnas ; indice2 ;  ) {
       indice2--;
       valor_1=picGrey[indice2+indice1*columnas];
       for ( ancho=ancho_max ; ancho ; ancho--) {	
         /* Don't excess limits! */
         posicion1=indice2+ancho;
         posicion1= (posicion1<0)         ? (-posicion1) :
                    (posicion1>=columnas) ? (posicion1-columnas) :
                                            posicion1;
		     posicion2=indice2-ancho;
         posicion2= (posicion2<0)         ? (-posicion2) :
                    (posicion2>=columnas) ? (posicion2-columnas) :
                                            posicion2;
         valor_1 += ( picGrey[posicion1+indice1*columnas] +
			          picGrey[posicion2+indice1*columnas] ) * gaus[ancho];
       }
       aux[indice2 + indice1*columnas]=valor_1;
     }
  }
  /* BLUR Y (calculated with a "mirror image" beyond the border )*/
  for ( indice1=columnas ; indice1 ; ) {
     indice1--;
     for ( indice2=filas ; indice2 ; ) {
       indice2--;
       valor_1=aux[indice1+indice2*columnas];
       for ( ancho=ancho_max ; ancho ; ancho--) {
         /* Don't excess limits! */
         posicion1 =indice2+ancho;
         posicion1 = (posicion1<0)         ? (-posicion1) :
                     (posicion1>=filas)    ? (posicion1-filas) :
                                              posicion1;
         posicion2 =indice2-ancho;
         posicion2 = (posicion2<0)         ? (-posicion2) :
                     (posicion2>=filas)    ? (posicion2-filas) :
                                              posicion2;
         valor_1 += (aux[indice1+posicion1*columnas] +
					 aux[indice1+posicion2*columnas] ) * gaus[ancho];
       }
       gausiana [indice1 + indice2*columnas]=valor_1/factor;
     }
  }
  
  /* FREE MEMORY */
  farfree(aux);
}
/*********************************************************************************/

