/*=====================================================================*/
/*                  RUTINAS DE VISION POR ORDENADOR VISIO232.H         */
/*                      POR MIGUEL ANGEL MARTINEZ ESTEBAN              */
/*=====================================================================*/

/***********************************************************************/
/* Modificado por Joaqun Mur ( abril 1997) jmur@robot3.cps.unizar.es  */
/***********************************************************************/

#include <memory.h>

#define IMPLEMENTATION_MIGUELAN
#include "miguelan.h"


/*=====================================================================*/
/*                      RUTINAS DE AYUDA Y CONVERSION                  */
/*=====================================================================*/


float valor_max_gradiente_32(float far *pt_imagen,int ncol,int nfil)

/* Calcula el valor maximo del modulo del gradiente de una imagen  */
/* y lo devuelve en la variable entera. Se usa para escalar luego  */
/* los sucesivos gradientes al valor m ximo para la representacion.*/
{
 int x,y;
 float gx=(float) 0.0;                         /* Almacena el valor pasada a x */
 float gy=(float) 0.0;                         /* Almacena el valor pasada a y */
 double auxiliar;
 float max_gradxy=(float) 0.0;

 for (y=1;y<(ncol-1);y++)
   for (x=1;x<(nfil-1);x++)
     {
      gx=(*(pt_imagen+(x+1)+y*nfil)-
	        *(pt_imagen+x+(y)*nfil));
      gy=(*(pt_imagen+x+(y+1)*nfil)-
	        *(pt_imagen+x+(y)*nfil));

      auxiliar=(double)gx*(double)gx+(double)gy*(double)gy;
      auxiliar=sqrt(auxiliar);               /* Calculo el G(x,y)           */
      if ( max_gradxy < (float)auxiliar )
	      max_gradxy=(float)auxiliar;
     }
 return max_gradxy;
}

/*********************************************************************************/

void gradiente_32_exp(float far *pt_imagen,
		  float *pt_gradientexy,
		  unsigned char far *pt_gradientetita,
		  float escala,char tipo,
		  int ncol,int nfil, int modo)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* dejando el gradiente de la misma en la direccion pasada pt_gradiende */
/* con numero columnas ncol y numero de filas nfil de la imagen         */
/* La escala es la obtenida con el procedimiento valos_max_gradiente    */
/* Los angulos los deja escalados de (0..256) equivale a (0..360).      */
/* NOTA: la imagen original y el resultado de la opearacin pueden ser  */
/* el mismo, es decir, si pasamos en pt_imagen y pt_gradientexy el mismo*/
/* puntero, el algoritmo calcular de forma correcta el gradiente, sobre*/
/* escribiendo la imagen original.                                      */
{
 float gx;                                  /* Almacena el valor pasada a x */
 float gy;                                  /* Almacena el valor pasada a y */
 float gradxy;                              /* Valor del gradiente    */
 double auxiliar;
 int x,y;
 float media,abs_gx, abs_gy;


  for (y=0;y<nfil-1;y++) {
    for (x=0;x<ncol-1;x++) {    
      /* Calculo normal de las componentes del gradiente (los bordes se calculan
         fuera de este bucle )                                                   */        
      gx= pt_imagen[(x+1)+(y+1)*ncol] - pt_imagen[x+y*ncol];
      /* Cambiamos el signo del gy para mantener el signo de los argumentos */
      gy= pt_imagen[x+(y+1)*ncol]-pt_imagen[(x+1)+y*ncol] ;

      /* Necesitamos el valor absoluto */
      abs_gx=(float) fabs(gx); abs_gy=(float) fabs(gy);
      
      if (tipo=='A') {
      /* Calculo del modulo del gradiente A : sqrt((gx*gx+gy*gy)/2)         */
        auxiliar=((double)gx*(double)gx+(double)gy*(double)gy)/2.0;
        gradxy=(float)sqrt(auxiliar)*escala;  /* Calculo el G(x,y)*/
      
      } else if (tipo=='B') {
      /* Calculo del modulo del gradiente B : Max {|gx|,|gy|}               */
        gradxy= escala * MAX (abs_gx,abs_gy);
      
      } else if (tipo=='C') {
      /* Calculo del modulo del gradiente ponderacion del tipo C          */
        media=(abs_gx + abs_gy);    /* media' = media*2 */
        /* gradxy=sqrt(0.2*media*media+1.8*gy*gy)*escala/sqrt(2.0); */
        gradxy=(float) sqrt(  0.9*((abs_gx>abs_gy) ? (gx*gx) : (gy*gy))
                    + 0.025*media*media)*escala;
      
      } else if (tipo=='D') {
      /* Calculo del modulo del gradiente ponderacion del tipo D          */
        media=(abs_gx + abs_gy);    /* media' = media*2 */
        /* gradxy=(0.8*abs(gx)+1.2*media)*escala/2; */
        gradxy=(float) ( 0.3*media+ 0.4*MAX(abs_gx,abs_gy))*escala;
     }

/* UMBRALIZO EL GRADIENTE */
      if (modo)            /* MODO != 0 -> UMBRALIZA GRADIENTE */
	      if ( (gradxy>UMBRAL_GRADD) && (gradxy<UMBRAL_GRADU) )
	          gradxy=(float)COLOR_GRAD_UMBRAL;
	      else 
            gradxy=(float) 0.0;

      pt_gradientexy[x+y*ncol]=gradxy;        /* Almaceno el valor de G(x,y) */

/* CALCULO DEL GRADIENTE DE ANGULOS SOLO SI GRADIENTE <> 0 */
/* Coloco el valor en 8 bits 256 valores escalados (0..256)=(0..360)*/
      if ( (gradxy > EPSILON ) && (abs_gx > EPSILON) ) {
          /* CALCULO NORMAL DEL ARGUMENTO 
	           angulo = arctang - 45
             pasar radianes [0..2*PI] a [0..256]  (factor escala)
             angulo no puede ser mayor de 255 (slo primera vuelta )
               -> quitar vueltas de 256 (se realiza automaticamente
                   en la conversin a unsigned char */
	        pt_gradientetita[x+y*ncol]= 
              (unsigned char) ROUND(( atan2( gy, gx ) + M_PI/4 )*256.0/( 2*M_PI));
	     }
       else  /* EL ANGULO ES 0 (por defecto), 45 o 225 GRADOS 
                (ventana de gradiente cruzado diagonal) */
         pt_gradientetita[x+y*ncol] = (gy > EPSILON) ?  96 :     /* 135 */
                                      (gy < EPSILON) ? 224 :     /* 315 */ 
                                                         0 ;     /* gx==gy==0 */       
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE DERECHO (x=nfil-1) */
    /* Si calculara el espejo de la imagen en los bordes concluiriamos que: */
    /* imagen reflejada en los bordes: */
    pt_gradientexy[(y+1)*ncol-1]= (float) /* Almaceno el valor de G(nfil,y) */
         fabs(pt_imagen[(y+1)*ncol-1] - pt_imagen[(y+2)*ncol-1])*escala;    
    /* EL ANGULO ES 0  */
    pt_gradientetita[(y+1)*ncol-1] = 0;    /*  0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE INFERIOR (y=ncol-1) */
 for (x=0;x<ncol-1;x++) {
   /* Si calculara el espejo de la imagen en los bordes concluiriamos que: */
    pt_gradientexy[x+(nfil-1)*ncol]= (float) /* Almaceno el valor de G(x,ncol) */   
         fabs(pt_imagen[x+(nfil-1)*ncol] - pt_imagen[x+1+(nfil-1)*ncol])*escala;    
    /* EL ANGULO ES 0  */
    pt_gradientetita[x+(nfil-1)*ncol] = 0;    /*  0 por defecto*/
 }
 
 /* Almaceno el valor de G(nfil,ncol)=0 */
 pt_gradientexy[ncol*nfil-1]=(float) 0.0;
 pt_gradientetita[ncol*nfil-1]=0;

}     /* fin del prodedimiento gradiente_32_exp */

/*********************************************************************************/

void gradiente_32_component(float far *pt_imagen,
		  float *pt_gradientexy,
		  unsigned char far *pt_gradientetita,
		  float escala,char tipo,
		  int ncol,int nfil, int modo)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* dejando el gradiente de la misma en la direccion pasada pt_gradiende */
/* con numero columnas ncol y numero de filas nfil de la imagen         */
/* La escala es la obtenida con el procedimiento valos_max_gradiente    */
/* Los angulos los deja escalados de (0..256) equivale a (0..360).      */
/* NOTA: la imagen original y el resultado de la opearacin pueden ser  */
/* el mismo, es decir, si pasamos en pt_imagen y pt_gradientexy el mismo*/
/* puntero, el algoritmo calcular de forma correcta el gradiente, sobre*/
/* escribiendo la imagen original.                                      */
{
 float gx;                                  /* Almacena el valor pasada a x */
 float gy;                                  /* Almacena el valor pasada a y */
 float gradxy;                              /* Valor del gradiente    */
 double auxiliar;
 int x,y;
 float media,abs_gx, abs_gy;
 float old_gradxy;

  for (y=0;y<nfil-1;y++) {
    for (x=0;x<ncol-1;x++) {    
      /* Calculo normal de las componentes del gradiente (los bordes se calculan
         fuera de este bucle )                                                   */        
      gx= pt_imagen[(x+1)+(y+1)*ncol] - pt_imagen[x+y*ncol];
      /* Cambiamos el signo del gy para mantener el signo de los argumentos */
      gy= pt_imagen[x+(y+1)*ncol]-pt_imagen[(x+1)+y*ncol] ;

      /* Necesitamos el valor absoluto */
      abs_gx=(float) fabs(gx); abs_gy=(float) fabs(gy);
      
      if (tipo=='A') {
      /* Calculo del modulo del gradiente A : sqrt((gx*gx+gy*gy)/2)         */
        auxiliar=((double)gx*(double)gx+(double)gy*(double)gy)/2.0;
        gradxy=(float)sqrt(auxiliar)*escala;  /* Calculo el G(x,y)*/
      
      } else if (tipo=='B') {
      /* Calculo del modulo del gradiente B : Max {|gx|,|gy|}               */
        gradxy= escala * MAX (abs_gx,abs_gy);
      
      } else if (tipo=='C') {
      /* Calculo del modulo del gradiente C ponderacion del tipo A          */
        media=(abs_gx + abs_gy);    /* media' = media*2 */
        /* gradxy=sqrt(0.2*media*media+1.8*gy*gy)*escala/sqrt(2.0); */
        gradxy=(float) sqrt(  0.9*((abs_gx>abs_gy) ? (gx*gx) : (gy*gy))
                    + 0.025*media*media)*escala;
      
      } else if (tipo=='D') {
      /* Calculo del modulo del gradiente D ponderacion del tipo A          */
        media=(abs_gx + abs_gy);    /* media' = media*2 */
        /* gradxy=(0.8*abs(gx)+1.2*abs(media))*escala/2.2; */
        gradxy=(float) ( 0.5454*media+ 0.3636*MAX(abs_gx,abs_gy))*escala;
     }

/* UMBRALIZO EL GRADIENTE */
      if (modo)            /* MODO != 0 -> UMBRALIZA GRADIENTE */
	      if ( (gradxy>UMBRAL_GRADD) && (gradxy<UMBRAL_GRADU) )
	          gradxy=(float)COLOR_GRAD_UMBRAL;
	      else 
            gradxy=(float)0.0;

      old_gradxy = pt_gradientexy[x+y*ncol];
      pt_gradientexy[x+y*ncol] += gradxy;  /* Almaceno el valor de G(x,y) */

/* CALCULO DEL GRADIENTE DE ANGULOS SOLO SI GRADIENTE <> 0 */
/* Coloco el valor en 8 bits 256 valores escalados (0..256)=(0..360)*/
      if ( gradxy > old_gradxy ) {
        if ( (gradxy > EPSILON ) && (abs_gx > EPSILON) ) {
            /* CALCULO NORMAL DEL ARGUMENTO 
	             angulo = arctang - 45
               pasar radianes [0..2*PI] a [0..256]  (factor escala)
               angulo no puede ser mayor de 255 (slo primera vuelta )
                 -> quitar vueltas de 256 (se realiza automaticamente
                     en la conversin a unsigned char */
	          pt_gradientetita[x+y*ncol]= 
                (unsigned char) ROUND(( atan2( gy, gx ) + M_PI/4 )*256.0/( 2*M_PI));
	       }
         else  /* EL ANGULO ES 0 (por defecto), 45 o 225 GRADOS 
                  (ventana de gradiente cruzado diagonal) */
           pt_gradientetita[x+y*ncol] = (gy > EPSILON) ?  96 :     /* 135 */
                                        (gy < EPSILON) ? 224 :     /* 315 */ 
                                                           0 ;     /* gx==gy==0 */       
      } //solo modifico tita si tenemos un modulo grande en esta pasada
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE DERECHO (x=nfil-1) */
    /* Si calculara el espejo de la imagen en los bordes concluiriamos que: */
    /* imagen reflejada en los bordes: */
    pt_gradientexy[(y+1)*ncol-1] +=  (float)/* Almaceno el valor de G(nfil,y) */
         fabs(pt_imagen[(y+1)*ncol-1] - pt_imagen[(y+2)*ncol-1])*escala;    
    /* EL ANGULO ES 0  */
    /*pt_gradientetita[(y+1)*ncol-1] = 0;      0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE INFERIOR (y=ncol-1) */
 for (x=0;x<ncol-1;x++) {
   /* Si calculara el espejo de la imagen en los bordes concluiriamos que: */
    pt_gradientexy[x+(nfil-1)*ncol] += (float) /* Almaceno el valor de G(x,ncol) */   
         fabs(pt_imagen[x+(nfil-1)*ncol] - pt_imagen[x+1+(nfil-1)*ncol])*escala;    
    /* EL ANGULO ES 0  */
    /*pt_gradientetita[x+(nfil-1)*ncol] = 0;      0 por defecto*/
 }
 
 /* Almaceno el valor de G(nfil,ncol)=0 
 pt_gradientexy[ncol*nfil-1]=0.0;
 pt_gradientetita[ncol*nfil-1]=0.0;     */

}     /* fin del prodedimiento gradiente_32_component */

/*********************************************************************************/

void doGradientComponent(byte *pt_imagen,
		  float *pt_gradientexy,
		  unsigned char *pt_gradientetita,
		  int ncol,int nfil, float factor)

/* Realiza el gradiente de la imagen pasada como puntero pt_imagen      */
/* dejando el gradiente de la misma en la direccion pasada pt_gradiende */
/* con numero columnas ncol y numero de filas nfil de la imagen         */
/* Los angulos los deja escalados de (0..256) equivale a (0..360).      */
/* NOTA: la imagen original y el resultado de la opearacin pueden ser  */
/* el mismo, es decir, si pasamos en pt_imagen y pt_gradientexy el mismo*/
/* puntero, el algoritmo calcular de forma correcta el gradiente, sobre*/
/* escribiendo la imagen original.                                      */

/* VERSIN MODIFICADA PARA TRATAR VARIAS COMPONENTES DEL COLOR          */
/* INICIALIZAR ANTES DE LLAMAR POR PRIMERA VEZ EL ALGORITMO             */
/* *pt_gradientexy (se va rellenando sucesivamente)                     */

{
 float gradxy;                              /* Valor del gradiente    */
 int x,y,gx,gy,abs_gx, abs_gy;
 float old_gradxy;


  for (y=0;y<nfil-1;y++) {
    for (x=0;x<ncol-1;x++) {    
      /* Calculo normal de las componentes del gradiente (los bordes se calculan
         fuera de este bucle )                                                   */        
      gx= pt_imagen[(x+1)+(y+1)*ncol] - pt_imagen[x+y*ncol];
      /* Cambiamos el signo del gy para mantener el signo de los argumentos */
      gy= pt_imagen[x+(y+1)*ncol]-pt_imagen[(x+1)+y*ncol] ;

      /* Necesitamos el valor absoluto */
      abs_gx= abs(gx); abs_gy= abs(gy);
      
      /* Calculo del modulo del gradiente A : sqrt((gx*gx+gy*gy)/2)         */
      gradxy=((float)sqrt((gx*gx+gy*gy)/2.0))*factor;
      /* Calculo el G(x,y)*/
      old_gradxy = pt_gradientexy[x+y*ncol];
      pt_gradientexy[x+y*ncol]+=gradxy;        /* Almaceno el valor de G(x,y) */

/* CALCULO DEL GRADIENTE DE ANGULOS SOLO SI GRADIENTE <> 0 */
/* Coloco el valor en 8 bits 256 valores escalados (0..256)=(0..360)*/
   if (gradxy > old_gradxy )
      if (abs_gx > EPSILON) {
          /* CALCULO NORMAL DEL ARGUMENTO 
	           angulo = arctang - 45
             pasar radianes [0..2*PI] a [0..256]  (factor escala)
             angulo no puede ser mayor de 255 (slo primera vuelta )
               -> quitar vueltas de 256 (se realiza automaticamente
                   en la conversin a unsigned char */
	        pt_gradientetita[x+y*ncol]= 
              (unsigned char) ROUND(( atan2( gy, gx ) + M_PI/4 )*256.0/( 2*M_PI));
	     }
       else  /* EL ANGULO ES 0 (por defecto), 45 o 225 GRADOS 
                (ventana de gradiente cruzado diagonal) */
         pt_gradientetita[x+y*ncol] = (gy > EPSILON) ?  96 :     /* 135 */
                                      (gy < EPSILON) ? 224 :     /* 315 */ 
                                                         0 ;     /* gx==gy==0 */       
    } /* fin del bucle for ( fin tratamiento de una fila) */

    /* BORDE DERECHO (x=nfil-1) */
    /* Si calculara el espejo de la imagen en los bordes concluiriamos que: */
    /* imagen reflejada en los bordes: */
    pt_gradientexy[(y+1)*ncol-1] += /* Almaceno el valor de G(nfil,y) */
         abs(pt_imagen[(y+1)*ncol-1] - pt_imagen[(y+2)*ncol-1]);    
    /* EL ANGULO ES 0  */
    //pt_gradientetita[(y+1)*ncol-1] = 0;    /*  0 por defecto*/
 }  /* fin del fucle for (fin tratamiento de columnas )*/

 /* BORDE INFERIOR (y=ncol-1) */
 for (x=0;x<ncol-1;x++) {
   /* Si calculara el espejo de la imagen en los bordes concluiriamos que: */
    pt_gradientexy[x+(nfil-1)*ncol] += /* Almaceno el valor de G(x,ncol) */   
         abs(pt_imagen[x+(nfil-1)*ncol] - pt_imagen[x+1+(nfil-1)*ncol]);    
    /* EL ANGULO ES 0  */
    //pt_gradientetita[x+(nfil-1)*ncol] = 0;    /*  0 por defecto*/
 }
 
 /* Almaceno el valor de G(nfil,ncol)=0 */
 //pt_gradientexy[ncol*nfil-1]=0.0;
 //pt_gradientetita[ncol*nfil-1]=0.0;

}     /* fin del prodedimiento doGradientComponent */


/*********************************************************************************/

