// dibapi.h
//

/* Proyecto Fin de Carrera
 * Departamento de Informatica e Ingeniera de Sistemas
 * Universidad de Zaragora
 *
 * rea:      Robtica y Automatizacin Avanzada.
 * Tema:      Visin por Computador.
 * Ttulo:    Diseo de un entorno de visin para 
 *            el tratamiento de imgenes.
 * Director:  Dr. Carlos Sags Blzquiz.
 * Autor:     Joaqun Mur Amada.
 * Promocin: 1991-1996.
*/

#ifndef _INC_DIBAPI
#define _INC_DIBAPI

#include "stdafx.h"

/* Handle to a DIB -Device Independent Bitmap- */
DECLARE_HANDLE(HDIB);

/* Handle to IMAGE, the object that contains the image data */
#define HIMAGE HDIB

/* Handle to a DDB -Device Dependent Bitmap-, a buffer to speed up display reviews */
//#define HDDB HBITMAP

/* DIB constants */
#define PALVERSION   0x300

/* DIB Macros*/

#define IS_WIN30_DIB(lpbi)  ((*(LPDWORD)(lpbi)) == sizeof(BITMAPINFOHEADER))
#define RECTWIDTH(lpRect)     ((lpRect)->right - (lpRect)->left)
#define RECTHEIGHT(lpRect)    ((lpRect)->bottom - (lpRect)->top)

// WIDTHBYTES performs DWORD-aligning of DIB scanlines.  The "bits"
// parameter is the bit count for the scanline (biWidth * biBitCount),
// and this macro returns the number of DWORD-aligned bytes needed
// to hold those bits.

#define WIDTHBYTES(bits)    (((bits) + 31) / 32 * 4)

static HICON hcurSave;
/* Macros to display/remove hourglass cursor for lengthy operations */
#define StartWait() hcurSave = SetCursor(LoadCursor(NULL,IDC_WAIT))
#define EndWait()   SetCursor(hcurSave)

/* Header signatutes for various resources */
#define BFT_ICON   0x4349   /* 'IC' */
#define BFT_BITMAP 0x4d42   /* 'BM' */
#define BFT_CURSOR 0x5450   /* 'PT' */

/* macro to determine if resource is a DIB */
#define ISDIB(bft) ((bft) == BFT_BITMAP)

/* Macro to align given value to the closest DWORD (unsigned long ) */
#define ALIGNULONG(i)   ((i+3)/4*4)

#define PALVERSION      0x300
#define MAXPALETTE      256       /* max. # supported palette entries */

/* Function prototypes */

LPSTR     WINAPI  FindDIBBits (LPSTR lpbi);
DWORD     WINAPI  DIBWidth (LPSTR lpDIB);
DWORD     WINAPI  DIBHeight (LPSTR lpDIB);
WORD      WINAPI  PaletteSize ( LPBITMAPINFOHEADER lpbi);
WORD      WINAPI  DIBNumColors ( LPBITMAPINFOHEADER lpbi);

HGLOBAL   WINAPI  CopyHandle (HGLOBAL h);
HPALETTE  WINAPI  CopyPalette (HPALETTE hpal);
HBITMAP   WINAPI  CopyBitmap (HBITMAP hbm);
HBITMAP   WINAPI  CropBitmap (HBITMAP hbm, PRECT prc);

HANDLE    WINAPI  DibFromBitmap ( HBITMAP hbm, DWORD biStyle, WORD biBits, HPALETTE hpal);
HBITMAP   WINAPI  BitmapFromDib ( HANDLE hdib, HPALETTE hpal);

BOOL      WINAPI  PaintDIB (HDC, LPRECT, HDIB, LPRECT, CPalette* pPal);
BOOL      WINAPI  DrawBitmap ( HDC hdc, INT x, INT y, HBITMAP hbm, DWORD rop);
BOOL      WINAPI  DibBlt ( HDC hdc, INT x0, INT y0, INT dx, INT dy,
                           HANDLE hdib, INT x1, INT y1, LONG rop);
BOOL      WINAPI  StretchDibBlt ( HDC hdc, INT x, INT y, INT dx, INT dy, HANDLE hdib,
                                   INT x0, INT y0, INT dx0, INT dy0, LONG rop);                                  

HPALETTE  WINAPI  CreateBIPalette (LPBITMAPINFOHEADER lpbi);
BOOL      WINAPI  CreateDIBPalette(HDIB hDIB, CPalette* cPal);
HPALETTE  WINAPI  CreateDibPalette (HANDLE hbi);

//VOID      WINAPI  AppPaint (HWND hWnd, HDC hDC, INT x, INT y);

#endif //!_INC_DIBAPI
