// multipad.cpp : Defines the class behaviors for the Multipad application.
//
// This is a part of the Microsoft Foundation Classes C++ library.
// Copyright (C) 1992-1995 Microsoft Corporation
// All rights reserved.
//
// This source code is only intended as a supplement to the
// Microsoft Foundation Classes Reference and related
// electronic documentation provided with the library.
// See these sources for detailed information regarding the
// Microsoft Foundation Classes product.

#include "stdafx.h"
#include "multipad.h"
#include "VisioApp.h"
//#include <locale.h>

IMPLEMENT_DYNCREATE(CPadDoc, CDocument)

BEGIN_MESSAGE_MAP(CPadDoc, CDocument)
	//{{AFX_MSG_MAP(CPadDoc)
	ON_COMMAND(ID_DELETECONTENTS, OnDeletecontents)
	//}}AFX_MSG_MAP
  /*
  */
END_MESSAGE_MAP()

void CPadDoc::Serialize(CArchive& ar)
{
	((CEditView*)m_viewList.GetHead())->SerializeRaw(ar);
}

CPadView* CPadDoc::GetView()
{
   POSITION pos = GetFirstViewPosition();
   if (pos != NULL) {
      CPadView* pView = (CPadView*) GetNextView(pos);
      ASSERT_VALID (pView);
      return (CPadView*) pView; 
   } else return NULL;
}


/////////////////////////////////////////////////////////////////////////////
void CPadDoc::OnDeletecontents() 
{
	// TODO: Add your command handler code here
  ((CEditView*)m_viewList.GetHead())->DeleteContents();
}

/////////////////////////////////////////////////////////////////////////////
BOOL CPadDoc::CanCloseFrame( CFrameWnd* pFrame )
{
  CPadView* pPadView = (CPadView*) m_viewList.GetHead();
  // if vscanf hasn't finised, don't allow close frame
  // (otherwise the program fails!!)
  if (pPadView->IsScanning()) return FALSE;

  // Free pPadView
  CVisioApp* pApp = (CVisioApp*)AfxGetApp();
  if ( pPadView == pApp->m_pPadView ) {
      /* Don't ask to save document on close */
      SetModifiedFlag(FALSE);
  }
	return CDocument::CanCloseFrame(pFrame);
}
/////////////////////////////////////////////////////////////////////////////

int print(const char *format, ...)
{
  va_list args;
  int retval;
 
  va_start(args, format);
  if (format) 
      retval=((CVisioApp*)AfxGetApp())->GetTextView()
                 ->vprintf( format, args);
  va_end(args);

  return retval;
}

int scan(const char *format, ...)
{
  va_list args;
  int retval;
 
  va_start(args, format);
      retval=((CVisioApp*)AfxGetApp())->GetTextView()
                 ->vscanf( format, args);
  va_end(args);

  return retval;
}

////////////////////////////////////////////////////////////////////////////
CPadView::CPadView() : m_bScanning(FALSE), m_bReadOnly(FALSE), m_ScanChar(0)
{
}

CPadView::~CPadView()
{
}


IMPLEMENT_DYNCREATE(CPadView, CEditView)
BEGIN_MESSAGE_MAP(CPadView, CEditView)
	//{{AFX_MSG_MAP(CPadView)
	ON_WM_CHAR()
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_CONTROL_REFLECT(EN_MAXTEXT, OnMaxtExt)
	ON_CONTROL_REFLECT(EN_ERRSPACE, OnErrSpace)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// We use the standard C form for multiple parameters

#define _LINE_FEED_

int CPadView::vprintf(const char *format, va_list argptr )
{
  char buffer[256];
  int retval;
 
  if (format) 
      retval=vsprintf(buffer, format, argptr);
  else
     return 0;
 
#ifdef _LINE_FEED_
  char buffer2[300];
  char *dst=buffer2, *src=buffer;
  /* check single formfeed characters are translated
   * to carriage return  plus form-feed*/
  for (; *src; src++, dst++) {
    if ((*dst = *src) == 10) /* form feed */
      if (src[-1] != 13) { /* carriage return */ 
        *dst=13;
        dst++;
        *dst=10; 
      }
  }
  *dst = '\0';
#endif //_LINE_FEED_

	CEdit& rEdit = GetEditCtrl();
  if (&rEdit == NULL) return 0;
  UINT nchars = GetBufferLength( );
  rEdit.SetSel( (int) nchars, (int) nchars);

#ifdef _LINE_FEED_
  rEdit.ReplaceSel(buffer2);
#else //_LINE_FEED_
  rEdit.ReplaceSel(buffer);
#endif //_LINE_FEED_

  // Don't ask to save document on close
  GetDocument()->SetModifiedFlag(FALSE);

  return retval;
}

void CPadView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	if (m_bScanning) {
    if ( nChar == VK_CANCEL || nChar == VK_RETURN) {
      // End Modal Loop (scanning state)
      EndModalLoop( 1 );
    } else {
      /* Are we writing before the scanning line ? */
      int nStartChar, nEndChar;
      CEdit& rEdit = GetEditCtrl();
      rEdit.GetSel(nStartChar, nEndChar);
      // Set real state
      if ( m_bReadOnly && (nStartChar>m_ScanChar))
        rEdit.SetReadOnly(m_bReadOnly = FALSE);
      else if (!m_bReadOnly && (nStartChar<m_ScanChar))
        rEdit.SetReadOnly(m_bReadOnly = TRUE);
    }
  }
	CEditView::OnChar(nChar, nRepCnt, nFlags);
}

void CPadView::OnUpdateEditPaste(CCmdUI* pCmdUI) 
{
	// TODO: Add your command update UI handler code here
  CEdit& rEdit = GetEditCtrl();
  if (m_bScanning) {
    int nStartChar, nEndChar;  
    rEdit.GetSel(nStartChar, nEndChar);
    if ( m_bReadOnly && (nStartChar>m_ScanChar))
      rEdit.SetReadOnly(m_bReadOnly = FALSE);
    else if (!m_bReadOnly && (nStartChar<m_ScanChar))
      rEdit.SetReadOnly(m_bReadOnly = TRUE);
    pCmdUI->Enable(!m_bReadOnly);
    rEdit.SetReadOnly(nStartChar<m_ScanChar);
  } else {
    pCmdUI->Enable(TRUE);
  }
}

void CPadView::OnErrSpace() 
{
	// TODO: Add your control notification handler code here
  int aceptar =
	MessageBox("Se ha producido un error por falta de memoria.\n"
    "Pulse 'Aceptar' para borrar "
    "el contenido del visualizador y liberar espacio",
    "Fallo de memoria",MB_ICONEXCLAMATION | MB_OKCANCEL );
  if (aceptar == IDOK) 
    DeleteContents(); // Borra el contenido de la vista
}

void CPadView::OnMaxtExt() 
{
	// TODO: Add your control notification handler code here
  int aceptar =
  MessageBox("Se ha alcanzado la capacidad mxima del visualizador "
    "de texto (aprox. 64.000 caracteres).\n Pulse 'Aceptar' para borrar "
    "el contenido del visualizador y liberar espacio",
    "Capacidad mxima",MB_ICONEXCLAMATION | MB_OKCANCEL );
  if (aceptar == IDOK) 
    DeleteContents(); // Borra el contenido de la vista
}

void CPadView::StopScanning()
{
  if (m_bScanning) { // Finish scanning process!
      EndModalLoop( 1 );
      LONG nTimes = 100;
        // wait until scanning process is finished!
        // (be sure the window is closed)
      //do {} while (AfxGetApp()->OnIdle(nTimes)); 
      AfxGetApp()->OnIdle(nTimes);
      AfxGetApp()->OnIdle(nTimes);
      AfxGetApp()->OnIdle(nTimes);
  }  	
}
/////////////////////////////////////////////////////////////////////////////
// CPadFrame

IMPLEMENT_DYNCREATE(CPadFrame, CMDIChildWnd)

CMenu CPadFrame::m_menu;

CPadFrame::CPadFrame()
{
}

CPadFrame::~CPadFrame()
{
}


BEGIN_MESSAGE_MAP(CPadFrame, CMDIChildWnd)
	//{{AFX_MSG_MAP(CPadFrame)
	ON_WM_CLOSE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPadFrame message handlers

void CPadFrame::OnClose() 
{
	// TODO: Add your message handler code here and/or call default
  CPadView* pPadView = (CPadView*) GetActiveView();
  pPadView->StopScanning();
  CVisioApp* pApp = (CVisioApp*)AfxGetApp();
  if ( pPadView == pApp->m_pPadView ) {
      pApp->m_pPadView = NULL;
      /* Don't ask to save document on close */
      pPadView->GetDocument()->SetModifiedFlag(FALSE);
  }
	CMDIChildWnd::OnClose();
}


BOOL CPadFrame::Create(  LPCTSTR lpszCDibViewClass, LPCTSTR lpszWindowName, 
                         DWORD dwStyle/* = 0 */, const RECT& rect/* = rectDefault */,
                         CMDIFrameWnd* pParentWnd /* = NULL */,
                         CCreateContext* pContext /* = NULL */) 

{
	// TODO: Add your specialized code here and/or call the base class

  // Setup the shared menu
	if (m_menu.m_hMenu == NULL)
    	if (!m_menu.LoadMenu(IDR_TEXTTYPE))
		    return FALSE;

  // when the list output frame window is active, use this menu
	m_hMenuShared = m_menu.m_hMenu;

	// Register a custom WndClass and create a window.
	// This must be done because CImgBasicDoc has a custom icon.

  if (lpszCDibViewClass == NULL) 
	  lpszCDibViewClass =
		  AfxRegisterWndClass(CS_HREDRAW | CS_VREDRAW,
			LoadCursor(NULL, IDC_ARROW),
			(HBRUSH) (COLOR_WINDOW+1),
			LoadIcon(AfxGetInstanceHandle(), MAKEINTRESOURCE(IDR_TEXTTYPE)));

	// create the CMDIChildWnd
	return CMDIChildWnd::Create(lpszCDibViewClass, lpszWindowName,
                              dwStyle, rect, pParentWnd, pContext);

}

