/*********************************************************************************\
*     File:                  P I C 2 D I B  . C P P                               *
*                                                                                 *
*     i/o routines for .BMP files (MS Windows 3.x)                                *
*     Standard   C++                                                              *
*                                                                                 *
*                                                                                 *
\*********************************************************************************/

/* Proyecto Fin de Carrera
 * Departamento de Informatica e Ingeniera de Sistemas
 * Universidad de Zaragora
 *
 * rea:      Robtica y Automatizacin Avanzada.
 * Tema:      Visin por Computador.
 * Ttulo:    Diseo de un entorno de visin para 
 *            el tratamiento de imgenes.
 * Director:  Dr. Carlos Sags Blzquiz.
 * Autor:     Joaqun Mur Amada.
 * Promocin: 1991-1996.
*/

#include "stdafx.h"
#include "..\\res\\resource.h"
#include "dibdoc.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif


/*******************************************/
static unsigned int getshort( byte* &fp)
{
  unsigned int c, c1;
  c = *fp++;  c1 = *fp++;
  return c | (c1 << 8);
}


/*******************************************/
static unsigned int getint( byte* &fp)
{
  unsigned int c, c1, c2, c3;
  c = *fp++; c1 = *fp++; c2 = *fp++;  c3 = *fp++;
  return c | (c1 << 8) | (c2 << 16) | (c3 << 24);
}


/*******************************************/
static void putshort( byte* &fp, int i)
{
  *fp++ = (byte) ((unsigned int ) i) & 0xff;
  *fp++ = (byte) (((unsigned int) i)>>8) & 0xff;
}


/*******************************************/
static void putint(byte* &fp, int i)
{
  *fp++ = (byte) ((unsigned int ) i)      & 0xff;  
  *fp++ = (byte) (((unsigned int) i)>>8)  & 0xff;
  *fp++ = (byte) (((unsigned int) i)>>16) & 0xff;
  *fp++ = (byte) (((unsigned int) i)>>24) & 0xff;
}


/*******************************************************************************
********************************************************************************/

  
    /* comments on error handling:
   not being able to malloc is a Fatal Error.  The program is aborted. */
#ifndef BI_RGB
#define BI_RGB  0
#endif
#ifndef BI_RLE8
#define BI_RLE8 1
#endif
#ifndef BI_RLE4
#define BI_RLE4 2
#endif

#define WIN_OS2_OLD 12
#define WIN_NEW     40
#define OS2_NEW     64

#define UINT32 DWORD

 int   SetIMAGE1   (byte *, byte *, u_int, u_int);
 int   SetIMAGE4   (byte *, byte *, u_int, u_int, u_int);
 int   SetIMAGE8   (byte *, byte *, u_int, u_int, u_int);
 int   SetIMAGE24  (byte *, byte *, u_int, u_int);
 void  GetHDIB1    (byte *, byte *, u_int, u_int);
 void  GetHDIB4    (byte *, byte *, u_int, u_int);
 void  GetHDIB8    (byte *, byte *, u_int, u_int);
 void  GetHDIB24   (byte *, byte *, u_int, u_int);


/***********************************************************************/
/*                          B M P   -->    P I C                       */
/***********************************************************************/
int CDibDoc::SetIMAGE(HDIB hDIB)
{
  int          rv;
  unsigned int biSize, biWidth, biHeight, biPlanes;
  unsigned int biBitCount, biCompression, biSizeImage, biXPelsPerMeter;
  unsigned int biYPelsPerMeter, biClrUsed, biClrImportant;
  byte         *pic24, *pic8;
  byte *ptr;
  char          buf[128];
  int cmaplen; //num of colors


  //Check for a non-null pointer
  if (hDIB==NULL) return 0;

  WaitCursor();

  byte* pDIB = (byte*) GlobalLock(hDIB);

  /* returns '1' on success */

  pic8 = pic24 = (byte *) NULL;

    /*
	 * Get a pointer to the DIB memory, the first of which contains
	 * a BITMAPINFO structure
	 */
  
  ptr = pDIB;
  biSize = getint(ptr);

  if ( biSize == WIN_NEW || biSize == OS2_NEW ) {
    biWidth         = getint(ptr);
    biHeight        = getint(ptr);
    biPlanes        = getshort(ptr);
    biBitCount      = getshort(ptr);
    biCompression   = getint(ptr);
    biSizeImage     = getint(ptr);
    biXPelsPerMeter = getint(ptr);
    biYPelsPerMeter = getint(ptr);
    biClrUsed       = getint(ptr);
    biClrImportant  = getint(ptr);
  }

  else {    /* old bitmap format */
    biWidth         = getshort(ptr);          /* Types have changed ! */
    biHeight        = getshort(ptr);
    biPlanes        = getshort(ptr);
    biBitCount      = getshort(ptr);
    /* We have to compute the rest of values for old bitmaps */
    biCompression   = BI_RGB;
    /* Not in old versions so have to compute them*/
    biSizeImage = (((biPlanes * biBitCount*biWidth)+31)/32)*4*biHeight;    
    biCompression   = BI_RGB; 
    biXPelsPerMeter = biYPelsPerMeter = 0;
    biClrUsed       = biClrImportant  = 0;
  }

  if (DEBUG>1) {
    debug_printf("\t\tbiSize=%d, biWidth=%d, biHeight=%d, biPlanes=%d\n",
	    biSize, biWidth, biHeight, biPlanes);
    debug_printf("\t\tbiBitCount=%d, biCompression=%d, biSizeImage=%d\n",
	    biBitCount, biCompression, biSizeImage);
    debug_printf("\t\tbiX,YPelsPerMeter=%d,%d  biClrUsed=%d, biClrImp=%d\n",
	    biXPelsPerMeter, biYPelsPerMeter, biClrUsed, biClrImportant);
  }

  /* error checking */
  if ((biBitCount!=1 && biBitCount!=4 && biBitCount!=8 && biBitCount!=24) || 
      biPlanes!=1 || biCompression>BI_RLE4) 
  {
    ArrowCursor();
    GlobalUnlock(hDIB);
    sprintf(buf,"SetIMAGE\nDatos DIB en memoria Corruptos!\n"
                "(bitCount=%d, Planes=%d, Compression=%d)",
	              biBitCount, biPlanes, biCompression);
    Warning(buf);    
    return 0;
  }

  if (((biBitCount==1 || biBitCount==24) && biCompression != BI_RGB) ||
      (biBitCount==4 && biCompression==BI_RLE8) ||
      (biBitCount==8 && biCompression==BI_RLE4))
  {
    ArrowCursor();
    GlobalUnlock(hDIB);
    sprintf(buf,"SetIMAGE\nBogus BMP data pointer!  (bitCount=%d, Compression=%d)",
	    biBitCount, biCompression);
    Error( buf);
    return 0;
  }

  ptr = pDIB + biSize;

  /* load up colormap, if any */
  if (biBitCount!=24) 
  {
    int i;

    cmaplen = (biClrUsed) ? biClrUsed : 1 << biBitCount;
    for (i=0; i<cmaplen; i++) {
      bMap[i] = *ptr++;
      gMap[i] = *ptr++;
      rMap[i] = *ptr++;
      if (biSize != WIN_OS2_OLD) {
        ptr++;
      }
    }

    if (DEBUG>1)
      debug_printf("SetIMAGE:  BMP colormap:  (RGB order)\n");
   }

  /*
  if (biSize != WIN_OS2_OLD) {
    // Waste any unused bytes between the colour map (if present)
    // and the start of the actual bitmap data.
    ptr += bPad;
  }
  */

  /* create pic8 or pic24 */

  if (biBitCount==24) {
    pic24 = (byte *) calloc((size_t) biWidth * biHeight * 3, (size_t) 1);
    if (!pic24) {
      ArrowCursor();
      GlobalUnlock(hDIB);
      Error("SetIMAGE\nFallo al requerir memoria para 'pic24'");
      return 0;
    }
  }
  else {
    pic8 = (byte *) calloc((size_t) biWidth * biHeight, (size_t) 1);
    if (!pic8) {
      ArrowCursor();
      GlobalUnlock(hDIB);
      Error("SetIMAGE\nFallo al requerir memoria para 'pic8'");
      return 0;
    }
  }

  WaitCursor();

  /* load up the image */
  if      (biBitCount == 1) rv = SetIMAGE1(ptr,pic8,biWidth,biHeight);
  else if (biBitCount == 4) rv = SetIMAGE4(ptr,pic8,biWidth,biHeight,
					  biCompression);
  else if (biBitCount == 8) rv = SetIMAGE8(ptr,pic8,biWidth,biHeight,
					  biCompression);
  else                      rv = SetIMAGE24(ptr,pic24,biWidth,biHeight);

  if (rv) {
    Error("SetIMAGE\nDatos en memoria parecen estar corruptos."
                "Rellenados con 0's.\n");
  }

  // Free IOfile's pic and save real image in case undo    
  StoreNewPic(NULL,PIC8,FALSE); // free origPic; pic -> origPic

  if (biBitCount == 24) {
    pic  = pic24;
    picType = PIC24;
    NumCols=numcols=mono=0;
  }
  else {
    pic  = pic8;
    picType = PIC8;
    mono=0;
    NumCols=numcols=cmaplen;
  }

  pWIDE = biWidth;  pHIGH = biHeight;
  
  ArrowCursor();
  GlobalUnlock(hDIB);

  // Update all views
  InitDibData(TRUE);
    
  return 1;
}  


/*******************************************/
int SetIMAGE1(byte *fp, byte *pic8, u_int w, u_int h)
{
  int   i,j,c,bitnum,padw;
  byte *pp;
  c = 0;
  padw = ((w + 31)/32) * 32;  /* 'w', padded to be a multiple of 32 */

  for (i=h-1; i>=0; i--) {
    pp = pic8 + (i * w);
    for (j=bitnum=0; j<padw; j++,bitnum++) {
      if ((bitnum&7) == 0) { /* read the next byte */
	      c = *fp++; 
	      bitnum = 0;
      }
      
      if (j<w) {
	      *pp++ = (byte) (c & 0x80) ? 1 : 0;
	      c <<= 1;
      }
    }
  }
  return 0;
}  



/*******************************************/
int SetIMAGE4(byte *fp, byte *pic8, u_int w, u_int h, u_int comp)
{
  int   i,j,c,c1,x,y,nybnum,padw;
  byte *pp;
  
  
  c = c1 = 0;
  
  if (comp == BI_RGB) {   /* read uncompressed data */
    padw = ((w + 7)/8) * 8; /* 'w' padded to a multiple of 8pix (32 bits) */
    
    for (i=h-1; i>=0; i--) {
      pp = pic8 + (i * w);

      for (j=nybnum=0; j<padw; j++,nybnum++) {
	      if ((nybnum & 1) == 0) { /* read next byte */
	        c = *fp++;
	        nybnum = 0;
	      }
	
	      if (j<w) {
	        *pp++ = (byte) (c & 0xf0) >> 4;
	        c <<= 4;
	      }
      }
    }
  }
  
  else if (comp == BI_RLE4) {  /* read RLE4 compressed data */
    x = y = 0;  
    pp = pic8 + x + (h-y-1)*w;
    
    while (y<h) {
      c = *fp++;
      
      if (c) {                                   /* encoded mode */
	    c1 = *fp++;
	    for (i=0; i<c; i++,x++,pp++) 
	      *pp = (byte) (i&1) ? (c1 & 0x0f) : ((c1>>4)&0x0f);
        }
  
      else {    /* c==0x00  :  escape codes */
	    c = *fp++;
	    
	    if      (c == 0x00) {                    /* end of line */
	      x=0;  y++;  pp = pic8 + x + (h-y-1)*w;
	    } 
	
	else if (c == 0x01) break;               /* end of pic8 */
	
	else if (c == 0x02) {                    /* delta */
	  c = *fp++;  x += c;
	  c = *fp++;  y += c;
	  pp = pic8 + x + (h-y-1)*w;
	}
	
	else {                                   /* absolute mode */
	  for (i=0; i<c; i++, x++, pp++) {
	    if ((i&1) == 0) c1 = *fp++;
	    *pp = (byte) (i&1) ? (c1 & 0x0f) : ((c1>>4)&0x0f);
	  }
	  
	  if (((c&3)==1) || ((c&3)==2)) *fp++;  /* read pad byte */
	}
      }  /* escape processing */
    }  /* while */
  }
  
  else {
    debug_printf("unknown BMP compression type 0x%0x\n", comp);
    return 1;
  }
  return 0;
}  



/*******************************************/
int SetIMAGE8(byte *fp, byte *pic8, u_int w, u_int h, u_int comp)
{
  int   i,j,c,c1,padw,x,y;
  byte *pp;
  
  if (comp == BI_RGB) {   /* read uncompressed data */
    padw = ((w + 3)/4) * 4; /* 'w' padded to a multiple of 4pix (32 bits) */

    for (i=h-1; i>=0; i--) {
      pp = pic8 + (i * w);

      for (j=0; j<padw; j++) {
	    c = *fp++;                                
	    if (j<w) *pp++ = (byte) c;
      }
    }
  }

  else if (comp == BI_RLE8) {  /* read RLE8 compressed data */
    x = y = 0;  
    pp = pic8 + x + (h-y-1)*w;

    while (y<h) {
      c = *fp++;

      if (c) {                                   /* encoded mode */
	    c1 = *fp++;
	    for (i=0; i<c; i++,x++,pp++) *pp = (byte) c1;
      }

      else {    /* c==0x00  :  escape codes */
	    c = *fp++;

	    if      (c == 0x00) {                    /* end of line */
	      x=0;  y++;  pp = pic8 + x + (h-y-1)*w;
	  } 

	  else if (c == 0x01) break;               /* end of pic8 */

	  else if (c == 0x02) {                    /* delta */
	    c = *fp++;  x += c;
	    c = *fp++;  y += c;
	    pp = pic8 + x + (h-y-1)*w;
	  }
      
      else {                                   /* absolute mode */
	    for (i=0; i<c; i++, x++, pp++) {
	      c1  = (byte) *fp++;
	      *pp = (byte) c1;
	    }
	    
	    if (c & 1) *fp++;  /* odd length run: read an extra pad byte */
	  }
     }  /* escape processing */
    }  /* while */
  }
  
  else {
    debug_printf("unknown BMP compression type 0x%0x\n", comp);
    return 1;
  }

  return 0;
}  



/*******************************************/
int SetIMAGE24(byte *fp, byte *pic24, u_int w, u_int h)
{
  int   i,j,padb;
  byte *pp;

  padb = (4 - ((w*3) % 4)) & 0x03;  /* # of pad bytes to read at EOscanline */

  for (i=h-1; i>=0; i--) {
    pp = pic24 + (i * w * 3);
    for (j=0; j<w; j++) {
      pp[2] = (byte) *fp++;   /* blue */
      pp[1] = (byte) *fp++;   /* green */
      pp[0] = (byte) *fp++;   /* red */
      pp += 3;
    }

    for (j=0; j<padb; j++) *fp++;
  }

  return 0;
}  




/***********************************************************************/
/*                          P I C   -->    B M P                       */
/***********************************************************************/
HDIB CDibDoc::GetHDIB ()
{ 
  /* format and color type with which the image is being saved */
  byte *thepic, *rp, *gp, *bp;
  int   ptype=PIC8, pfree=0,colType=F_REDUCED;
  TRACE("\n\nGetHDIB\n\n");
  // Does the DC support palettes?
  if (  UsePalette ) { 
    /* retrieve a colormaped image */
    /*NumCols=MAX_ALLOCABLE_COLS;  Maximum alocable colors */
    thepic = GetImage(&ptype, colType, &pfree, &NumCols, &rp, &gp, &bp,
                        MAX_ALLOCABLE_COLS);
    HDIB hdib = GetHDIB ( thepic ,ptype , pWIDE, pHIGH,rp, gp, bp, NumCols );
    if (pfree) free (thepic); // delete temporal image
    return hdib;
  }
  //Update # of colors in DibDoc with IOfile' info
  NumCols=numcols;
  return GetHDIB ( pic ,picType , pWIDE, pHIGH, rMap, gMap, bMap, NumCols ); 
}


HDIB CDibDoc::GetHDIB ( byte *pic824, int ptype, u_int w, u_int h,
                          byte *rmap, byte *gmap, byte *bmap,
                          u_int ncolors )
{
  /*
   * if PIC8 the program writes an uncompressed 1-, 4- or 8-bit image
   * (depending on the value of ncolors)
   *
   * if PIC24 the program writes an uncompressed 24-bit image
   */

  /* This algorithm don't sort the colormap since it is supposed to be if
   * the image is PIC8 */

  u_int i, nbits=0, bperlin=0;
  byte *dib, *fp;

  if (ptype == PIC24) {  /* is F_FULLCOLOR */
    nbits = 24;
    ncolors = 0; 
  }

  else if (ptype == PIC8) {
    
    /* determine how many bits per pixel we'll be writing */
    if ( ncolors <= 2) nbits = 1;
    else if (ncolors<=16) nbits = 4;
    else nbits = 8;

  } // end of PIC8


  bperlin = ((w * nbits + 31) / 32) * 4;   /* # bytes written per line */

  
  /* compute filesize and write it */
  i = 40 +                /* size of bitmap info header */
      (ncolors * 4) +          /* size of colormap */
      bperlin * h;        /* size of image data */

  /* get memory for the bitmap */
  HDIB hDIB = (HDIB) GlobalAlloc(GHND, (size_t) i);
  dib = fp = (byte *) GlobalLock((HGLOBAL) hDIB);
  if (!fp) {
    FatalError("unable to malloc in GetHDIB()");
    GlobalFree(hDIB);
    return NULL;
  }
  
  
  /******* BITMAPINFO structure *********
  typedef struct tagBITMAPINFO { // bmi 
   BITMAPINFOHEADER bmiHeader; 
   RGBQUAD          bmiColors[1]; 
  } BITMAPINFO; 
  */

  /* bitmap info header (BITMAPINFO->BITMAPINFOHEADER)*/
  putint(fp, 40);         /* biSize: size of bitmap info header */
  putint(fp, w);          /* biWidth */
  putint(fp, h);          /* biHeight */
  putshort(fp, 1);        /* biPlanes:  must be '1' */
  putshort(fp, nbits);    /* biBitCount: 1,4,8, or 24 */
  putint(fp, BI_RGB);     /* biCompression:  BI_RGB, BI_RLE8 or BI_RLE4 */
  putint(fp, bperlin*h);  /* biSizeImage:  size of raw image data */
  putint(fp, 75 * 39);    /* biXPelsPerMeter: (75dpi * 39" per meter) */
  putint(fp, 75 * 39);    /* biYPelsPerMeter: (75dpi * 39" per meter) */
  putint(fp, ncolors);         /* biClrUsed: # of colors used in cmap */
  putint(fp, ncolors);         /* biClrImportant: same as above */


  /* write out the colormap (BITMAPINFO->RGBQUAD)*/
  for (i=0; i<ncolors; i++) {
    *fp++ = bmap[i];
    *fp++ = gmap[i];
    *fp++ = rmap[i];
    *fp++ = 0;
  }

  /* write out the image (pixel array) */

  if      (nbits ==  1) GetHDIB1 (fp, pic824, w, h);
  else if (nbits ==  4) GetHDIB4 (fp, pic824, w, h);
  else if (nbits ==  8) GetHDIB8 (fp, pic824, w, h);
  else if (nbits == 24) GetHDIB24(fp, pic824, w, h);
  GlobalUnlock(hDIB);
  return hDIB;
}


	  
	  
/*******************************************/
void GetHDIB1(byte *fp, byte *pic8, u_int w, u_int h)
{
  u_int j,k,bytes,pad,c;
  byte* pp;
  
  bytes= (w>>3)+1;    /* this bytes+1 can be written without care about padding */
  pad =  ((32-(w & 31))&31)>>3;   /* (32 - w%32)/31 'w', padded to be a multiple of 32 */
                              
  for (; h;) {  /* row loop */
    pp =  pic8 + (--h)*w ;
    for (j=bytes; --j;) {  /* column loop -non padded bytes */
      /* write the next DWORD */
      c = 0;
      for (k=8; --k;) {
        c |= (*pp )? 1 : 0;
        c <<= 1;
        pp++;
      }
      /* write the next DWORD */
	    *fp++ = c|((*pp )? 1 : 0);
      pp++;
    }
    /* padded byte */
    if (k= w&7) {
      c = 0;
      for (;k;k--) {
        c |= (*pp )? 1 : 0;
        c <<= 1;
        pp++;
      }
      /* write last DWORD */
	    *fp++ = (c|( (*pp ) ? 1 : 0))<<(8-(w&7));
    }
    /* last entire bytes */
    for(k=pad;k;k--)
      *fp++ = 0;
  }
}  



/*******************************************/
void GetHDIB4(byte *fp, byte *pic8, u_int w, u_int h)
{
  u_int j,bytes,pad;
  byte c;
  byte* pp;
  
  bytes= (w>>1)+1;         /* w/2+1 this bytes+1 can be written without care
                              about padding */
  pad =  ((8-(w&7))&7)>>1; /* (8 - w%8)/2 bytes padded to be a multiple of 32 */
                              
  for (; h;) {  /* row loop */
    pp =  pic8 + --h*w ;
    for (j=bytes; --j;) {  /* column loop -non padded bytes */
      /* write the next byte */
      c = *pp<<4;
      pp++;
      *fp++ = c | *pp;
      pp++;
    }
    
    if (w&1) *fp++ = *pp<<4; 

    /* padded DWORD (end of row) */
    for (j=pad; j;j--)
      *fp++ = 0;
    
  }
}  

/*******************************************/
void GetHDIB8(byte *fp, byte *pic8, u_int w, u_int h)
{
  u_int j,pad;
  byte* pp;
  
  pad = (4 - (w & 3))&3;   /* 'w' padded to a multiple of 4pix (32 bits) */
                              
  for (;h;) {  /* row loop */
    pp = pic8 + --h*w;
    /* column loop -non padded bytes- */
    for (j=w; j; j--) *fp++ = *pp++;  
    
    /* padded padded (end of row) */
    for (j=pad; j; j--) *fp++ = 0;  
  }
}


/*******************************************/
void GetHDIB24(byte *fp, byte *pic24, u_int w, u_int h)
{
  int   j,padb;
  byte *pp;

  padb = (4 - ((w*3) % 4)) & 0x03;  /* # of pad bytes to write at EOscanline */

  for (;h;) {
    pp = pic24 + (--h * w * 3);
    for (j=w; j; j--) {
      *fp++ = pp[2];
      *fp++ = pp[1];
      *fp++ = pp[0];
      pp += 3;
    }

    for (j=padb; j; j--) *fp++ = 0;
  }
}  

/////////////////////////////////////////////////////////////////////////

