/***
*sscanf.c - read formatted data from string
*
*       Copyright (c) 1985-1995, Microsoft Corporation. All rights reserved.
*
*Purpose:
*       defines scanf() - reads formatted data from string
*
*******************************************************************************/
#include "stdafx.h"
#include "multipad.h"

/***
*int sscanf(string, format, ...) - read formatted data from string
*
*Purpose:
*       Reads formatted data from string into arguments.  _input does the real
*       work here.  Sets up a FILE so file i/o operations can be used, makes
*       string look like a huge buffer to it, but _filbuf will refuse to refill
*       it if it is exhausted.
*
*       Allocate the 'fake' _iob[] entryit statically instead of on
*       the stack so that other routines can assume that _iob[] entries are in
*       are in DGROUP and, thus, are near.
*
*       Multi-thread: (1) Since there is no stream, this routine must never try
*       to get the stream lock (i.e., there is no stream lock either).  (2)
*       Also, since there is only one staticly allocated 'fake' iob, we must
*       lock/unlock to prevent collisions.
*
*Entry:
*       char *string - string to read data from
*       char *format - format string
*       followed by list of pointers to storage for the data read.  The number
*       and type are controlled by the format string.
*
*Exit:
*       returns number of fields read and assigned
*
*Exceptions:
*
*******************************************************************************/

//extern "C" int __cdecl _input(FILE *, const unsigned char *, va_list);
/* Nmero mximo de parmetros extra que se puede pasar a la funcin scanf */
#define MAX_VA_ARGS 20

int CPadView::vscanf ( const char *format, va_list arglist)
/*
 * 'V'ariable 'S'tring 'SCAN', 'F'ormatted
 */
{
  //va_list arglist;
  /* Don't do nothing if we are
   * scanning yet */
  if (m_bScanning) return 0;

  m_bScanning = TRUE;
  m_ScanChar = GetBufferLength();
  
  // Set cursor in last char of view
  GetEditCtrl().SetSel( m_ScanChar, m_ScanChar);

  CFrameWnd* pWnd = (CMDIChildWnd*) GetParentFrame();
  pWnd->BringWindowToTop( );
  TRACE("Lanzado GetParentFrame()->BringWindowToTop()","CPadView::scanf");
  if (  pWnd->IsIconic() ) {
    WINDOWPLACEMENT lpwndpl;
    pWnd->GetWindowPlacement( &lpwndpl ) ;
    lpwndpl.showCmd = SW_RESTORE;
    pWnd->SetWindowPlacement( &lpwndpl ) ;
  };

  // enter modal loop
	RunModalLoop(MLF_SHOWONIDLE);
  // modal loop exited
	int retval;
  {
    // execute sscanf code with the text above the last
    // position before CView::scanf funcion call
	  ASSERT_VALID(this);
	  LPCTSTR string = LockBuffer();
	  ASSERT(string != NULL);
    {          
      /* array with arguments (they really are pointers to
       * chars, ints, strings, floats...) so they can be
       * cast to void pointer */
      void* arg[MAX_VA_ARGS];
      char *ptr = (char*) format;
      int nargs = 0;
      //va_start(arglist, format);

      /* compute number of arguments with format specifiers */
      /* (one parameter per '%') */
      for(; *ptr ; ptr++)
        if (*ptr=='%') {
          if (nargs==MAX_VA_ARGS) {
            MessageBox("El nmero de parmetros pasados a CPadView::scanf es excesivo");
            return(0);
          }
          arg[nargs] = va_arg(arglist, void*);
          nargs++;
        }

      if(nargs<=3) { /* reduced version to speed up*/
        for(int i=nargs; i<3; i++) {
          arg[nargs] = NULL;
        }
        retval = ::sscanf( (char *) string + m_ScanChar,format,
                 arg[0],arg[1],arg[2]);
      } else { /* complete version, but a bit slow */
        for(int i=nargs; i<MAX_VA_ARGS; i++) {
          arg[nargs] = NULL;
        }
        retval = ::sscanf( (char *) string + m_ScanChar,format,
          arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],arg[6],arg[7],arg[8],arg[9],
          arg[10],arg[11],arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],arg[18],arg[19]
          );
      }

      va_end (arglist);
    }
    UnlockBuffer();
	  ASSERT_VALID(this);
  }
  
  GetEditCtrl().SetReadOnly(FALSE);
  m_ScanChar = 0; // Safe Value
  m_bScanning = FALSE;

  // Don't ask to save document on close
  GetDocument()->SetModifiedFlag(FALSE);

  // number of scanned parameters
  return(retval);
}

/*******************************************************************************/
